// BenLincoln.DaVinci.GUI - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.GUI.

// BenLincoln.DaVinci.GUI is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci.GUI is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci.GUI (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace BenLincoln.DaVinci.GUI
{
    public class ClickableThumbnailPictureBox : PictureBox
    {
        // minimum number of milliseconds between image opening events
        // for some reason, the context menus cause images to open twice
        // even though the event handler is only set up once.
        protected const int _MinimumTimeBetweenImageOpeningEvents = 50;

        // This value controls whether or not ToolTips for this control are enabled
        // It is currently set to false because of a bug in .NET 2.0 in which ToolTips
        // leak approximately 2 GDI handles. Unlike the (patched) bug of the same type 
        // in .NET 1.1, explicitly calling Dispose() on the ToolTip does not function
        // as an effective workaround
        protected const bool _EnableToolTips = false;

        protected string[] _ImagePaths;
        protected int _PrimaryImageIndex;
        protected int _ThumbnailSourceImageIndex;
        protected string _ImageName;
        protected Hashtable _MenuToIndexHashtable;
        //protected ArrayList _ToolTipList;
        protected int _ImageMargin;
        //protected Image _ThumbnailImage;
        protected ClickableThumbnailContextMenu _ContextMenu;

        protected bool _PopupIsOpen;

        protected DateTime _LastImageOpenEventTime;

        protected ClickableThumbnailToolTip _ToolTip;

        protected bool _ParentDisplayShouldSlide;

        #region Properties

        //public SlidingThumbnailDisplay ParentDisplay
        //{
        //    get
        //    {
        //        return _ParentDisplay;
        //    }
        //    set
        //    {
        //        _ParentDisplay = value;
        //    }
        //}

        public string[] ImagePaths
        {
            get
            {
                return _ImagePaths;
            }
            set
            {
                _ImagePaths = value;
                if ((_ImagePaths != null) && (_ImagePaths.Length > 1))
                {
                    Array.Sort(_ImagePaths);
                }
            }
        }

        public int PrimaryImageIndex
        {
            get
            {
                return _PrimaryImageIndex;
            }
            set
            {
                _PrimaryImageIndex = value;
            }
        }

        public int ThumbnailSourceImageIndex
        {
            get
            {
                return _ThumbnailSourceImageIndex;
            }
            set
            {
                _ThumbnailSourceImageIndex = value;
            }
        }

        public string ImageName
        {
            get
            {
                return _ImageName;
            }
            set
            {
                _ImageName = value;
            }
        }

        public int ImageMargin
        {
            get
            {
                return _ImageMargin;
            }
            set
            {
                _ImageMargin = value;
            }
        }

        public bool ParentDisplayShouldSlide
        {
            get
            {
                return _ParentDisplayShouldSlide;
            }
        }

        #endregion

        public ClickableThumbnailPictureBox()
            : base()
        {
            _PopupIsOpen = false;
            _ParentDisplayShouldSlide = true;
            //_ToolTip = new ClickableThumbnailToolTip();
            _MenuToIndexHashtable = new Hashtable();
            //_ToolTipList = new ArrayList();

            //this.BackColor = Color.Black;
            //this.BackColor = Color.White;
            this.BackColor = Color.FromArgb(220, 220, 220);
            this.SizeMode = PictureBoxSizeMode.CenterImage;

            this.DoubleClick += ClickableThumbnailPictureBox_Click;
            this.MouseEnter += ClickableThumbnailPictureBox_MouseEnter;
            this.MouseLeave += ClickableThumbnailPictureBox_MouseLeave;

            ImageMargin = 2;
            _LastImageOpenEventTime = DateTime.Now;
        }

        //~ClickableThumbnailPictureBox()
        //{
        //    this.Dispose();
        //}

        protected override void Dispose(bool disposing)
        {
            DisposeFunction(disposing);
        }


        // note: do not try to make a delegate for Dispose(), because it will make the 
        // .NET Framework hang onto the object indefinitely, creating a User object leak.
        // I had the delegate code here originally because during testing, I got an 
        // "invalid cross-thread operation" exception message when it hit the base.Dispose() 
        // call, but apparently it's still not the right thing to do.

        //delegate void DisposeCallback(bool disposing);

        protected void DisposeFunction(bool disposing)
        {
            //if (this.InvokeRequired)
            //{
            //    DisposeCallback c = new DisposeCallback(DisposeFunction);
            //    this.Invoke(c, new object[] { disposing });
            //}
            //else
            //{
                if (_ToolTip != null)
                {
                    try
                    {
                        _ToolTip.RemoveAll();
                        _ToolTip.Dispose();
                        _ToolTip = null;
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error disposing of object: " + ex.Message);
                        // do nothing
                    }
                }

                try
                {
                    _MenuToIndexHashtable.Clear();
                    _MenuToIndexHashtable = null;
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error disposing of object: " + ex.Message);
                    // do nothing
                }

                if (this.Image != null)
                {
                    try
                    {
                        this.Image.Dispose();
                        this.Image = null;
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error disposing of object: " + ex.Message);
                        // do nothing
                    }
                }

                //if (_ThumbnailImage != null)
                //{
                //    try
                //    {
                //        _ThumbnailImage.Dispose();
                //        _ThumbnailImage = null;
                //    }
                //    catch (Exception ex)
                //    {
                //        Console.WriteLine("Error disposing of object: " + ex.Message);
                //        // do nothing
                //    }
                //}

                if (_ContextMenu != null)
                {
                    try
                    {
                        _ContextMenu.Dispose();
                        _ContextMenu = null;
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error disposing of object: " + ex.Message);
                        // do nothing
                    }
                }

                this.DoubleClick -= ClickableThumbnailPictureBox_Click;
                this.MouseEnter -= ClickableThumbnailPictureBox_MouseEnter;
                this.MouseLeave -= ClickableThumbnailPictureBox_MouseLeave;

                _ImagePaths = null;
                _ImageName = "";

                try
                {
                    this.Events.Dispose();
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error disposing of delegate list: " + ex.Message);
                    // do nothing
                }

                try
                {
                    base.Dispose(disposing);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error calling base.Dispose: " + ex.Message);
                    // do nothing
                }
            //}
        }

        protected void ClickableThumbnailPictureBox_MouseLeave(object sender, EventArgs e)
        {
            if (!_PopupIsOpen)
            {
                SetParentDisplaySlideState(true);
            }
        }

        protected void ClickableThumbnailPictureBox_MouseEnter(object sender, EventArgs e)
        {
            SetParentDisplaySlideState(false);
        }

        protected void SetParentDisplaySlideState(bool newState)
        {
            //if (ParentDisplay != null)
            //{
            //    ParentDisplay.SlideEnabled = newState;
            //}
            _ParentDisplayShouldSlide = newState;
        }

        public ClickableThumbnailPictureBox(string[] imagePaths, int primaryImageIndex, int thumbnailSourceImageIndex)
            : this()
        {
            ImagePaths = imagePaths;
            PrimaryImageIndex = primaryImageIndex;
            ThumbnailSourceImageIndex = thumbnailSourceImageIndex;
            ReloadConfiguration();
        }

        public void ReloadConfiguration()
        {
            bool proceed = true;
            string errorMessage = "";

            if ((ImagePaths == null) || (ImagePaths.Length == 0))
            {
                errorMessage = "The list of image paths cannot be null or empty.";
                //throw new Exception(errorMessage);
                proceed = false;
            }

            if ((PrimaryImageIndex < 0) || (PrimaryImageIndex < 0))
            {
                errorMessage = "The primary image index must be specified, and must be within the bounds of the image path list array.";
                //throw new Exception(errorMessage);
                proceed = false;
            }

            if ((ThumbnailSourceImageIndex < 0) || (ThumbnailSourceImageIndex < 0))
            {
                errorMessage = "The thumbnail source image index must be specified, and must be within the bounds of the image path list array.";
                //throw new Exception(errorMessage);
                proceed = false;
            }

            if (proceed)
            {
                SetToolTip(Path.GetFileNameWithoutExtension(ImagePaths[PrimaryImageIndex]));
                BuildContextMenu();
                CreateThumbnailImage();
            }
            else
            {
                SetErrorImage(errorMessage);
            }
        }

        protected void BuildContextMenu()
        {
            if (_ContextMenu != null)
            {
                _ContextMenu.Dispose();
            }
            _ContextMenu = new ClickableThumbnailContextMenu();

            //cmImage.Tag = DateTime.Now.ToBinary().ToString() + " - " + rand.Next(26354).ToString();

            for (int i = 0; i < ImagePaths.Length; i++)
            {
                string imageName = Path.GetFileName(ImagePaths[i]);
                ClickableThumbnailMenuItem var = new ClickableThumbnailMenuItem(imageName);
                var.Tag = i.ToString() + " Parent";

                ClickableThumbnailMenuItem varOpen = new ClickableThumbnailMenuItem("Open With Default Program");
                varOpen.Tag = i.ToString() + " Open";
                varOpen.SetClickEventHandler(ContextMenu_OpenFile);
                _MenuToIndexHashtable.Add(varOpen.GetHashCode(), i);
                var.MenuItems.Add(varOpen);

                ClickableThumbnailMenuItem varCopy = new ClickableThumbnailMenuItem("Copy Path To Clipboard");
                varCopy.Tag = i.ToString() + " Copy";
                varCopy.SetClickEventHandler(ContextMenu_CopyFilePath);
                _MenuToIndexHashtable.Add(varCopy.GetHashCode(), i);
                var.MenuItems.Add(varCopy);

                _ContextMenu.MenuItems.Add(var);
            }

            _ContextMenu.SetPopupEventHandler(cmImage_Popup);
            _ContextMenu.SetCollapseEventHandler(cmImage_Collapse);

            this.ContextMenu = _ContextMenu;
        }

        public bool ThumbnailCallback()
        {
            return true;
        }

        void cmImage_Popup(object sender, EventArgs e)
        {
            SetParentDisplaySlideState(false);
            _PopupIsOpen = true;
        }

        void cmImage_Collapse(object sender, EventArgs e)
        {
            SetParentDisplaySlideState(true);
            _PopupIsOpen = false;
        }

        protected void CreateThumbnailImage()
        {
            if (!File.Exists(ImagePaths[ThumbnailSourceImageIndex]))
            {
                SetErrorImage("Unable to locate the file '" + ImagePaths[ThumbnailSourceImageIndex] + "'.");
            }
            else
            {
                try
                {
                    Image tempImage = Image.FromFile(ImagePaths[ThumbnailSourceImageIndex]);

                    float scalingFactor = (float)(this.Height - (2.0f * _ImageMargin)) / (float)tempImage.Height;

                    if (tempImage.Width > tempImage.Height)
                    {
                        // landscape orientation
                        scalingFactor = (float)(this.Width - (2.0f * _ImageMargin)) / (float)tempImage.Width;
                    }
                    else
                    {
                        // portrait orientation
                        // no change necessary, scalingFactor is already set
                    }

                    int newWidth = (int)((float)tempImage.Width * scalingFactor);
                    int newHeight = (int)((float)tempImage.Height * scalingFactor);

                    this.Image = tempImage.GetThumbnailImage(newWidth, newHeight, new Image.GetThumbnailImageAbort(ThumbnailCallback), IntPtr.Zero);

                    tempImage.Dispose();
                    tempImage = null;
                }
                catch (Exception ex)
                {
                    SetErrorImage("Error: " + ex.Message);
                }
            }
        }

        public void SetPrimaryImage(string primaryImagePath)
        {
            int piIndex = IndexOf(primaryImagePath);
            if (piIndex > -1)
            {
                PrimaryImageIndex = piIndex;
            }
        }

        public void SetThumbnailSourceImage(string thumbnailSourceImagePath)
        {
            int piIndex = IndexOf(thumbnailSourceImagePath);
            if (piIndex > -1)
            {
                ThumbnailSourceImageIndex = piIndex;
            }
        }

        public int IndexOf(string imagePath)
        {
            for (int i = 0; i < ImagePaths.Length; i++)
            {
                if (ImagePaths[i] == imagePath)
                {
                    return i;
                }
            }
            return -1;
        }

        protected void SetToolTip(string message)
        {
            if (_ToolTip != null)
            {
                _ToolTip.Dispose();
            }
            // note: see the comments where this const is defined before enabling creation of tooltips
            if (_EnableToolTips)
            {
                _ToolTip = new ClickableThumbnailToolTip();
                _ToolTip.SetToolTip(this, message);
            }
        }

        protected void SetErrorImage(string errorMessage)
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();
            Stream iStream = currentAssembly.GetManifestResourceStream("BenLincoln.DaVinci.GUI.ImageError.PNG");
            this.Image = Image.FromStream(iStream);
            iStream.Close();
            this.SizeMode = PictureBoxSizeMode.StretchImage;
            SetToolTip(errorMessage);
        }

        protected void DisplayErrorPopup(string errorMessage)
        {
            MessageBox.Show(errorMessage, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }


        #region Menu Event Handlers

        //protected void ContextMenu_OpenPrimaryFile(object sender, System.EventArgs e)
        //{
        //    OpenFile(PrimaryImageIndex);
        //}

        //protected void ContextMenu_CopyPrimaryFilePath(object sender, System.EventArgs e)
        //{
        //    CopyFullPathToClipBoard(PrimaryImageIndex);
        //}

        protected void ContextMenu_OpenFile(object sender, System.EventArgs e)
        {
            int senderID = sender.GetHashCode();
            //MessageBox.Show(senderID.ToString());
            if (_MenuToIndexHashtable.Contains(senderID))
            {
                int targetID = (int)_MenuToIndexHashtable[senderID];
                OpenFile(targetID);
            }
            else
            {
                //throw new Exception("Unable to locate the hash code for the sending object in the mapping table.");
                //SetErrorImage("Unable to locate the hash code for the sending object in the mapping table.");
                DisplayErrorPopup("Unable to locate the hash code for the sending object in the mapping table.");
            }
        }

        protected void ContextMenu_CopyFilePath(object sender, System.EventArgs e)
        {
            int senderID = sender.GetHashCode();
            if (_MenuToIndexHashtable.Contains(senderID))
            {
                int targetID = (int)_MenuToIndexHashtable[senderID];
                CopyFullPathToClipBoard(targetID);
            }
            else
            {
                //throw new Exception("Unable to locate the hash code for the sending object in the mapping table.");
                //SetErrorImage("Unable to locate the hash code for the sending object in the mapping table.");
                DisplayErrorPopup("Unable to locate the hash code for the sending object in the mapping table.");
            }
        }

        #endregion


        #region General Event Handlers

        private void ClickableThumbnailPictureBox_Click(object sender, EventArgs e)
        {
            OpenFile(PrimaryImageIndex);
        }

        #endregion


        protected void OpenFile(int fileIndex)
        {
            TimeSpan interval = DateTime.Now - _LastImageOpenEventTime;
            //if (interval.TotalMilliseconds > _MinimumTimeBetweenImageOpeningEvents)
            //{
                _LastImageOpenEventTime = DateTime.Now;
                if (File.Exists(ImagePaths[fileIndex]))
                {
                    try
                    {
                        System.Diagnostics.Process.Start(ImagePaths[fileIndex]);
                    }
                    catch (Exception ex)
                    {
                        //throw new Exception("Unable to open the file '" + ImagePaths[fileIndex] +
                        //    "'. The specific error message was '" + ex.Message + "'.", ex);
                        //SetErrorImage("Unable to open the file '" + ImagePaths[fileIndex] +
                        //    "'. The specific error message was '" + ex.Message + "'.");
                        DisplayErrorPopup("Unable to open the file '" + ImagePaths[fileIndex] +
                            "'. The specific error message was '" + ex.Message + "'.");
                    }
                }
                else
                {
                    //throw new Exception("Unable to locate the file '" + ImagePaths[fileIndex] + "'.");
                    //SetErrorImage("Unable to locate the file '" + ImagePaths[fileIndex] + "'.");
                    DisplayErrorPopup("Unable to locate the file '" + ImagePaths[fileIndex] + "'.");
                }
            //}
        }

        protected void CopyFullPathToClipBoard(int fileIndex)
        {
            Clipboard.SetText(ImagePaths[fileIndex]);
        }

        private void InitializeComponent()
        {
            ((System.ComponentModel.ISupportInitialize)(this)).BeginInit();
            this.SuspendLayout();
            ((System.ComponentModel.ISupportInitialize)(this)).EndInit();
            this.ResumeLayout(false);

        }




    }
}
