// BenLincoln.DaVinci.GUI - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.GUI.

// BenLincoln.DaVinci.GUI is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci.GUI is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci.GUI (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci.GUI
{
    public class ProcessableItemCollection : CollectionBase, IBindingList
    {
        protected PropertyDescriptor _SortProperty;
        protected ListSortDirection _SortDirection;
        protected ListChangedEventArgs resetEvent = new ListChangedEventArgs(ListChangedType.Reset, -1);
        protected ListChangedEventHandler _OnListChanged;


        public ProcessableItemCollection()
        {
        }

        public ProcessableItemCollection(ProcessableItemCollection sourceCollection)
            :this()
        {
            for (int i = 0; i < sourceCollection.Count; i++)
            {
                this.Add(new ProcessableItem(sourceCollection[i]));
            }
        }

        public ProcessableItem this[int index]
        {
            get
            {
                return (ProcessableItem)this.List[index];
            }
            set
            {
                ProcessableItem pi = value;
                if (Contains(pi))
                {
                    // might need to do something here?
                }
                else
                {
                    pi.Position = GetNextUnusedPosition();
                }
                pi.ParentCollection = this;
                
                this.List[index] = pi;
            }
        }

        public void Swap(int position1, int position2)
        {
            int index1 = -1;
            int index2 = -1;

            for (int i = 0; i < this.Count; i++)
            {
                if (this[i].Position == position1)
                {
                    index1 = i;
                }
                if (this[i].Position == position2)
                {
                    index2 = i;
                }
            }

            if ((index1 != index2) && (index1 > -1) && (index2 > -1))
            {
                int tempPosition = this[index1].Position;
                this[index1].Position = this[index2].Position;
                this[index2].Position = tempPosition;
            }
        }

        protected int GetNextUnusedPosition()
        {
            UpdatePositions();
            int result = 1;
            for (int i = 0; i < this.Count; i++)
            {
                if (this[i].Position >= result)
                {
                    result = this[i].Position + 1;
                }
            }
            return result;
        }

        protected void UpdatePositions()
        {
            if (this.Count > 0)
            {
                int[] allPositions = new int[this.Count];
                for (int i = 0; i < this.Count; i++)
                {
                    allPositions[i] = this[i].Position;
                }
                Array.Sort(allPositions);
                for (int j = 0; j < this.Count; j++)
                {
                    for (int k = 0; k < this.Count; k++)
                    {
                        if (this[k].Position == allPositions[j])
                        {
                            this[k].Position = j + 1;
                        }
                    }
                }
            }
        }

        public ProcessableItem this[string itemName]
        {
            get
            {
                int index = IndexOf(itemName);
                if (index == -1)
                {
                    throw new KeyNotFoundException("No item with the name '" + itemName + "' was found in the collection.");
                }
                return this[index];
            }
            set
            {
                int index = IndexOf(itemName);
                if (index == -1)
                {
                    throw new KeyNotFoundException("No item with the name '" + itemName + "' was found in the collection.");
                }
                //ProcessableItem pi = value;
                //pi.ParentCollection = this;
                //this.List[index] = pi;
                // use the accessor in this class so the value-setting code doesn't need to be duplicated
                this[index] = value;
            }
        }

        public int IndexOf(ProcessableItem item)
        {
            return base.List.IndexOf(item);
        }

        public int IndexOf(string itemName)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (this[i].Name == itemName)
                {
                    return i;
                }
            }
            return -1;
        }

        public int IndexOf(int position)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (this[i].Position == position)
                {
                    return i;
                }
            }
            return -1;
        }

        public int Add(ProcessableItem item)
        {
            return this.Add(item, GetNextUnusedPosition());
        }

        public int Add(ProcessableItem item, int position)
        {
            ProcessableItem pi = item;
            pi.ParentCollection = this;
            pi.Position = position;
            return this.List.Add(pi);
        }

        public void Remove(ProcessableItem item)
        {
            this.List.Remove(item);
            UpdatePositions();
        }

        public void CopyTo(Array array, int index)
        {
            this.List.CopyTo(array, index);
        }

        public void AddRange(ProcessableItemCollection coll)
        {
            for (int i = 0; i < coll.Count; i++)
            {
                this.List.Add(coll[i]);
            }
        }

        public void AddRange(ProcessableItem[] coll)
        {
            foreach (ProcessableItem item in coll)
            {
                this.List.Add(item);
            }
        }

        public bool Contains(ProcessableItem item)
        {
            return this.List.Contains(item);
        }

        public bool Contains(string itemName)
        {
            if (IndexOf(itemName) > -1)
            {
                return true;
            }
            return false;
        }

        public void Insert(int index, ProcessableItem item)
        {
            ProcessableItem pi = item;
            pi.ParentCollection = this;
            pi.Position = GetNextUnusedPosition();
            this.List.Insert(index, pi);
        }

        public void Sort()
        {
            InnerList.Sort();
        }

        //public ProcessableItemEnumerator GetEnumerator()
        //{
        //    return new ProcessableItemEnumerator(this);
        //}


        #region IBindingList interface

        bool IBindingList.AllowEdit
        {
            get { return false; }
        }

        bool IBindingList.AllowNew
        {
            get { return false; }
        }

        bool IBindingList.AllowRemove
        {
            get { return true; }
        }

        bool IBindingList.SupportsChangeNotification
        {
            get { return true; }
        }

        bool IBindingList.SupportsSearching
        {
            get { return false; }
        }

        bool IBindingList.SupportsSorting
        {
            get { return true; }
        }

        object IBindingList.AddNew()
        {
            throw new NotSupportedException();
        }

        void IBindingList.AddIndex(PropertyDescriptor property) 

        {
            throw new NotSupportedException();
        }

        void IBindingList.RemoveIndex(PropertyDescriptor property)
        {
            throw new NotSupportedException();
        }

        int IBindingList.Find(PropertyDescriptor property, object key)
        {
            throw new NotSupportedException();
        }


        bool IBindingList.IsSorted
        {
            get
            {
                return (_SortProperty != null);
            }
        }

        ListSortDirection IBindingList.SortDirection
        {
            get
            {
                return _SortDirection;
            }
        }

        PropertyDescriptor IBindingList.SortProperty
        {
            get
            {
                return _SortProperty;
            }
        }

        void IBindingList.ApplySort(PropertyDescriptor property, ListSortDirection direction)
        {
            _SortProperty = property;
            _SortDirection = direction;
            this.InnerList.Sort(new GenericObjectComparer(_SortProperty, _SortDirection));
        }

        void IBindingList.RemoveSort()
        {
            _SortProperty = null;
        }

        public event ListChangedEventHandler ListChanged
        {
            add
            {
                _OnListChanged += value;
            }
            remove
            {
                _OnListChanged -= value;
            }
        }

        protected override void OnClearComplete()
        {
            OnListChanged(resetEvent);
        }

        protected override void OnInsertComplete(int index, object value)
        {
            UpdatePositions();
            //ProcessableItem pi = (ProcessableItem)value;
            //pi.Parent = this;
            OnListChanged(new ListChangedEventArgs(ListChangedType.ItemAdded, index));
        }

        protected override void OnRemoveComplete(int index, object value)
        {
            UpdatePositions();
            //ProcessableItem pi = (ProcessableItem)value;
            //pi.Parent = this;
            OnListChanged(new ListChangedEventArgs(ListChangedType.ItemDeleted, index));
        }

        protected override void OnSetComplete(int index, object oldValue, object newValue)
        {
            UpdatePositions();
            //if (oldValue != newValue)
            //{
                OnListChanged(new ListChangedEventArgs(ListChangedType.ItemAdded, index));
            //}
        }

        protected virtual void OnListChanged(ListChangedEventArgs ev)
        {
            if (_OnListChanged != null)
            {
                _OnListChanged(this, ev);
            }
        }

        public string[] GetDirectoryList()
        {
            string[] result = new string[this.Count];

            for (int i = 0; i < this.Count; i++)
            {
                result[i] = this[IndexOf(i + 1)].Name;
            }

            return result;
        }

        #endregion

    }
}
