// BenLincoln.DaVinci.GUI - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.GUI.

// BenLincoln.DaVinci.GUI is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci.GUI is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci.GUI (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace BenLincoln.DaVinci.GUI
{
    public partial class SlidingThumbnailDisplay : UserControl, IDisposable
    {
        protected enum AnimationState
        {
            Normal,
            SlidingDelay
        }

        // number of pixels between thumbnails
        protected const int _ThumbnailSeparation = 1;
        
        // default number of pixels that thumbnails move between updates
        protected const int _DefaultThumbnailSlideDistance = 1;

        // the maximum number of pixels that thumbnails are allowed to slide in a single update
        // (only used when IncreaseSlideSpeedWithQueueLength is set to true)
        protected int _MaxThumbnailSlideDistance;

        // default maximum slide distance
        protected const int _DefaultMaxThumbnailSlideDistance = 50;

        //protected const int _ThumbnailSlideDistance = 1;
        //protected const int _ThumbnailSlideDistance = 50;
        protected int _ThumbnailSlideDistance;

        protected bool _IncreaseSlideSpeedWithQueueLength;

        // milliseconds between updates
        //protected const int _UpdateInterval = 30;
        protected const int _UpdateInterval = 10;

        // delay (in milliseconds) when a thumbnail finishes sliding into the display
        //protected const int _FinishedSlidingDelay = 750;
        // this is currently set to zero because it made the animation look jerky instead of 
        // like an assembly line, which is why I added it in the first place
        protected const int _FinishedSlidingDelay = 0;

        // retain enough thumbnails to draw this wide of a control if necessary
        protected int _ThumbnailConcourseWidth;


        protected System.Windows.Forms.Timer _AnimationTimer;

        protected ArrayList _ThumbnailControlHashes;

        protected ArrayList _QueuedThumbnails;

        protected AnimationState _CurrentAnimationState;

        protected int _FinishedSlidingDelayIteration;

        protected bool _SlideEnabled;

        protected bool _SlidingDelayComplete;

        protected DateTime _LastAnimationTimestamp;

        protected bool _Disposed;

        //protected static readonly object _AnimationLockObject = new object();
        protected readonly object _AnimationLockObject = new object();

        #region Properties

        public bool SlideEnabled
        {
            get
            {
                return _SlideEnabled;
            }
            set
            {
                _SlideEnabled = value;
            }
        }

        public bool IncreaseSlideSpeedWithQueueLength
        {
            get
            {
                return _IncreaseSlideSpeedWithQueueLength;
            }
            set
            {
                _IncreaseSlideSpeedWithQueueLength = value;
            }
        }

        #endregion

        public SlidingThumbnailDisplay()
        {
            _Disposed = false;
            _ThumbnailSlideDistance = _DefaultThumbnailSlideDistance;
            SetMaxThumbnailSlideDistance();
            IncreaseSlideSpeedWithQueueLength = true;
            UpdateWidth();
            this.Resize += SlidingThumbnailDisplay_Resize;
            InitializeComponent();
            MouseEnter += SlidingThumbnailDisplay_MouseEnter;
            MouseLeave += SlidingThumbnailDisplay_MouseLeave;

            //this.BackColor = Color.Black;
            this.BackColor = Color.FromArgb(30, 30, 30);

            _CurrentAnimationState = AnimationState.Normal;

            _ThumbnailControlHashes = new ArrayList();
            _QueuedThumbnails = new ArrayList();
            _FinishedSlidingDelayIteration = 0;

            _LastAnimationTimestamp = DateTime.Now;

            _AnimationTimer = new System.Windows.Forms.Timer();

            _AnimationTimer.Interval = _UpdateInterval;

            _AnimationTimer.Tick += _AnimationTimer_Tick;

            _SlidingDelayComplete = false;

            // wait a random number of milliseconds so that multiple displays are 
            // not literally trying to update simultaneously
            Random rand = new Random();
            Thread.Sleep(rand.Next(100));

            SlideEnabled = true;
            _AnimationTimer.Enabled = true;
        }

        //~SlidingThumbnailDisplay()
        //{
        //    this.Dispose();
        //}

        protected override void Dispose(bool disposing)
        {
            // If you need thread safety, use a lock around these 
            // operations, as well as in your methods that use the resource.
            if (!_Disposed)
            {
                //if (disposing)
                //{
                    if (_AnimationTimer != null)
                    {
                        _AnimationTimer.Tick -= _AnimationTimer_Tick;
                        _AnimationTimer.Dispose();
                        _AnimationTimer = null;
                    }
                //}

                this.Resize -= SlidingThumbnailDisplay_Resize;
                MouseEnter -= SlidingThumbnailDisplay_MouseEnter;
                MouseLeave -= SlidingThumbnailDisplay_MouseLeave;

                _ThumbnailControlHashes = null;
                if (_QueuedThumbnails != null)
                {
                    if (_QueuedThumbnails.Count > 0)
                    {
                        for (int i = 0; i < _QueuedThumbnails.Count; i++)
                        {
                            ClickableThumbnailPictureBox qt = (ClickableThumbnailPictureBox)_QueuedThumbnails[i];
                            qt.Dispose();
                        }
                    }
                    _QueuedThumbnails = null;
                }

                for (int i = 0; i < this.Controls.Count; i++)
                {
                    if (this.Controls[i] != null)
                    {
                        try
                        {
                            Control currentControl = this.Controls[i];
                            currentControl.Dispose();
                            this.Controls.Remove(currentControl);
                            currentControl = null;
                        }
                        catch (Exception ex)
                        {
                            // do nothing
                        }
                    }
                }

                try
                {
                    this.Events.Dispose();
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error disposing of delegate list: " + ex.Message);
                    // do nothing
                }

                _Disposed = true;
            }
            base.Dispose(true);
        }

        void SlidingThumbnailDisplay_Resize(object sender, EventArgs e)
        {
            UpdateWidth();
            SetMaxThumbnailSlideDistance();
        }

        protected void UpdateWidth()
        {
            int newWidth = this.Width + this.Height;
            if ((_ThumbnailConcourseWidth < newWidth))
            {
                _ThumbnailConcourseWidth = newWidth;
            }
        }

        protected void SetMaxThumbnailSlideDistance()
        {
            if (_IncreaseSlideSpeedWithQueueLength)
            {
                _MaxThumbnailSlideDistance = GetThumbnailDimension();
                if (_MaxThumbnailSlideDistance < 0)
                {
                    _MaxThumbnailSlideDistance = 1;
                }
            }
            else
            {
                _MaxThumbnailSlideDistance = _DefaultThumbnailSlideDistance;
            }
        }

        //public void QueueThumbnail(ClickableThumbnailPictureBox newThumbnail)
        //{
        //    _QueuedThumbnails.Add(newThumbnail);
        //}

        public void QueueThumbnail(string[] imagePaths, int primaryImageIndex, int thumbnailSourceImageIndex)
        {
            if (this.Enabled)
            {
                ClickableThumbnailPictureBox newThumbnail = GetNewThumbnail(imagePaths);
                newThumbnail.PrimaryImageIndex = primaryImageIndex;
                newThumbnail.ThumbnailSourceImageIndex = thumbnailSourceImageIndex;
                newThumbnail.ReloadConfiguration();
                try
                {
                    _QueuedThumbnails.Add(newThumbnail);
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
        }

        public void QueueThumbnail(string[] imagePaths, string primaryImagePath, string thumbnailSourceImagePath)
        {
            if (this.Enabled)
            {
                ClickableThumbnailPictureBox newThumbnail = GetNewThumbnail(imagePaths);
                newThumbnail.SetPrimaryImage(primaryImagePath);
                newThumbnail.SetThumbnailSourceImage(thumbnailSourceImagePath);
                newThumbnail.ReloadConfiguration();
                try
                {
                    _QueuedThumbnails.Add(newThumbnail);
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
        }

        protected ClickableThumbnailPictureBox GetNewThumbnail(string[] imagePaths)
        {
            ClickableThumbnailPictureBox newThumbnail = new ClickableThumbnailPictureBox();
            //newThumbnail.ParentDisplay = this;
            newThumbnail.ImagePaths = imagePaths;
            newThumbnail.Height = GetThumbnailDimension();
            //newThumbnail.Height = this.Height;
            newThumbnail.Width = newThumbnail.Height;
            newThumbnail.Anchor = AnchorStyles.Right;
            return newThumbnail;
        }

        protected int GetThumbnailDimension()
        {
            return this.Height - (4 * _ThumbnailSeparation);
        }

        public void AddNextThumbnailToControl(int maxXPosition)
        {
            if (this.Enabled)
            {
                if (_QueuedThumbnails.Count > 0)
                {
                    try
                    {
                        ClickableThumbnailPictureBox newThumbnail = (ClickableThumbnailPictureBox)_QueuedThumbnails[0];
                        //newThumbnail.Left = this.Width;
                        newThumbnail.Left = maxXPosition + _ThumbnailSeparation;
                        newThumbnail.Top = (2 * _ThumbnailSeparation);
                        this.Controls.Add(newThumbnail);
                        _ThumbnailControlHashes.Add(newThumbnail.GetHashCode());
                        _QueuedThumbnails.RemoveAt(0);
                    }
                    catch (Exception ex)
                    {
                        throw new Exception("Unable to add a new thumbnail image to the display.", ex);
                    }
                }
            }
        }

        protected void Animate()
        {
            if (this.Enabled)
            {
                lock (_AnimationLockObject)
                {
                    TimeSpan uDiff = (DateTime.Now - _LastAnimationTimestamp);
                    if (uDiff.TotalMilliseconds < _UpdateInterval)
                    {
                        // do nothing - something triggered multiple updates before one was finished
                    }
                    else
                    {
                        if (_IncreaseSlideSpeedWithQueueLength)
                        {
                            SetMaxThumbnailSlideDistance();
                            _ThumbnailSlideDistance = _DefaultThumbnailSlideDistance + _QueuedThumbnails.Count;
                            if (_ThumbnailSlideDistance > _MaxThumbnailSlideDistance)
                            {
                                _ThumbnailSlideDistance = _MaxThumbnailSlideDistance;
                            }
                        }
                        else
                        {
                            _ThumbnailSlideDistance = _DefaultThumbnailSlideDistance;
                        }

                        bool slide = true;
                        if (!SlideEnabled)
                        {
                            slide = false;
                        }
                        if (slide)
                        {
                            for (int i = 0; i < this.Controls.Count; i++)
                            {
                                if ((this.Controls[i] != null) && (this.Controls[i] is ClickableThumbnailPictureBox))
                                {
                                    try
                                    {
                                        ClickableThumbnailPictureBox cpb = (ClickableThumbnailPictureBox)this.Controls[i];
                                        if (!cpb.ParentDisplayShouldSlide)
                                        {
                                            slide = false;
                                        }
                                    }
                                    catch (Exception ex)
                                    {
                                        // do nothing
                                    }
                                }
                            }
                        }
                        if (slide)
                        {
                            switch (_CurrentAnimationState)
                            {
                                case AnimationState.SlidingDelay:
                                    _FinishedSlidingDelayIteration++;
                                    if (_FinishedSlidingDelayIteration > (_FinishedSlidingDelay / _UpdateInterval))
                                    {
                                        _CurrentAnimationState = AnimationState.Normal;
                                        _SlidingDelayComplete = true;
                                        _FinishedSlidingDelayIteration = 0;
                                    }
                                    break;

                                case AnimationState.Normal:
                                    int maximumXPosition = 0;
                                    int minimumXPositionThreshold = this.Width - (_ThumbnailConcourseWidth + this.Height);

                                    ArrayList toBeRemoved = new ArrayList();

                                    bool foundThumbnail = false;

                                    foreach (int hashCode in _ThumbnailControlHashes)
                                    {
                                        for (int i = 0; i < this.Controls.Count; i++)
                                        {
                                            int currentHashCode = this.Controls[i].GetHashCode();
                                            if (currentHashCode == hashCode)
                                            {
                                                if (this.Controls[i].Location.X > maximumXPosition)
                                                {
                                                    maximumXPosition = this.Controls[i].Location.X;
                                                    foundThumbnail = true;
                                                }

                                                if (this.Controls[i].Location.X < minimumXPositionThreshold)
                                                {
                                                    toBeRemoved.Add(currentHashCode);
                                                }
                                            }
                                        }
                                    }

                                    // remove any thumbnail that's slid left beyond the threshold
                                    foreach (int removeHash in toBeRemoved)
                                    {
                                        for (int i = 0; i < this.Controls.Count; i++)
                                        {
                                            int currentHashCode = this.Controls[i].GetHashCode();
                                            if (currentHashCode == removeHash)
                                            {
                                                this.Controls[i].Dispose();
                                                //this.Controls.RemoveAt(i);
                                                this.Controls.Remove(this.Controls[i]);
                                                //break;
                                            }
                                        }
                                        _ThumbnailControlHashes.Remove(removeHash);
                                    }
                                    toBeRemoved.Clear();

                                    // it's the right edge of the thumbnails that actually matters
                                    maximumXPosition += this.Height;

                                    // is the current state such that the thumbnails should slide?
                                    bool doSlide = true;

                                    if (maximumXPosition <= (this.Width - _ThumbnailSeparation))
                                    {
                                        if (_QueuedThumbnails.Count == 0)
                                        {
                                            doSlide = false;
                                            _SlidingDelayComplete = true;
                                        }
                                        else
                                        {
                                            if (_SlidingDelayComplete)
                                            {
                                                if (foundThumbnail)
                                                {
                                                    AddNextThumbnailToControl(maximumXPosition);
                                                }
                                                else
                                                {
                                                    AddNextThumbnailToControl(this.Width);
                                                }
                                                _SlidingDelayComplete = false;
                                            }
                                            else
                                            {
                                                doSlide = false;
                                                _CurrentAnimationState = AnimationState.SlidingDelay;
                                            }
                                        }
                                    }

                                    if (doSlide)
                                    {
                                        ArrayList currentHashes = new ArrayList(_ThumbnailControlHashes);
                                        foreach (int hashCode in currentHashes)
                                        {
                                            bool foundHash = false;
                                            for (int i = 0; i < this.Controls.Count; i++)
                                            {
                                                int currentHashCode = this.Controls[i].GetHashCode();
                                                if (currentHashCode == hashCode)
                                                {
                                                    this.Controls[i].Left -= _ThumbnailSlideDistance;
                                                    foundHash = true;
                                                }
                                            }
                                            // prune orphaned hashes
                                            if (!foundHash)
                                            {
                                                _ThumbnailControlHashes.Remove(hashCode);
                                            }
                                        }
                                        currentHashes.Clear();
                                        currentHashes = null;
                                    }

                                    break;
                            }
                            _LastAnimationTimestamp = DateTime.Now;
                        }
                    }
                }
            }
        }

        #region Event Handlers

        private void SlidingThumbnailDisplay_MouseEnter(object sender, EventArgs e)
        {
            SlideEnabled = false;
        }

        private void SlidingThumbnailDisplay_MouseLeave(object sender, EventArgs e)
        {
           SlideEnabled = true;
        }

        protected void _AnimationTimer_Tick(object sender, EventArgs e)
        {
            Animate();
        }

        #endregion
    }
}
