// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class ColourGradientImage : MultiChannelImage
    {
        protected GreyscaleChannel _BaseChannel;
        protected ColourGradient _ColourGradient;

        #region Properties

        public GreyscaleChannel BaseChannel
        {
            get
            {
                return _BaseChannel;
            }
            set
            {
                _BaseChannel = value;
                _Channels = new GreyscaleChannel[] { _BaseChannel };
                GenerateName();
            }
        }

        public ColourGradient ColourGradient
        {
            get
            {
                return _ColourGradient;
            }
            set
            {
                _ColourGradient = value;
                GenerateName();
            }
        }

        #endregion

        public ColourGradientImage()
            : base()
        {
            _NamePrefix = "FC-Gr-";
            _TemporaryObjectNamePrefix = "gr_temp_";
        }

        public ColourGradientImage(GreyscaleChannel baseChannel, ColourGradient gradient)
            : this()
        {
            BaseChannel = baseChannel;
            ColourGradient = gradient;
            IsCustomImage = true;
        }

        protected override void _GenerateName()
        {
            StringBuilder nameBuilder = new StringBuilder(_NamePrefix);

            string bcName = "Undefined";
            if ((BaseChannel != null) && (BaseChannel.ImageName != null) && (BaseChannel.ImageName != ""))
            {
                bcName = BaseChannel.ImageName;
            }
            string grName = "Undefined";
            if ((ColourGradient != null) && (ColourGradient.GradientName != null) && (ColourGradient.GradientName != ""))
            {
                grName = ColourGradient.GradientName;
            }

            nameBuilder.Append("(");
            nameBuilder.Append(bcName);
            nameBuilder.Append(")");
            nameBuilder.Append(_NameComponentSeparator);
            //nameBuilder.Append("(");
            nameBuilder.Append(grName);
            //nameBuilder.Append(")");
            nameBuilder.Append(_NameSuffix);
            _ImageName = nameBuilder.ToString();
        }

        public override string GetCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            string filePath = GetFilePathWithoutExtension();
            StringBuilder builder = new StringBuilder();
            // if the option to recreate existing images is not enabled and the 
            // file already exists, do nothing.
            if ((!regenerateExistingFiles) && AllOutputFilesExist(filePath, OutputFormats))
            {
                builder.Append("# Information: all file-type variations of the gradient-mapped false colour image '");
                builder.Append(ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }
            _WillBeCreatedByScript = true;

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' colour gradient image");
            builder.Append(Environment.NewLine);
            builder.Append("# Apply the gradient '");
            builder.Append(ColourGradient.GradientName);
            builder.Append("' to the the greyscale image '");
            builder.Append(BaseChannel.ImageName);
            builder.Append("'");
            builder.Append(Environment.NewLine);

            if ((regenerateExistingFiles) && (AnyOutputFilesExist(filePath, OutputFormats)))
            {
                builder.Append("# Warning: one or more variations of the target file already exist, but will be overwritten as the option to regenerate existing files is enabled.");
                builder.Append(Environment.NewLine);
                _WillOverwriteExistingData = true;
            }

            string bcScriptObjectName = _TemporaryObjectNamePrefix + "base_channel";
            string grScriptObjectName = _TemporaryObjectNamePrefix + "gradient";

            builder.Append(BaseChannel.GetReadScriptCode(bcScriptObjectName));
            builder.Append(ColourGradient.GetCreationScriptCode(grScriptObjectName));

            builder.Append(scriptTargetObjectName);
            builder.Append(" = applypalette(");
            builder.Append(bcScriptObjectName);
            builder.Append(", ");
            builder.Append(grScriptObjectName);
            builder.Append(", min = 0.0, max = 1.0)");
            builder.Append(Environment.NewLine);

            // write the image to disk
            builder.Append(GetImageFileWriteCode(scriptTargetObjectName, regenerateExistingFiles));

            // de-allocate temporary objects
            builder.Append(GetDeallocationScriptCode(bcScriptObjectName));
            builder.Append(GetDeallocationScriptCode(grScriptObjectName));

            return builder.ToString();
        }

    }
}
