// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class ImageCube
    {
        protected Hashtable _SpectralBands;
        protected Hashtable _GreyscaleChannels;
        protected GreyscaleChannelSet[] _GreyscaleChannelSets;
        protected GreyscaleChannelSet[] _StatisticalCubeSets;
        protected ThreeChannelImage[] _ThreeChannelImages;
        protected LumaColourImage[] _TintedChannels;
        protected ColourGradientImage[] _ColourGradientImages;
        protected LumaColourImage[] _LumaColourImages;
        protected DCSImage[] _DCSImages;

        //protected GreyscaleChannel[] _RatioChannels;
        //protected GreyscaleChannel[] _DifferenceChannels;
        protected CalculatedChannelSet[] _CalculatedChannels;

        protected string _BaseInputDirectory;
        protected string _BaseOutputDirectory;

        protected TMSBInputConfiguration _AbstractInputConfig;
        protected TMSBInputConfiguration _InputConfig;
        protected TMSBColourConfiguration _ColourConfig;
        protected TMSBProcessingConfiguration _ProcessingConfig;
        protected TMSBOutputConfiguration _OutputConfig;

        //protected const int _MaxRecursionDepthCalculatedChannels = 2;

        protected TMSBEngine _ParentEngine;

        protected TransformationProfileCollection _TransformationProfiles;

        #region Properties

        public GreyscaleChannelSet[] GreyscaleChannelSets
        {
            get
            {
                return _GreyscaleChannelSets;
            }
            set
            {
                _GreyscaleChannelSets = value;
            }
        }

        public GreyscaleChannelSet[] StatisticalCubeSets
        {
            get
            {
                return _StatisticalCubeSets;
            }
            set
            {
                _StatisticalCubeSets = value;
            }
        }

        public ThreeChannelImage[] ThreeChannelImages
        {
            get
            {
                return _ThreeChannelImages;
            }
            set
            {
                _ThreeChannelImages = value;
            }
        }

        public LumaColourImage[] TintedChannels
        {
            get
            {
                return _TintedChannels;
            }
            set
            {
                _TintedChannels = value;
            }
        }

        public ColourGradientImage[] ColourGradientImages
        {
            get
            {
                return _ColourGradientImages;
            }
            set
            {
                _ColourGradientImages = value;
            }
        }

        public LumaColourImage[] LumaColourImages
        {
            get
            {
                return _LumaColourImages;
            }
            set
            {
                _LumaColourImages = value;
            }
        }

        public DCSImage[] DCSImages
        {
            get
            {
                return _DCSImages;
            }
            set
            {
                _DCSImages = value;
            }
        }

        public TMSBInputConfiguration AbstractInputConfig
        {
            get
            {
                return _AbstractInputConfig;
            }
            set
            {
                _AbstractInputConfig = value;
                SetInputConfiguration();
            }
        }

        /// <summary>
        /// This property is read-only because it is created based on the AbstractInputConfig object.
        /// To modify the input configuration, set the value of AbstractInputConfig and this object
        /// will be updated automatically.
        /// </summary>
        public TMSBInputConfiguration InputConfig
        {
            get
            {
                return _InputConfig;
            }
            //set
            //{
            //    _InputConfig = value;
            //}
        }

        public TMSBColourConfiguration ColourConfig
        {
            get
            {
                return _ColourConfig;
            }
            set
            {
                _ColourConfig = value;
            }
        }

        public TMSBProcessingConfiguration ProcessingConfig
        {
            get
            {
                return _ProcessingConfig;
            }
            set
            {
                _ProcessingConfig = value;
            }
        }

        public TMSBOutputConfiguration OutputConfig
        {
            get
            {
                return _OutputConfig;
            }
            set
            {
                _OutputConfig = value;
            }
        }

        public TransformationProfileCollection TransformationProfiles
        {
            get
            {
                return _TransformationProfiles;
            }
            set
            {
                _TransformationProfiles = value;
            }
        }

        #endregion

        public ImageCube(TMSBEngine parentEngine)
        {
            _ParentEngine = parentEngine;
            _SpectralBands = new Hashtable();
            _GreyscaleChannels = new Hashtable();
            _GreyscaleChannelSets = new GreyscaleChannelSet[0];
            _StatisticalCubeSets = new GreyscaleChannelSet[0];

            _ThreeChannelImages = new ThreeChannelImage[0];
            _LumaColourImages = new LumaColourImage[0];
            _TintedChannels = new LumaColourImage[0];
            _ColourGradientImages = new ColourGradientImage[0];
            _DCSImages = new DCSImage[0];

            _BaseInputDirectory = "";
            _BaseOutputDirectory = "";

            _InputConfig = new TMSBInputConfiguration();
            ColourConfig = new TMSBColourConfiguration();
            _ProcessingConfig = new TMSBProcessingConfiguration();
            _OutputConfig = new TMSBOutputConfiguration();

            _CalculatedChannels = new CalculatedChannelSet[0];
            TransformationProfiles = new TransformationProfileCollection();
        }

        public ImageCube(TMSBEngine parentEngine, string baseInputDirectory, string baseOutputDirectory, 
            TransformationProfileCollection transformConfig, TMSBInputConfiguration inputConfig,
            TMSBColourConfiguration colourConfig, TMSBProcessingConfiguration processingConfig, TMSBOutputConfiguration outputConfig)
            : this(parentEngine)
        {
            _BaseInputDirectory = baseInputDirectory;
            _BaseOutputDirectory = baseOutputDirectory;
            TransformationProfiles = transformConfig;
            AbstractInputConfig = inputConfig;
            ColourConfig = colourConfig;
            ProcessingConfig = processingConfig;
            OutputConfig = outputConfig;
        }

        protected ImageCube(string tempDirectory, TransformationProfileCollection transformConfig, TMSBInputConfiguration inputConfig,
            TMSBColourConfiguration colourConfig, TMSBProcessingConfiguration processingConfig, TMSBOutputConfiguration outputConfig)
            : this(null)
        {
            _BaseInputDirectory = tempDirectory;
            _BaseOutputDirectory = tempDirectory;
            TransformationProfiles = transformConfig;
            _InputConfig = inputConfig;
            ColourConfig = colourConfig;
            ProcessingConfig = processingConfig;
            OutputConfig = outputConfig;
            AddBandsFromInputConfig();
        }

        public static ImageCube GetTemporaryCube(string tempDirectory, TransformationProfileCollection transformConfig, TMSBInputConfiguration inputConfig,
            TMSBColourConfiguration colourConfig, TMSBProcessingConfiguration processingConfig, TMSBOutputConfiguration outputConfig)
        {
            return new ImageCube(tempDirectory, transformConfig, inputConfig, colourConfig, processingConfig, outputConfig);
        }

        ~ImageCube()
        {
            _ParentEngine = null;
        }

        public void AddGreyscaleChannelSet(GreyscaleChannelSet newSet)
        {
            Array.Resize(ref _GreyscaleChannelSets, _GreyscaleChannelSets.Length + 1);
            GreyscaleChannelSets[GreyscaleChannelSets.Length - 1] = newSet;
        }

        protected void SetInputConfiguration()
        {
            _InputConfig = TMSBInputConfiguration.CreateActualizedInputConfig(_AbstractInputConfig, _BaseInputDirectory);
            //foreach (SpectralBand band in InputConfig.SpectralBands)
            //{
            //    AddSpectralBand(band);
            //}
            AddBandsFromInputConfig();
        }

        protected void AddBandsFromInputConfig()
        {
            foreach (SpectralBand band in InputConfig.SpectralBands)
            {
                AddSpectralBand(band);
            }
        }

        //protected void BuildTintMappings()
        //{
        //    // iterate through the list of tint mappings
        //    // if the referenced channel exists, add any mappings associated with it
        //    // associate the base plane of all tint mappings with a master object
        //    foreach (TintMapping tintMapping in ProcessingConfig.TintMappings)
        //    {
        //        string channelName = "";
        //        bool foundChannel = true;
        //        try
        //        {
        //            channelName = tintMapping.Channel.GetChannelName(this);
        //        }
        //        catch (Exception ex)
        //        {
        //            CreateDebugMessage("The band or channel referenced for a tint mapping ('" + tintMapping.Channel.ReferenceName +
        //                "') does not appear to exist in the image cube. The tinting colour specified will not be used.");
        //            foundChannel = false;
        //        }
        //        if ((foundChannel == true) && (channelName == ""))
        //        {
        //            CreateDebugMessage("Error: an empty string was passed as a reference to a channel as part of a tint mapping. This mapping will be ignored.");
        //            foundChannel = false;
        //        }

        //        if (foundChannel)
        //        {
        //            GreyscaleChannel chan = GetGreyscaleChannel(channelName);
        //            foreach (string tintName in tintMapping.TintNames)
        //            {
        //                if (ProcessingConfig.ContainsTintingColour(tintName))
        //                {
        //                    TintingColour colour = ProcessingConfig.GetTintingColour(tintName);
        //                    chan.AddTintingColour(colour);
        //                }
        //            }
        //            UpdateGreyscaleChannel(chan);
        //        }

        //        //if (ContainsGreyscaleChannel(tintMapping.Channel.ReferenceName))
        //        //{
        //        //    GreyscaleChannel chan = GetGreyscaleChannel(tintMapping.ChannelName);
        //        //    foreach (string tintName in tintMapping.TintNames)
        //        //    {
        //        //        if (InputConfig.ContainsTintingColour(tintName))
        //        //        {
        //        //            TintingColour colour = InputConfig.GetTintingColour(tintName);
        //        //            chan.AddTintingColour(colour);
        //        //        }
        //        //    }
        //        //    UpdateGreyscaleChannel(chan);
        //        //}
        //    }
        //}

        public void GenerateChannels()
        {
            // generate greyscale channels
            CreateDebugMessage("Generating channels.");

            GreyscaleChannel[] channels = GetChannelPermutations();
            foreach (GreyscaleChannel channel in channels)
            {
                //_GreyscaleChannels.Add(channel.ImageName, channel);
                AddGreyscaleChannel(channel);
            }

            CreateDebugMessage("Done generating channels.");
        }

        public void GeneratePermutations()
        {
            CreateDebugMessage("Generating permutations.");
            GreyscaleChannel[] channels = GetSortedGreyscaleChannels();

            // generate three-channel images
            if (!ProcessingConfig.GenerateThreeChannelImages)
            {
                CreateDebugMessage("The GenerateThreeChannelImages option is disabled. Skipping creation of any three-channel images.");
            }
            else
            {
                CreateDebugMessage("The GenerateThreeChannelImages option is enabled.");
                _ThreeChannelImages = GetThreeChannelImages(channels);
            }

            // generate tinted channels
            if (!ProcessingConfig.GenerateTintedImages)
            {
                CreateDebugMessage("The GenerateTintedImages option is disabled. Skipping creation of any tinted images.");
            }
            else
            {
                CreateDebugMessage("The GenerateTintedImages option is enabled.");
                _TintedChannels = GetTintedChannels(channels);
            }

            // generate gradient images
            if (!ProcessingConfig.GenerateGradientImages)
            {
                CreateDebugMessage("The GenerateGradientImages option is disabled. Skipping creation of any tinted images.");
            }
            else
            {
                CreateDebugMessage("The GenerateGradientImages option is enabled.");
                ColourGradientImages = GetColourGradientImages();
            }

            // DCS image permutations
            if (!ProcessingConfig.GenerateDCSImages)
            {
                CreateDebugMessage("The GenerateDCSImages option is disabled. Skipping creation of any DCS images.");
            }
            else
            {
                CreateDebugMessage("The GenerateDCSImages option is enabled.");
                if (ProcessingConfig.GenerateDCSImagesFromCombinedChannels || ProcessingConfig.GenerateDCSImagesFromIndividualChannels || ProcessingConfig.GenerateDCSImagesFromCustomThreeChannelImages)
                {
                    _DCSImages = GetDCSImages();
                }
            }

            // generate other luma/colour variations
            if ((!ProcessingConfig.GenerateDCSEnhancedContrast) && (!ProcessingConfig.GenerateLumaColourImages))
            {
                CreateDebugMessage("The GenerateLumaColourImages option is disabled. Skipping creation of any Luma/Colour images.");
            }
            else
            {
                CreateDebugMessage("The GenerateLumaColourImages option is enabled.");
                _LumaColourImages = GetLumaColourImages();
            }

            CreateDebugMessage("Done generating permutations.");
        }

        public void AddSpectralBand(SpectralBand newBand)
        {
            if (_SpectralBands.Contains(newBand.BandName))
            {
                throw new ArgumentException("The image cube already contains a band named '" + newBand.BandName + "'.");
            }
            _SpectralBands.Add(newBand.BandName, newBand);
        }

        public SpectralBand GetSpectralBand(string bandName)
        {
            if (!_SpectralBands.Contains(bandName))
            {
                throw new ArgumentException("The image cube does not contain a band named '" + bandName + "'.");
            }
            return (SpectralBand)_SpectralBands[bandName];
        }

        public bool ContainsBand(string bandName)
        {
            return _SpectralBands.Contains(bandName);
        }

        public string[] GetBandNames()
        {
            SpectralBand[] bandList = GetSortedBands();
            string[] bandNames = new string[bandList.Length];
            for (int i = 0; i < bandList.Length; i++)
            {
                SpectralBand currentBand = (SpectralBand)bandList[i];
                bandNames[i] = currentBand.BandName;
            }
            return bandNames;
        }

        public SpectralBand[] GetSortedBands()
        {
            ArrayList bandList = new ArrayList();
            foreach (string key in _SpectralBands.Keys)
            {
                bandList.Add(GetSpectralBand(key));
            }
            bandList.Sort();
            SpectralBand b = new SpectralBand("proto");
            return (SpectralBand[])bandList.ToArray(b.GetType());
        }

        //protected string GetCalculatedPermutationName(string channel1Name, string channel2Name, BandCombineMethod combineMethod)
        //{
        //    StringBuilder builder = new StringBuilder();
        //    switch (combineMethod)
        //    {
        //        case BandCombineMethod.Ratio:
        //            builder.Append("Ratio-");
        //            break;
        //        case BandCombineMethod.Difference:
        //            builder.Append("Difference-");
        //            break;
        //        default:
        //            throw new ArgumentException("The band combination method '" + combineMethod.ToString() + "' is not supported by this function.");
        //            break;
        //    }
        //    builder.Append(OutputImage.WrapNameComponentIfComplex(channel1Name));
        //    switch (combineMethod)
        //    {
        //        case BandCombineMethod.Ratio:
        //            builder.Append("_To_");
        //            break;
        //        case BandCombineMethod.Difference:
        //            builder.Append("-");
        //            break;
        //    }
        //    builder.Append(OutputImage.WrapNameComponentIfComplex(channel2Name));
        //    return builder.ToString();
        //}

        //protected string GetCalculatedPermutationScriptCode(string channel1Name, string channel2Name, BandCombineMethod combineMethod, CalculatedChannelCombinationConfig combineConfig)
        //{
        //    StringBuilder builder = new StringBuilder();
        //    builder.Append("= ");
        //    if (combineConfig.AbsoluteValueOutput)
        //    {
        //        builder.Append("abs(");
        //    }
        //    switch (combineMethod)
        //    {
        //        case BandCombineMethod.Ratio:
        //            // Clipped in order to prevent the extreme outliers from preventing a useful image being created
        //            // (due to the inability of the aggressive normalization function to find a match without using an 
        //            // extremely high-resolution histogram)
        //            //builder.Append("clipData((%ChannelByName:CHANNEL1% / %ChannelByName:CHANNEL2%), clipMin = 0.0, clipMax = 100.0)");
        //            builder.Append("%ChannelByName:CHANNEL1% / %ChannelByName:CHANNEL2%");
        //            break;
        //        case BandCombineMethod.Difference:
        //            builder.Append("%ChannelByName:CHANNEL1% - %ChannelByName:CHANNEL2%");
        //            break;
        //        default:
        //            throw new ArgumentException("The band combination method '" + combineMethod.ToString() + "' is not supported by this function.");
        //            break;
        //    }
        //    if (combineConfig.AbsoluteValueOutput)
        //    {
        //        builder.Append(")");
        //    }
        //    string scriptCode = builder.ToString();
        //    scriptCode = scriptCode.Replace("CHANNEL1", channel1Name);
        //    scriptCode = scriptCode.Replace("CHANNEL2", channel2Name);
        //    return scriptCode;
        //}

        public GreyscaleChannel[] GetCalculatedChannelPermutations(GreyscaleChannel[] existingChannels, CalculatedChannelCombinationConfig combineConfig, int currentRecursionDepth)
        {
            ArrayList updatedList = new ArrayList();
            ArrayList nameList = new ArrayList();

            int currentPermCount = existingChannels.Length;

            foreach (GreyscaleChannel chan in existingChannels)
            {
                updatedList.Add(chan);
                nameList.Add(chan.ImageName);
            }

            for (int c1Num = 0; c1Num < (existingChannels.Length - 1); c1Num++)
            {
                GreyscaleChannel c1 = existingChannels[c1Num];
                for (int c2Num = c1Num + 1; c2Num < existingChannels.Length; c2Num++)
                {
                    GreyscaleChannel c2 = existingChannels[c2Num];
                    if (c1.ImageName != c2.ImageName)
                    {
                        if (c1.HasMutuallyExclusiveBands(c2))
                        {
                            CreateDebugMessage("Channel '" + c1.ImageName + "' is based on one or more bands which are mutually exclusive with one or more bands in channel '" +
                                c2.ImageName + "'. They will not be combined in this calculated permutation.");
                        }
                        else
                        {
                            AddPermutationIfNew(ref updatedList, ref nameList, c1, c2, combineConfig);
                            // add the inverse version if so configured
                            if (combineConfig.GenerateReciprocal)
                            {
                                AddPermutationIfNew(ref updatedList, ref nameList, c2, c1, combineConfig);
                            }
                        }
                    }
                }
            }

            int newPermCount = updatedList.Count;
            GreyscaleChannel protochannel = new GreyscaleChannel(1, BandCombineMethod.Average);

            if ((currentPermCount == newPermCount) || (currentRecursionDepth >= combineConfig.MaxRecursionDepth))
            {
                // unwind to the previous level
                return (GreyscaleChannel[])updatedList.ToArray(protochannel.GetType());
            }
            else
            {
                // recurse to the next level

                return GetCalculatedChannelPermutations((GreyscaleChannel[])updatedList.ToArray(protochannel.GetType()), combineConfig, currentRecursionDepth + 1);
            }
        }

        protected void AddPermutationIfNew(ref ArrayList updatedList, ref ArrayList nameList, GreyscaleChannel c1, GreyscaleChannel c2, CalculatedChannelCombinationConfig combineConfig)
        {
            string newPermName = combineConfig.GetCalculatedPermutationName(c1.ImageName, c2.ImageName);
            if (nameList.Contains(newPermName))
            {
                CreateDebugMessage("Calculated Permutation '" + newPermName + "' already exists and will not be added again.");
            }
            else
            {
                CreateDebugMessage("Calculated Permutation '" + newPermName + "' does not exist yet and will be added.");
                GreyscaleChannel newPerm = new GreyscaleChannel();
                newPerm.AddReferencedChannel(c1);
                newPerm.AddReferencedChannel(c2);
                newPerm.UserDefinedName = newPermName;
                //newPerm.BandCombineMode = BandCombineMethod.Custom | combineMethod;
                newPerm.BandCombineMode = BandCombineMethod.Custom;
                newPerm.CustomBandCombinationScriptCode = combineConfig.GetCalculatedPermutationScriptCode(c1.ImageName, c2.ImageName);
                newPerm.InputTransformations = combineConfig.InputTransformations;
                newPerm.OutputTransformations = combineConfig.OutputTransformations;
                newPerm.UseAsThreeChannelChannel = combineConfig.UseAsStandardThreeChannelChannel;
                newPerm.UseAsLuminanceChannel = combineConfig.UseAsLuminanceChannel;
                newPerm.ParentCube = this;
                newPerm.OutputFormats = OutputConfig.OutputFormats;
                newPerm.CustomDescription = CalculatedChannelCombinationConfig.GetCalculatedChannelCombinationMethodName(combineConfig.CombineMethod);
                updatedList.Add(newPerm);
                nameList.Add(newPermName);
            }
        }

        public GreyscaleChannel[] GetChannelPermutations()
        {
            SpectralBand[] bands = GetSortedBands();

            SpectralBand protoband = new SpectralBand("prototype");

            ArrayList permutationList = new ArrayList();
            GreyscaleChannel protochannel = new GreyscaleChannel(1, BandCombineMethod.Average);

            // generate channels based on all bands in the cube, if any such options are enabled
            if (!ProcessingConfig.GenerateStatisticalChannels)
            {
                CreateDebugMessage("The GenerateStatisticalChannels option is disabled. Skipping creation of any statistical greyscale channels.");
            }
            else
            {
                CreateDebugMessage("The GenerateStatisticalChannels option is enabled.");
                if (ProcessingConfig.StatisticalGreyscaleCubeConfigurations.Length > 0)
                {
                    _StatisticalCubeSets = new GreyscaleChannelSet[ProcessingConfig.StatisticalGreyscaleCubeConfigurations.Length];
                    for (int i = 0; i < ProcessingConfig.StatisticalGreyscaleCubeConfigurations.Length; i++)
                    {
                        StatisticalCubeConfig cscc = ProcessingConfig.StatisticalGreyscaleCubeConfigurations[i];
                        GreyscaleChannelSet allBandSet = new GreyscaleChannelSet();
                        allBandSet.ChannelSetName = cscc.CubeName + " Statistical Greyscale Channels";
                        allBandSet.MomentCubeWindowSize = cscc.WindowSize;
                        allBandSet.UseAsThreeChannelChannels = cscc.UseAsThreeChannelChannels;
                        foreach (CombinedBandConfig statConfig in cscc.CombinedBandConfigurations)
                        {
                            GreyscaleChannel statChannel = new GreyscaleChannel(bands.Length, statConfig.BandCombineMode);
                            statChannel.UserDefinedName = cscc.CubeName + GreyscaleChannel.GetBandCombineModeNameSuffix(statConfig.BandCombineMode);
                            statChannel.SpectralBands = bands;
                            statChannel.UsesAllSpectralBands = true;
                            //statChannel.RenameAllBandsFiles = OutputConfig.RenameAllBandsFiles;
                            statChannel.OutputFormats = OutputConfig.OutputFormats;
                            statChannel.OutputTransformations = statConfig.OutputTransformations;
                            //statChannel.OutputNormalization = statConfig.OutputNormalization;
                            statChannel.UseAsLuminanceChannel = statConfig.UseAsLuminanceChannel;
                            allBandSet.AddChannel(statChannel);
                            permutationList.Add(statChannel);
                        }
                        AddGreyscaleChannelSet(allBandSet);
                        _StatisticalCubeSets[i] = allBandSet;
                    }
                }
            }

            CreateDebugMessage("Generating spectral band permutations.");
            for (int startBand = 0; startBand < bands.Length; startBand++)
            {
                SpectralBandFlagList fl = new SpectralBandFlagList();
                for (int bandNum = startBand; bandNum < bands.Length; bandNum++)
                {
                    fl.AddBand(bands[bandNum]);
                }

                while (fl.GetBand(0).Flag == true)
                {
                    ArrayList componentList = new ArrayList();
                    bool isAllBands = false;
                    bool isRGB = false;
                    if (startBand == 0)
                    {
                        isAllBands = true;
                    }
                    for (int i = 0; i < fl.GetLength(); i++)
                    {
                        SpectralBandWithFlag bwf = fl.GetBand(i);
                        if (bwf.Flag == true)
                        {
                            componentList.Add(bwf.Band);
                        }
                        else
                        {
                            isAllBands = false;
                        }
                    }

                    SpectralBand[] bandsForChannel = (SpectralBand[])componentList.ToArray(protoband.GetType());
                    GreyscaleChannelSet currentBandSet = new GreyscaleChannelSet();
                    if (componentList.Count == 1)
                    {
                        GreyscaleChannel permChannel = new GreyscaleChannel(bandsForChannel, BandCombineMethod.Average);
                        permChannel.OutputFormats = OutputConfig.OutputFormats;
                        permChannel.UseAsLuminanceChannel = bandsForChannel[0].UseAsLumaChannel;
                        permutationList.Add(permChannel);
                        currentBandSet.AddChannel(permChannel);
                    }
                    else
                    {
                            if ((componentList.Count == 3) &&
                                (bandsForChannel[0].BandName == InputConfig.RedBandName) &&
                                (bandsForChannel[1].BandName == InputConfig.GreenBandName) &&
                                (bandsForChannel[2].BandName == InputConfig.BlueBandName))
                            {
                                isRGB = true;
                            }

                            foreach (CombinedBandConfig channelConfig in ProcessingConfig.ChannelCombinationMethods)
                            {
                                if ((isRGB) || ((!isRGB) && (channelConfig.BandCombineMode != BandCombineMethod.Luminance)))
                                {
                                    GreyscaleChannel currentChannel = new GreyscaleChannel(bandsForChannel.Length, channelConfig.BandCombineMode);
                                    currentChannel.SpectralBands = bandsForChannel;
                                    currentChannel.UsesAllSpectralBands = false;
                                    currentChannel.OutputFormats = OutputConfig.OutputFormats;
                                    currentChannel.OutputTransformations = channelConfig.OutputTransformations;
                                    //currentChannel.OutputNormalization = channelConfig.OutputNormalization;
                                    currentChannel.UseAsLuminanceChannel = channelConfig.UseAsLuminanceChannel;
                                    currentBandSet.AddChannel(currentChannel);
                                    permutationList.Add(currentChannel);
                                }
                            }
                    }
                    if (currentBandSet.Channels.Length > 0)
                    {
                        AddGreyscaleChannelSet(currentBandSet);
                    }
                    fl.IncrementFlags();
                }
            }

            // calculated channels (ratio, difference, etc)
            // to do: add an option to create these from all other bands
            // IE this will need to go down to the end of this method
            if (!ProcessingConfig.GenerateCalculatedChannels)
            {
                CreateDebugMessage("The GenerateCalculatedChannels option is disabled. Skipping creation of any calculated greyscale channels.");
            }
            else
            {
                CreateDebugMessage("The GenerateCalculatedChannels option is enabled.");
                _CalculatedChannels = new CalculatedChannelSet[ProcessingConfig.CalculatedChannelConfigurations.Length];
                for (int cc = 0; cc < ProcessingConfig.CalculatedChannelConfigurations.Length; cc++)
                {
                    _CalculatedChannels[cc] = new CalculatedChannelSet();
                    _CalculatedChannels[cc].Config = ProcessingConfig.CalculatedChannelConfigurations[cc];
                }

                if (_CalculatedChannels.Length > 0)
                {
                    for (int ccSetNum = 0; ccSetNum < _CalculatedChannels.Length; ccSetNum++)
                    {
                        ArrayList ibChannelList = new ArrayList();
                        foreach (GreyscaleChannel chan in permutationList)
                        {
                            if (chan.SpectralBands.Length == 1)
                            {
                                ibChannelList.Add(chan);
                            }
                        }
                        GreyscaleChannel[] ibChannels = (GreyscaleChannel[])ibChannelList.ToArray(protochannel.GetType());

                        _CalculatedChannels[ccSetNum].Channels = GetCalculatedChannelPermutations(ibChannels, _CalculatedChannels[ccSetNum].Config, 1);

                        foreach (GreyscaleChannel cChan in _CalculatedChannels[ccSetNum].Channels)
                        {
                            if (cChan.ReferencedChannels.Length == 2)
                            {
                                GreyscaleChannelSet ccSet = new GreyscaleChannelSet();
                                ccSet.AddChannel(cChan);
                                permutationList.Add(cChan);
                                AddGreyscaleChannelSet(ccSet);
                            }
                        }
                    }
                }
            }


            // custom-defined channels
            if (!ProcessingConfig.GenerateCustomChannels)
            {
                CreateDebugMessage("The GenerateCustomChannels option is disabled. Skipping creation of any custom-defined greyscale channels.");
            }
            else
            {
                CreateDebugMessage("The GenerateCustomChannels option is enabled.");
                foreach (CustomGreyscaleChannelConfig cust in ProcessingConfig.CustomGreyscaleChannelConfigurations)
                {
                    try
                    {
                        GreyscaleChannel customChannel = new GreyscaleChannel();
                        switch (cust.ChannelType)
                        {
                            case CustomGreyscaleChannelType.BandCategory:
                                customChannel = new GreyscaleChannel(InputConfig.GetBandsByCategory(cust.Category), cust.BandCombineMode);
                                break;
                            case CustomGreyscaleChannelType.CustomScriptCode:
                                customChannel.CustomBandCombinationScriptCode = cust.CustomScriptCode;
                                customChannel.BandCombineMode = BandCombineMethod.Custom;
                                break;
                        }
                        customChannel.UserDefinedName = cust.ChannelName;
                        customChannel.UseAsLuminanceChannel = cust.UseAsLuminanceChannel;
                        customChannel.UseAsThreeChannelChannel = cust.UseAsThreeChannelChannel;
                        //customChannel.OutputNormalization = cust.OutputNormalization;
                        customChannel.InputTransformations = cust.InputTransformations;
                        customChannel.OutputTransformations = cust.OutputTransformations;
                        customChannel.OutputFormats = OutputConfig.OutputFormats;
                        customChannel.ParentCube = this;
                        customChannel.IsCustomImage = true;
                        //_GreyscaleChannels.Add(customChannel.ImageName, customChannel);
                        if (customChannel.PrerequisitesExist())
                        {
                            GreyscaleChannelSet currentBandSet = new GreyscaleChannelSet();
                            permutationList.Add(customChannel);
                            currentBandSet.AddChannel(customChannel);
                            AddGreyscaleChannelSet(currentBandSet);
                        }
                        else
                        {
                            CreateDebugMessage("One or more of the prerequisites for the custom channel '" + customChannel.ImageName +
                                "' is missing from the input data. This channel and any images which depend on it will not be created.");
                        }
                    }
                    catch (Exception ex)
                    {
                        // Do nothing for now
                        CreateDebugMessage("Error instantiating the custom greyscale channel '" + cust.ChannelName + "': '" + ex.Message + "'.");
                    }
                }
            }

            return (GreyscaleChannel[])permutationList.ToArray(protochannel.GetType());
        }

        public int GetMinimumBandNumber()
        {
            ArrayList bandNumbers = GetBandNumbers();
            return (int)bandNumbers[0];
        }

        public int GetMaximumBandNumber()
        {
            ArrayList bandNumbers = GetBandNumbers();
            return (int)bandNumbers[bandNumbers.Count - 1];
        }

        protected ArrayList GetBandNumbers()
        {
            ArrayList bandNumbers = new ArrayList(_SpectralBands.Count);
            foreach (SpectralBand band in _SpectralBands.Values)
            {
                bandNumbers.Add(band.BandNumber);
            }
            bandNumbers.Sort();
            return bandNumbers;
        }

        public void AddGreyscaleChannel(GreyscaleChannel newChannel)
        {
            if (_GreyscaleChannels.Contains(newChannel.ImageName))
            {
                throw new ArgumentException("The image cube already contains a channel named '" + newChannel.ImageName + "'.");
            }
            _GreyscaleChannels.Add(newChannel.ImageName, newChannel);
        }

        public GreyscaleChannel GetGreyscaleChannel(string channelName)
        {
            if (!_GreyscaleChannels.Contains(channelName))
            {
                throw new ArgumentException("The image cube does not contain a channel named '" + channelName + "'.");
            }
            return (GreyscaleChannel)_GreyscaleChannels[channelName];
        }

        public void UpdateGreyscaleChannel(GreyscaleChannel channel)
        {
            if (!_GreyscaleChannels.Contains(channel.ImageName))
            {
                AddGreyscaleChannel(channel);
                return;
            }
            _GreyscaleChannels[channel.ImageName] = channel;
        }

        public bool ContainsGreyscaleChannel(string channelName)
        {
            //return _GreyscaleChannels.Contains(channelName);
            for (int i = 0; i < _GreyscaleChannelSets.Length; i++)
            {
                if (_GreyscaleChannelSets[i].ContainsChannel(channelName))
                {
                    return true;
                }
            }
            return false;
        }

        public string[] GetGreyscaleChannelNames()
        {
            //GreyscaleChannel[] channelList = GetSortedGreyscaleChannels();
            //string[] channelNames = new string[channelList.Length];
            //for (int i = 0; i < channelList.Length; i++)
            //{
            //    GreyscaleChannel currentChannel = (GreyscaleChannel)channelList[i];
            //    channelNames[i] = currentChannel.ImageName;
            //}
            //return channelNames;
            ArrayList names = new ArrayList();
            for (int i = 0; i < _GreyscaleChannelSets.Length; i++)
            {
                string[] cNames = _GreyscaleChannelSets[i].GetChannelNames();
                for (int j = 0; j < cNames.Length; j++)
                {
                    names.Add(cNames[j]);
                }
            }
            string[] proto = new string[] { "1", "2" };
            return (string[])names.ToArray(proto.GetType());
        }

        public GreyscaleChannel[] GetSortedGreyscaleChannels()
        {
            CreateDebugMessage("Obtaining sorted greyscale channels.");
            ArrayList channelList = new ArrayList();
            foreach (string key in _GreyscaleChannels.Keys)
            {
                channelList.Add(GetGreyscaleChannel(key));
            }
            channelList.Sort();
            GreyscaleChannel b = new GreyscaleChannel(1, BandCombineMethod.Average);
            return (GreyscaleChannel[])channelList.ToArray(b.GetType());
            CreateDebugMessage("Done obtaining sorted greyscale channels.");
        }

        public ThreeChannelImage[] GetStatisticalThreeChannelImages(GreyscaleChannel[] sourceChannels)
        {
            ArrayList imageList = new ArrayList();

            CreateDebugMessage("Begin statistical three channel image permutations.");

            int redEnd = (sourceChannels.Length - 2);
            if (ProcessingConfig.GenerateStatisticalImageThreeChannelPermutationsUnrestricted)
            {
                redEnd = sourceChannels.Length;
            }

            for (int redChannelNum = 0; redChannelNum < redEnd; redChannelNum++)
            {
                int greenBegin = redChannelNum + 1;
                int greenEnd = (sourceChannels.Length - 1);

                if (ProcessingConfig.GenerateStatisticalImageThreeChannelPermutationsUnrestricted)
                {
                    greenBegin = 0;
                    greenEnd = sourceChannels.Length;
                }

                for (int greenChannelNum = greenBegin; greenChannelNum < greenEnd; greenChannelNum++)
                {
                    if (redChannelNum == greenChannelNum)
                    {
                        CreateDebugMessage("The red and green channels are identical. This branch of permutation will not be used.");
                    }
                    else
                    {
                        int blueBegin = greenChannelNum + 1;

                        if (ProcessingConfig.GenerateStatisticalImageThreeChannelPermutationsUnrestricted)
                        {
                            blueBegin = 0;
                        }

                        for (int blueChannelNum = blueBegin; blueChannelNum < (sourceChannels.Length); blueChannelNum++)
                        {
                            bool includeImage = true;
                            if (redChannelNum == blueChannelNum)
                            {
                                CreateDebugMessage("The red and blue channels are identical. This branch of permutation will not be used.");
                                includeImage = false;
                            }
                            if (greenChannelNum == blueChannelNum)
                            {
                                CreateDebugMessage("The green and blue channels are identical. This branch of permutation will not be used.");
                                includeImage = false;
                            }

                            if (includeImage)
                            {
                                GreyscaleChannel[] tcic = new GreyscaleChannel[3];
                                tcic[0] = sourceChannels[redChannelNum];
                                tcic[1] = sourceChannels[greenChannelNum];
                                tcic[2] = sourceChannels[blueChannelNum];

                                ThreeChannelImage tci = new ThreeChannelImage(tcic);

                                //tci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                                tci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                                tci.OutputFormats = OutputConfig.OutputFormats;
                                tci.UseInStandardLumaColourImages = ProcessingConfig.GenerateLumaColourVariationsFromStatisticalChannels;
                                tci.ParentCube = this;
                                tci.NamePrefix = "FC-3C-Statistical-";
                                imageList.Add(tci);
                            }
                        }
                    }
                }
            }

            CreateDebugMessage("End statistical three channel image permutations.");

            ThreeChannelImage tcProto = new ThreeChannelImage();
            return (ThreeChannelImage[])imageList.ToArray(tcProto.GetType());
        }


        public ThreeChannelImage[] GetCalculatedThreeChannelImages(GreyscaleChannel[] sourceChannels, CalculatedChannelCombinationConfig combineConfig)
        {
            ArrayList imageList = new ArrayList();

            SpectralBand[] bands = GetSortedBands();

            CreateDebugMessage("Begin calculated three channel image permutations - method '" + combineConfig.CombineMethod.ToString() +
                "'.");

            // I really, really tried to think of a way to do this using a recursive function.
            for (int redBand1 = 0; redBand1 < bands.Length; redBand1++)
            {
                CreateDebugMessage("\tRed band 1: '" + bands[redBand1].BandName + "'.");
                for (int redBand2 = 1; redBand2 < bands.Length; redBand2++)
                {
                    CreateDebugMessage("\t\tRed band 2: '" + bands[redBand2].BandName + "'.");
                    bool continueEvalLZeroRed = true;
                    if (redBand1 == redBand2)
                    {
                        CreateDebugMessage("The two red bands are identical. This branch of permutation will not be used.");
                        continueEvalLZeroRed = false;
                    }
                    if ((continueEvalLZeroRed) && (redBand1 > redBand2))
                    {
                        CreateDebugMessage("The first red band is greater than the second. This branch of permutation will not be used.");
                        continueEvalLZeroRed = false;
                    }
                    if (continueEvalLZeroRed)
                    {
                        for (int greenBand1 = 0; greenBand1 < bands.Length; greenBand1++)
                        {
                            CreateDebugMessage("\t\t\tGreen band 1: '" + bands[greenBand1].BandName + "'.");
                            for (int greenBand2 = 1; greenBand2 < bands.Length; greenBand2++)
                            {
                                CreateDebugMessage("\t\t\t\tGreen band 2: '" + bands[greenBand2].BandName + "'.");
                                bool continueEvalL1 = true;

                                if (greenBand1 == greenBand2)
                                {
                                    CreateDebugMessage("The two green bands are identical. This branch of permutation will not be used.");
                                    continueEvalL1 = false;
                                }

                                if ((continueEvalL1) && (greenBand1 > greenBand2))
                                {
                                    CreateDebugMessage("The first green band is greater than the second. This branch of permutation will not be used.");
                                    continueEvalL1 = false;
                                }

                                if ((continueEvalL1) && (redBand1 == greenBand1) && (redBand2 == greenBand2))
                                {
                                    CreateDebugMessage("The red and green channels of this calculated three-channel combination are identical. " +
                                        "This branch of permutation will not be used.");
                                    continueEvalL1 = false;
                                }

                                if (continueEvalL1)
                                {
                                    for (int blueBand1 = 0; blueBand1 < bands.Length; blueBand1++)
                                    {
                                        CreateDebugMessage("\t\t\t\t\tBlue band 1: '" + bands[blueBand1].BandName + "'.");
                                        for (int blueBand2 = 1; blueBand2 < bands.Length; blueBand2++)
                                        {
                                            CreateDebugMessage("\t\t\t\t\t\tBlue band 2: '" + bands[blueBand2].BandName + "'.");
                                            bool continueEvalL2 = true;

                                            if (blueBand1 == blueBand2)
                                            {
                                                CreateDebugMessage("The two blue bands are identical. This branch of permutation will not be used.");
                                                continueEvalL2 = false;
                                            }

                                            if ((continueEvalL2) && (blueBand1 > blueBand2))
                                            {
                                                CreateDebugMessage("The first blue band is greater than the second. This branch of permutation will not be used.");
                                                continueEvalL2 = false;
                                            }

                                            if ((continueEvalL2) && (redBand1 == blueBand1) && (redBand2 == blueBand2))
                                            {
                                                CreateDebugMessage("The red and blue channels of this calculated three-channel combination are identical. " +
                                                    "This branch of permutation will not be used.");
                                                continueEvalL2 = false;
                                            }

                                            if ((continueEvalL2) && (greenBand1 == blueBand1) && (greenBand2 == blueBand2))
                                            {
                                                CreateDebugMessage("The green and blue channels of this calculated three-channel combination are identical. " +
                                                    "This branch of permutation will not be used.");
                                                continueEvalL2 = false;
                                            }

                                            if ((continueEvalL2))
                                            {
                                                bool includeImage = false;

                                                if ((combineConfig.GenerateThreeChannelSimilarMotion) && (IsSimilarMotion(redBand1, redBand2, greenBand1, greenBand2, blueBand1, blueBand2)))
                                                {
                                                    CreateDebugMessage("This is a case of similar motion, and the option to generate similar motion combinations is enabled. " +
                                                        "This variation will be included.");
                                                    includeImage = true;
                                                }

                                                if ((!includeImage) && (combineConfig.GenerateThreeChannelObliqueMotion) && (IsObliqueMotion(redBand1, redBand2, greenBand1, greenBand2, blueBand1, blueBand2)))
                                                {
                                                    CreateDebugMessage("This is a case of oblique motion, and the option to generate oblique motion combinations is enabled. " +
                                                        "This variation will be included.");
                                                    includeImage = true;
                                                }

                                                if ((!includeImage) && (combineConfig.GenerateThreeChannelOtherPermutations))
                                                {
                                                    CreateDebugMessage("This combination does not conform to a known pattern, but the option to generate other permutations is enabled. " +
                                                        "This variation will be included.");
                                                    includeImage = true;
                                                }

                                                if (includeImage)
                                                {
                                                    ThreeChannelImage tci = GetCalculatedThreeChanelImage(bands, sourceChannels, combineConfig, redBand1, redBand2, greenBand1, greenBand2, blueBand1, blueBand2);

                                                    if (tci == null)
                                                    {
                                                        CreateDebugMessage("Unable to obtain the channels for this image permutation.");
                                                    }
                                                    else
                                                    {
                                                        imageList.Add(tci);
                                                    }

                                                    if (combineConfig.GenerateThreeChannelReciprocal)
                                                    {
                                                        CreateDebugMessage("The option to generate reciprocal images is enabled. Attempting to locate the channels for the reciprocal of the current image.");
                                                        ThreeChannelImage tcir = GetCalculatedThreeChanelImage(bands, sourceChannels, combineConfig, redBand2, redBand1, greenBand2, greenBand1, blueBand2, blueBand1);
                                                        if (tcir == null)
                                                        {
                                                            CreateDebugMessage("Unable to obtain the channels for the reciprocal of this image permutation.");
                                                        }
                                                        else
                                                        {
                                                            imageList.Add(tcir);
                                                        }
                                                    }
                                                }

                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            CreateDebugMessage("End calculated three channel image permutations - method '" + combineConfig.CombineMethod.ToString() +
                "'.");

            ThreeChannelImage tcProto = new ThreeChannelImage();
            return (ThreeChannelImage[])imageList.ToArray(tcProto.GetType());
        }

        protected ThreeChannelImage GetCalculatedThreeChanelImage(SpectralBand[] bands, GreyscaleChannel[] sourceChannels, CalculatedChannelCombinationConfig combineConfig, int redBand1, int redBand2, int greenBand1, int greenBand2, int blueBand1, int blueBand2)
        {
            GreyscaleChannel[] fcic = new GreyscaleChannel[3];
            bool continueProcessing = true;

            CreateDebugMessage("Attempting to locate the channels for the image combination:");
            CreateDebugMessage("\tRed channel band 1: '" + bands[redBand1].BandName + "'.");
            CreateDebugMessage("\tRed channel band 2: '" + bands[redBand2].BandName + "'.");
            CreateDebugMessage("\tGreen channel band 1: '" + bands[greenBand1].BandName + "'.");
            CreateDebugMessage("\tGreen channel band 2: '" + bands[greenBand2].BandName + "'.");
            CreateDebugMessage("\tBlue channel band 1: '" + bands[blueBand1].BandName + "'.");
            CreateDebugMessage("\tBlue channel band 2: '" + bands[blueBand2].BandName + "'.");

            fcic[0] = GetCalculatedChannel(sourceChannels, bands[redBand1].BandName, bands[redBand2].BandName);
            if (fcic[0] == null)
            {
                CreateDebugMessage("Unable to locate the red channel for this image in the available greyscale channels.");
                continueProcessing = false;
            }

            if (continueProcessing)
            {
                fcic[1] = GetCalculatedChannel(sourceChannels, bands[greenBand1].BandName, bands[greenBand2].BandName);
                if (fcic[1] == null)
                {
                    CreateDebugMessage("Unable to locate the green channel for this image in the available greyscale channels.");
                    continueProcessing = false;
                }
            }

            if (continueProcessing)
            {
                fcic[2] = GetCalculatedChannel(sourceChannels, bands[blueBand1].BandName, bands[blueBand2].BandName);
                if (fcic[2] == null)
                {
                    CreateDebugMessage("Unable to locate the blue channel for this image in the available greyscale channels.");
                    continueProcessing = false;
                }
            }

            if (continueProcessing)
            {
                ThreeChannelImage fci = new ThreeChannelImage(fcic);

                if ((fci.HasMutuallyExclusiveBands()))
                {
                    CreateDebugMessage("# Debug: image '" + fci.ImageName + "' contains mutually exclusive bands and will not be included.");
                    continueProcessing = false;
                }
                //if ((continueProcessing) && (fci.ContainsDisallowedChannels()))
                //{
                //    CreateDebugMessage("# Debug: image '" + fci.ImageName + "' contains bands which are not allowed in three-channel images. It will not be included.");
                //    continueProcessing = false;
                //}

                if (continueProcessing)
                {
                    fci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                    //fci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                    fci.OutputFormats = OutputConfig.OutputFormats;
                    fci.UseInStandardLumaColourImages = combineConfig.UseThreeChannelPermuatationsInLumaColourImages;
                    fci.ParentCube = this;
                    fci.NamePrefix = "FC-3C-Calculated-";
                    return fci;
                }
            }

            return null;
        }

        protected bool IsSimilarMotion(int redBand1, int redBand2, int greenBand1, int greenBand2, int blueBand1, int blueBand2)
        {
            bool result = true;

            // technically, the "is greater than" portion of these checks aren't a violation of similar motion in the context of musical theory, but including these variations
            // doesn't make much sense from an imaging perspective, and generates a very high number of functionally-equivalent images
            if (redBand1 >= greenBand1)
            {
                CreateDebugMessage("The first red band is greater than or equal to the first green band. This is not a case of similar motion.");
                result = false;
            }
            if (redBand2 >= greenBand2)
            {
                CreateDebugMessage("The second red band is greater than or equal to the second green band. This is not a case of similar motion.");
                result = false;
            }

            if (redBand1 >= blueBand1)
            {
                CreateDebugMessage("The first red band is greater than or equal to the first blue band. This is not a case of similar motion.");
                result = false;
            }
            if (redBand2 >= blueBand2)
            {
                CreateDebugMessage("The second red band is greater than or equal to the second blue band. This is not a case of similar motion.");
                result = false;
            }

            if (greenBand1 >= blueBand1)
            {
                CreateDebugMessage("The first green band is greater than or equal to the first blue band. This is not a case of similar motion.");
                result = false;
            }
            if (greenBand2 >= blueBand2)
            {
                CreateDebugMessage("The second green band is greater than or equal to the second blue band. This is not a case of similar motion.");
                result = false;
            }

            return result;
        }

        protected bool IsObliqueMotion(int redBand1, int redBand2, int greenBand1, int greenBand2, int blueBand1, int blueBand2)
        {
            bool result = true;
            bool foundInitialMatch = false;
            if ((redBand1 == greenBand1) && (greenBand1 == blueBand1))
            {
                foundInitialMatch = true;
                if ((redBand2 == greenBand2) || (redBand2 == blueBand2) || (greenBand2 == blueBand2))
                {
                    CreateDebugMessage("The first bands of each channel match, but one or more of the second bands of each channel also match. This is not a case of oblique motion.");
                    result = false;
                }
                // technically, these next three checks aren't a violation of oblique motion in the context of musical theory, but including these variations
                // doesn't make much sense from an imaging perspective, and generates a very high number of functionally-equivalent images
                if (redBand2 >= greenBand2)
                {
                    CreateDebugMessage("The second red band is greater than or equal to the second green band. This is not a case of oblique motion.");
                    result = false;
                }
                if (redBand2 >= blueBand2)
                {
                    CreateDebugMessage("The second red band is greater than or equal to the second blue band. This is not a case of oblique motion.");
                    result = false;
                }
                if (greenBand2 >= blueBand2)
                {
                    CreateDebugMessage("The second green band is greater than or equal to the second blue band. This is not a case of oblique motion.");
                    result = false;
                }
            }

            if ((redBand2 == greenBand2) && (greenBand2 == blueBand2))
            {
                foundInitialMatch = true;
                if ((redBand1 == greenBand1) || (redBand1 == blueBand1) || (greenBand1 == blueBand1))
                {
                    CreateDebugMessage("The second bands of each channel match, but one or more of the first bands of each channel also match. This is not a case of oblique motion.");
                    result = false;
                }
                // technically, these next three checks aren't a violation of oblique motion in the context of musical theory, but including these variations
                // doesn't make much sense from an imaging perspective, and generates a very high number of functionally-equivalent images
                if (redBand1 >= greenBand1)
                {
                    CreateDebugMessage("The second red band is greater than or equal to the second green band. This is not a case of oblique motion.");
                    result = false;
                }
                if (redBand1 >= blueBand1)
                {
                    CreateDebugMessage("The second red band is greater than or equal to the second blue band. This is not a case of oblique motion.");
                    result = false;
                }
                if (greenBand1 >= blueBand1)
                {
                    CreateDebugMessage("The second green band is greater than or equal to the second blue band. This is not a case of oblique motion.");
                    result = false;
                }
            }

            if (!foundInitialMatch)
            {
                result = false;
            }

            //if (redBand1 != greenBand1)
            //{
            //    CreateDebugMessage("The first red band is not equal to the first green band. This is not a case of oblique motion.");
            //    result = false;
            //}
            //if (redBand1 != blueBand1)
            //{
            //    CreateDebugMessage("The first red band is not equal to the first blue band. This is not a case of oblique motion.");
            //    result = false;
            //}



            return result;
        }

        protected GreyscaleChannel GetCalculatedChannel(GreyscaleChannel[] sourceChannels, string band1Name, string band2Name)
        {
            foreach (GreyscaleChannel chan in sourceChannels)
            {
                if ((chan.SpectralBands != null) && (chan.SpectralBands.Length == 2) && (chan.SpectralBands[0].BandName == band1Name) &&
                     (chan.SpectralBands[1].BandName == band2Name))
                {
                    return chan;
                }
            }
            CreateDebugMessage("The specified GreyscaleChannel array does not contain a channel whose first band is named '" +
                band1Name + "' and whose second band is named '" + band2Name + "'.");
            return null;
        }


        public ThreeChannelImage[] GetThreeChannelImages(GreyscaleChannel[] channels)
        {
            ArrayList imageList = new ArrayList();

            SpectralBand[] bands = GetSortedBands();

            //GreyscaleChannel[] channels = GetChannelPermutations();

            //int cubeMinBandNumber = GetMinimumBandNumber();
            int cubeMaxBandNumber = GetMaximumBandNumber();

            // generate three-channel images
            if (!ProcessingConfig.GenerateCustomThreeChannelImages)
            {
                CreateDebugMessage("The GenerateCustomThreeChannelImages option is disabled. Skipping creation of any custom three-channel images.");
            }
            else
            {
                CreateDebugMessage("The GenerateCustomThreeChannelImages option is enabled.");
                foreach (CustomThreeChannelImageConfig ctc in ProcessingConfig.CustomThreeChannelImageConfigurations)
                {
                    ctc.GetChannelsFromImageCube(this);
                    if (ctc.Generate)
                    {
                        GreyscaleChannel[] ctcChannels = new GreyscaleChannel[]
                        {
                            ctc.RedChannel,
                            ctc.GreenChannel,
                            ctc.BlueChannel
                        };
                        ThreeChannelImage ctci = new ThreeChannelImage(ctcChannels);
                        if (ctc.UserDefinedName != "")
                        {
                            ctci.UserDefinedName = ctc.UserDefinedName;
                        }
                        //ctci.OutputNormalization = ctc.OutputNormalization;
                        ctci.InputTransformations = ctc.InputTransformations;
                        ctci.OutputTransformations = ctc.OutputTransformations;
                        ctci.IncludeInSecondOrderImages = ctc.IncludeInSecondOrderImages;

                        ctci.IsCustomImage = true;

                        ctci.OutputFormats = OutputConfig.OutputFormats;
                        imageList.Add(ctci);
                    }
                }
            }


            if (!ProcessingConfig.GenerateStandardThreeChannelImages)
            {
                CreateDebugMessage("The GenerateStandardThreeChannelImages option is disabled. Skipping creation of any standard three-channel images.");
            }
            else
            {
                CreateDebugMessage("The GenerateStandardThreeChannelImages option is enabled.");
                for (int redChannelI = 0; redChannelI < channels.Length; redChannelI++)
                {
                    GreyscaleChannel redChannel = channels[redChannelI];
                    //int minBand = channel.GetMinimumBandNumber();
                    if (redChannel.UseAsThreeChannelChannel)
                    {
                        int maxBandRed = redChannel.GetMaximumBandNumber();
                        //int minDiff = minBand - cubeMinBandNumber;
                        int maxDiffRed = cubeMaxBandNumber - maxBandRed;
                        // there must be at least two channels after the current channel
                        // (to provide the data which will be mapped to green and blue)
                        if (maxDiffRed >= 2)
                        {
                            for (int greenChannelI = 0; greenChannelI < channels.Length; greenChannelI++)
                            {
                                GreyscaleChannel greenChannel = channels[greenChannelI];
                                if (greenChannel.UseAsThreeChannelChannel)
                                {
                                    int minBandGreen = greenChannel.GetMinimumBandNumber();
                                    if (minBandGreen > maxBandRed)
                                    {
                                        int maxBandGreen = greenChannel.GetMaximumBandNumber();
                                        int maxDiffGreen = cubeMaxBandNumber - maxBandGreen;
                                        // there must be at least one channel after the current channel
                                        // (to provide the data which will be mapped to blue)
                                        if (maxDiffGreen >= 1)
                                        {
                                            for (int blueChannelI = 0; blueChannelI < channels.Length; blueChannelI++)
                                            {
                                                GreyscaleChannel blueChannel = channels[blueChannelI];
                                                if (blueChannel.UseAsThreeChannelChannel)
                                                {
                                                    int minBandBlue = blueChannel.GetMinimumBandNumber();
                                                    if (minBandBlue > maxBandGreen)
                                                    {
                                                        GreyscaleChannel[] fcic = new GreyscaleChannel[3];
                                                        fcic[0] = redChannel;
                                                        fcic[1] = greenChannel;
                                                        fcic[2] = blueChannel;
                                                        ThreeChannelImage fci = new ThreeChannelImage(fcic);
                                                        if (OutputConfig.RenameRGBFiles)
                                                        {
                                                            fci.RemovePrefixIfRGB(InputConfig.RedBandName, InputConfig.GreenBandName, InputConfig.BlueBandName);
                                                        }
                                                        bool includeVariation = true;
                                                        if (fci.HasMutuallyExclusiveBands())
                                                        {
                                                            CreateDebugMessage("# Debug: image '" + fci.ImageName + "' contains mutually exclusive bands and will not be included.");
                                                            includeVariation = false;
                                                        }
                                                        if (fci.ContainsDisallowedChannels())
                                                        {
                                                            CreateDebugMessage("# Debug: image '" + fci.ImageName + "' contains bands which are not allowed in three-channel images. It will not be included.");
                                                            includeVariation = false;
                                                        }

                                                        if (includeVariation)
                                                        {
                                                            fci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                                                            //fci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                                                            fci.OutputFormats = OutputConfig.OutputFormats;
                                                            imageList.Add(fci);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // statistical permutations
            if (!ProcessingConfig.GenerateStatisticalImageThreeChannelPermutations)
            {
                CreateDebugMessage("The GenerateStatisticalImageThreeChannelPermutations option is disabled. Skipping creation of any statistical three-channel images.");
            }
            else
            {
                CreateDebugMessage("The GenerateStatisticalImageThreeChannelPermutations option is enabled.");
                foreach (GreyscaleChannelSet cgsc in _StatisticalCubeSets)
                {
                    if (cgsc.UseAsThreeChannelChannels)
                    {
                        foreach (ThreeChannelImage tci in GetStatisticalThreeChannelImages(cgsc.Channels))
                        {
                            imageList.Add(tci);
                        }
                    }
                }
            }


            if (!ProcessingConfig.GenerateCalculatedThreeChannelImages)
            {
                CreateDebugMessage("The GenerateCalculatedThreeChannelImages option is disabled. Skipping creation of any calculated three-channel images.");
            }
            else
            {
                CreateDebugMessage("The GenerateCalculatedThreeChannelImages option is enabled.");
                // calculated permutations
                foreach (CalculatedChannelSet ccSet in _CalculatedChannels)
                {
                    foreach (ThreeChannelImage tci in GetCalculatedThreeChannelImages(ccSet.Channels, ccSet.Config))
                    {
                        imageList.Add(tci);
                    }
                }
            }

            ThreeChannelImage tcProto = new ThreeChannelImage();
            return (ThreeChannelImage[])imageList.ToArray(tcProto.GetType());
        }

        public ColourGradientImage[] GetColourGradientImages()
        {
            ArrayList imageList = new ArrayList();

            foreach (TintMapping gMap in ProcessingConfig.GradientMappings)
            {
                string channelName = "";
                bool foundChannel = true;
                try
                {
                    channelName = gMap.Channel.GetChannelName(this);
                }
                catch (Exception ex)
                {
                    CreateDebugMessage("The band or channel referenced for the gradient-mapped image ('" + gMap.Channel.ReferenceName +
                        "') does not appear to exist in the image cube. This mapping will not be used.");
                    foundChannel = false;
                }
                if ((foundChannel == true) && (channelName == ""))
                {
                    CreateDebugMessage("Error: an empty string was passed as a reference to a channel as part of a gradient mapping. This mapping will be ignored.");
                    foundChannel = false;
                }

                bool foundGradient = true;
                //string gradientName = "";
                string gradientName = gMap.TintName;
                //if ((gMap.TintNames != null) && (gMap.TintNames.Length != 0) && (gMap.TintNames[0] != null))
                //{
                //    gradientName = gMap.TintNames[0];
                //}
                if (!ColourConfig.ContainsGradient(gradientName))
                {
                    foundGradient = false;
                }
                if ((foundGradient == true) && (gradientName == ""))
                {
                    CreateDebugMessage("Error: an empty string was passed as a reference to a gradient as part of a gradient mapping. This mapping will be ignored.");
                    foundGradient = false;
                }

                if (foundChannel)
                {
                    GreyscaleChannel chan = GetGreyscaleChannel(channelName);
                    ColourGradient grad = ColourConfig.GetGradient(gradientName);

                    ColourGradientImage gci = new ColourGradientImage(chan, grad);
                    gci.OutputFormats = OutputConfig.OutputFormats;
                    gci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                    //gci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                    gci.ParentCube = this;
                    imageList.Add(gci);
                }
            }

            ColourGradientImage grProto = new ColourGradientImage();
            return (ColourGradientImage[])imageList.ToArray(grProto.GetType());
        }

        public LumaColourImage[] GetTintedChannels(GreyscaleChannel[] channels)
        {
            ArrayList imageList = new ArrayList();

            //foreach (GreyscaleChannel channel in channels)
            //{
            //    TintingColour[] colours = channel.GetSortedColours();
            //    for (int colourNum = 0; colourNum < colours.Length; colourNum++)
            //    {
            //        LumaColourImage lci = new LumaColourImage(channel.ImageName, colours[colourNum].TintingColourName);
            //        //lci.OutputFormats[0].BitsPerChannel = OutputBitsPerChannel;
            //        lci.OutputFormats = OutputConfig.OutputFormats;
            //        lci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
            //        lci.IncludeInSecondOrderImages = colours[colourNum].IncludeInSecondOrderImages;
            //        lci.UseAsLuminanceChannel = lci.IncludeInSecondOrderImages;
            //        lci.UserDefinedName = channel.ImageName + String.Format("-Tinted-{0:00}", (colourNum + 1));
            //        lci.LumaChannel = channel;
            //        lci.TintingColourName = colours[colourNum].TintingColourName;
            //        imageList.Add(lci);
            //    }
            //}

            foreach (TintMapping tMap in ProcessingConfig.TintMappings)
            {
                string channelName = "";
                bool foundChannel = true;
                try
                {
                    channelName = tMap.Channel.GetChannelName(this);
                }
                catch (Exception ex)
                {
                    CreateDebugMessage("The band or channel referenced for the tinted image ('" + tMap.Channel.ReferenceName +
                        "') does not appear to exist in the image cube. This mapping will not be used.");
                    foundChannel = false;
                }
                if ((foundChannel == true) && (channelName == ""))
                {
                    CreateDebugMessage("Error: an empty string was passed as a reference to a channel as part of a tint mapping. This mapping will be ignored.");
                    foundChannel = false;
                }

                bool foundTint = true;

                if (!ColourConfig.ContainsTintingColour(tMap.TintName))
                {
                    foundTint = false;
                }
                if ((foundTint == true) && (tMap.TintName == ""))
                {
                    CreateDebugMessage("Error: an empty string was passed as a reference to a tinting colour as part of a tint mapping. This mapping will be ignored.");
                    foundTint = false;
                }

                if (foundChannel)
                {
                    GreyscaleChannel chan = GetGreyscaleChannel(channelName);
                    TintingColour tColour = ColourConfig.GetTintingColour(tMap.TintName);

                    LumaColourImage lci = new LumaColourImage(chan.ImageName, tColour.TintingColourName);

                    lci.OutputFormats = OutputConfig.OutputFormats;
                    //lci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                    lci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                    lci.ParentCube = this;

                    lci.IncludeInSecondOrderImages = tMap.IncludeInSecondOrderImages;
                    lci.UseAsLuminanceChannel = lci.IncludeInSecondOrderImages;
                    lci.UserDefinedName = "FC-Tn-" + OutputImage.WrapNameComponentIfComplex(chan.ImageName) + "-" + OutputImage.WrapNameComponentIfComplex(tColour.TintingColourName);
                    lci.LumaChannel = chan;
                    lci.TintingColour = tColour;
                    //lci.TintingColourName = tColour.TintingColourName;

                    imageList.Add(lci);
                }
            }

            LumaColourImage lcProto = new LumaColourImage();
            return (LumaColourImage[])imageList.ToArray(lcProto.GetType());
        }

        protected LumaColourImage[] GetLumaColourImages()
        {
            ArrayList imageList = new ArrayList();
            GreyscaleChannel[] channels = GetSortedGreyscaleChannels();

            if (ProcessingConfig.GenerateLumaColourImages)
            {
                // first, create the variations on the tinted channels
                if ((ProcessingConfig.GenerateLumaColourVariationsFromIndividualBands) && (ProcessingConfig.GenerateLumaColourVariationsFromTintedImages))
                {
                    foreach (LumaColourImage tintedChannel in _TintedChannels)
                    {
                        if (tintedChannel.IncludeInSecondOrderImages)
                        {
                            imageList = GetValidLumaColourVariations(imageList, channels, tintedChannel);
                        }
                    }
                }

                // variations on gradient-mapped images
                if (ProcessingConfig.GenerateLumaColourVariationsFromGradientMappedImages)
                {
                    foreach (ColourGradientImage cgi in ColourGradientImages)
                    {
                        imageList = GetValidLumaColourVariations(imageList, channels, cgi);
                    }
                }

                // next, create variations on the three-channel images
                if (ProcessingConfig.GenerateLumaColourVariationsFromThreeChannelImages)
                {
                    foreach (ThreeChannelImage threeChannel in _ThreeChannelImages)
                    {
                        if (threeChannel.IncludeInSecondOrderImages)
                        {
                            if ((!threeChannel.IsCustomImage) || (ProcessingConfig.GenerateLumaColourVariationsFromCustomThreeChannelImages))
                            {
                                imageList = GetValidLumaColourVariations(imageList, channels, threeChannel);
                            }
                        }
                    }
                }

                // if the options are enabled, generate variations from the DCS images
                if ((ProcessingConfig.GenerateDCSImagesFromIndividualChannels) && (ProcessingConfig.GenerateLumaColourVariationsFromDCSImages) && (_DCSImages.Length > 0))
                {
                    foreach (DCSImage dci in _DCSImages)
                    {
                        // third-order images - DSC colour, greyscale channel luminance
                        if (dci.UseInStandardLumaColourImages)
                        {
                            imageList = GetValidLumaColourVariations(imageList, channels, dci);
                        }
                    }
                }
            }

            // DCS-enhanced contrast
            if ((ProcessingConfig.GenerateDCSEnhancedContrast) && (_DCSImages.Length > 0))
            {
                foreach (DCSImage dci in _DCSImages)
                {
                    if (dci.DCSConfiguration.UseInDCSEnhancedContrast)
                    {
                        LumaColourImage dec = new LumaColourImage(dci.SourceImage.ImageName, dci.ImageName);
                        dec.OutputFormats = OutputConfig.OutputFormats;
                        //dec.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                        dec.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                        dec.LumaChannel = dci;
                        dec.ColourFilePath = dci.SourceImage.GetFilePathWithPrimaryExtension();
                        //dec.NamePrefix = "FC-DCS_EC-";
                        dec.UserDefinedName = "FC-DCS_EC-(" + dci.SourceImage.ImageName + ")-" + dci.GetDCSTypeName();
                        imageList.Add(dec);
                    }
                }
            }

            LumaColourImage lcProto = new LumaColourImage();
            return (LumaColourImage[])imageList.ToArray(lcProto.GetType());
        }

        protected bool IncludeLCVariation(GreyscaleChannel lumaChan, MultiChannelImage colourImage)
        {
            if (!lumaChan.UseAsLuminanceChannel)
            {
                CreateDebugMessage("The channel '" + lumaChan.ImageName + "' is configured to not be used as a luminance channel. It will not be used.");
                return false;
            }

            // reject combinations where the images contain mutually exclusive bands
            if ((!lumaChan.UsesAllSpectralBands) && (colourImage.HasMutuallyExclusiveBands(lumaChan)))
            {
                CreateDebugMessage("# Debug: luma/colour combination '" + lumaChan.ImageName + "' / '" + colourImage.ImageName + "' contains mutually exclusive bands and will not be included.");
                return false;
            }

            if (lumaChan.UsesAllSpectralBands)
            {
                if (!ProcessingConfig.GenerateLumaColourVariationsFromStatisticalChannels)
                {
                    CreateDebugMessage("The option to generate Colour/Luma variations from statistical channels is not enabled, and the " +
                        "luminance channel '" + lumaChan.ImageName + "' is a statistical channel. It will not be used.");
                    return false;
                }
            }

            if (!colourImage.IncludeInSecondOrderImages)
            {
                CreateDebugMessage("The colour image '" + colourImage.ImageName + "' has the flag to include in second-order images disabled. It will not be used.");
                return false;
            }

            // do not generate two effective copies of the R-G-B image
            if ((lumaChan.BandCombineMode == BandCombineMethod.Luminance) && (lumaChan.SpectralBands.Length == 3))
            {
                if ((colourImage.Channels.Length == 3) && (colourImage.Channels[0].SpectralBands.Length == 1) && 
                    (colourImage.Channels[1].SpectralBands.Length == 1) && (colourImage.Channels[2].SpectralBands.Length == 1))
                {
                    if ((lumaChan.SpectralBands[0].BandName == colourImage.Channels[0].SpectralBands[0].BandName) &&
                        (lumaChan.SpectralBands[1].BandName == colourImage.Channels[1].SpectralBands[0].BandName) &&
                        (lumaChan.SpectralBands[2].BandName == colourImage.Channels[2].SpectralBands[0].BandName))
                    {
                        return false;
                    }
                }
            }

            // checks against individual channels of the colour image
            foreach (GreyscaleChannel colourChan in colourImage.Channels)
            {
                if (colourChan.UsesAllSpectralBands)
                {
                    if (!ProcessingConfig.GenerateLumaColourVariationsFromStatisticalChannels)
                    {
                        CreateDebugMessage("The option to generate Colour/Luma variations from statistical channels is not enabled, and the " +
                            "colour channel '" + colourChan.ImageName + "' is a statistical channel. It will not be used.");
                        return false;
                    }
                }
                else
                {
                    if ((!ProcessingConfig.GenerateLumaColourVariationsFromCombinedChannels) && (colourChan.SpectralBands.Length > 1))
                    {
                        CreateDebugMessage("The option to generate Colour/Luma variations from combined channels is not enabled, and the " +
                            "colour channel '" + colourChan.ImageName + "' is a statistical channel. It will not be used.");
                        return false;
                    }

                    if (!lumaChan.UsesAllSpectralBands)
                    {
                        bool foundUnmatchedBand = false;
                        if (lumaChan.SpectralBands.Length != colourChan.SpectralBands.Length)
                        {
                            foundUnmatchedBand = true;
                        }
                        if ((lumaChan.SpectralBands.Length > 1) && (lumaChan.BandCombineMode != colourChan.BandCombineMode))
                        {
                            foundUnmatchedBand = true;
                        }
                        if (!foundUnmatchedBand)
                        {
                            foreach (SpectralBand colourBand in colourChan.SpectralBands)
                            {
                                foreach (SpectralBand lumaBand in lumaChan.SpectralBands)
                                {
                                    //if ((colourBand.BandName == lumaBand.BandName) && (!colourImage.IsCustomImage) && (!colourChan.IsCustomImage) && (!lumaChan.IsCustomImage))
                                    //{
                                    //    CreateDebugMessage("The channel '" + lumaChan.ImageName + "' contains the same band as the colour image '" + colourImage.ImageName + 
                                    //        "' ('" + colourBand.BandName + "'). This combination will not be used.");
                                    //    return false;
                                    //}
                                    if (colourBand.BandName != lumaBand.BandName)
                                    {
                                        foundUnmatchedBand = true;
                                    }
                                    if ((lumaChan.SpectralBands.Length == 1) && (!lumaBand.UseAsLumaChannel))
                                    {
                                        CreateDebugMessage("The channel '" + lumaChan.ImageName + "' is based on a single band ('" + lumaBand.BandName +
                                            "') which is not configured for use in a luminance channel. This combination will not be used.");
                                        return false;
                                    }
                                }
                            }
                        }
                        if (!foundUnmatchedBand)
                        {
                            CreateDebugMessage("The luminance image '" + lumaChan.ImageName + "' appears to be identical to one of the channels of the colour image '" + 
                                colourImage.ImageName + "'. This combination will not be used.");
                            return false;
                        }
                    }
                }

                if (!ProcessingConfig.GenerateLumaColourVariationsFromIndividualBands)
                {
                    if (lumaChan.SpectralBands.Length > 1)
                    {
                        CreateDebugMessage("The option to generate Colour/Luma variations from individual bands is not enabled, and the " +
                            "luma channel '" + lumaChan.ImageName + "' contains only a single band. It will not be used.");
                        return false;
                    }

                    if (colourChan.SpectralBands.Length == 1)
                    {
                        CreateDebugMessage("The option to generate Colour/Luma variations from individual bands is not enabled, and the " +
                            "colour channel '" + colourChan.ImageName + "' contains only a single band. It will not be used.");
                        return false;
                    }
                }

            }

            return true;
        }

        protected ArrayList GetValidLumaColourVariations(ArrayList baseList, GreyscaleChannel[] lumaChannels, MultiChannelImage colourImage)
        {
            foreach (GreyscaleChannel lumaChan in lumaChannels)
            {
                if (IncludeLCVariation(lumaChan, colourImage))
                {
                    LumaColourImage lci = new LumaColourImage(lumaChan.ImageName, colourImage.ImageName);
                    lci.OutputFormats = OutputConfig.OutputFormats;
                    //lci.OutputNormalization = ProcessingConfig.ColourImageOverallOutputNormalization;
                    lci.OutputTransformations = ProcessingConfig.ColourImageOutputTransformations;
                    lci.LumaChannel = lumaChan;
                    lci.ColourFilePath = colourImage.GetFilePathWithPrimaryExtension();
                    foreach (GreyscaleChannel colourChan in colourImage.Channels)
                    {
                        lci.AddChannel(colourChan);
                    }
                    baseList.Add(lci);
                }
            }
            return baseList;
        }

        protected DCSImage[] GetDCSImages()
        {
            ArrayList imageList = new ArrayList();

            foreach (ThreeChannelImage threeChannel in _ThreeChannelImages)
            {
                bool hasCombinationChannels = false;
                bool isAllChannelsImage = true;
                int spectralBandCount = 0;
                foreach (GreyscaleChannel greyChan in threeChannel.Channels)
                {
                    if (greyChan.SpectralBands.Length > 1)
                    {
                        hasCombinationChannels = true;
                    }
                    if (greyChan.SpectralBands.Length != _SpectralBands.Count)
                    {
                        isAllChannelsImage = false;
                    }
                }
                if ((hasCombinationChannels && ProcessingConfig.GenerateDCSImagesFromCombinedChannels)
                    ||
                    ((!hasCombinationChannels) && ProcessingConfig.GenerateDCSImagesFromIndividualChannels)
                    ||
                    ((!ProcessingConfig.GenerateDCSImagesFromCombinedChannels) && (ProcessingConfig.CustomThreeChannelImageConfigurations.Length > 0) && isAllChannelsImage))
                {
                    foreach (DCSParameterSet dcsType in ProcessingConfig.DCSTypes)
                    {
                        DCSImage dci = new DCSImage(threeChannel, dcsType);
                        dci.OutputFormats = OutputConfig.OutputFormats;

                        //bool includeVariation = true;

                        //// this should never happen
                        //if ((includeVariation) && (dci.HasMutuallyExclusiveBands()))
                        //{
                        //    CreateDebugMessage("# Debug: DCS image '" + dci.ImageName + "' contains mutually exclusive bands and will not be included.");
                        //    includeVariation = false;
                        //}
                        //// this should never happen
                        //if (dci.ContainsDisallowedChannels())
                        //{
                        //    CreateDebugMessage("# Debug: DCS image '" + dci.ImageName + "' contains bands which are not allowed in three-channel images. It will not be included.");
                        //    includeVariation = false;
                        //}

                        bool addPermutation = true;

                        if ((threeChannel.IsCustomImage) && (!ProcessingConfig.GenerateDCSImagesFromCustomThreeChannelImages))
                        {
                            addPermutation = false;
                        }

                        if (addPermutation)
                        {
                            imageList.Add(dci);
                        }
                    }
                }
            }

            MultiChannelImage mciProto = new MultiChannelImage();
            mciProto.UserDefinedName = "Prototype";
            DCSImage dcsProto = new DCSImage(mciProto);
            return (DCSImage[])imageList.ToArray(dcsProto.GetType());
        }

        public void CreateDebugMessage(string message)
        {
            //if (EnableDebugLogging)
            //{
            //    Console.WriteLine("DEBUG: " + message);
            //}
            if (_ParentEngine != null)
            {
                _ParentEngine.GenerateScriptDebugMessage(-1, message);
            }
        }

        public void CreateInformationalMessage(string message)
        {
            _ParentEngine.GenerateScriptInformationalMessage(-1, message);
        }

    }
}
