// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace BenLincoln.DaVinci
{
    public enum ImageFormat
    {
        TIFF,
        PNG,
        JPEG
    }

    public enum BitsPerChannel
    {
        EightByte,
        SixteenShort,
        ThirtyTwoFloat,
        SixtyFourDouble
    }

    public enum ScalingMode
    {
        None,
        LongestEdgeAbsolute,
        Percentage
    }

    public class OutputParameterSet
    {
        protected ImageFormat _FileFormat;
        protected BitsPerChannel _BitsPerChannel;
        protected ScalingMode _ScalingMode;
        protected float _ScalingValue;
        protected bool RegenerateExistingImages;

        #region Properties

        public ImageFormat FileFormat
        {
            get
            {
                return _FileFormat;
            }
            set
            {
                _FileFormat = value;
            }
        }

        public BitsPerChannel BitsPerChannel
        {
            get
            {
                return _BitsPerChannel;
            }
            set
            {
                _BitsPerChannel = value;
            }
        }

        public ScalingMode ScalingMode
        {
            get
            {
                return _ScalingMode;
            }
            set
            {
                _ScalingMode = value;
            }
        }

        public float ScalingValue
        {
            get
            {
                return _ScalingValue;
            }
            set
            {
                _ScalingValue = value;
            }
        }

        public bool RegenerateExistingFiles
        {
            get
            {
                return RegenerateExistingImages;
            }
            set
            {
                RegenerateExistingImages = value;
            }
        }

        #endregion

        public OutputParameterSet()
        {
            _FileFormat = ImageFormat.TIFF;
            _BitsPerChannel = BitsPerChannel.SixteenShort;
            _ScalingMode = ScalingMode.None;
            _ScalingValue = 1.0f;
        }

        public static string ImageFormatToDaVinciFileTypeString(ImageFormat imageFormat)
        {
            switch (imageFormat)
            {
                case ImageFormat.TIFF:
                    return "tiff";
                    break;
                case ImageFormat.PNG:
                    return "png";
                    break;
                case ImageFormat.JPEG:
                    return "jpeg";
                    break;
            }
            throw new ArgumentException("The image format '" + imageFormat.ToString() + "' does not identify a recognized output file type.");
        }

        public static string ImageFormatToFileExtension(ImageFormat imageFormat)
        {
            switch (imageFormat)
            {
                case ImageFormat.TIFF:
                    return ".TIF";
                    break;
                case ImageFormat.PNG:
                    return ".PNG";
                    break;
                case ImageFormat.JPEG:
                    return ".JPG";
                    break;
            }
            throw new ArgumentException("The image format '" + imageFormat.ToString() + "' does not identify a recognized output file type.");
        }

        public static ImageFormat ImageFormatFromString(string formatString)
        {
            string fString = formatString.Trim().ToLower();
            switch (fString)
            {
                case "tiff":
                    return ImageFormat.TIFF;
                    break;
                case "jpg":
                    return ImageFormat.JPEG;
                    break;
                case "jpeg":
                    return ImageFormat.JPEG;
                    break;
                case "png":
                    return ImageFormat.PNG;
                    break;
            }
            throw new ArgumentException("The string '" + formatString + "' does not identify a recognized output file format.");
        }

        public static BitsPerChannel BitsPerChannelFromString(string bitsPerChannelString)
        {
            string bpcString = bitsPerChannelString.Trim().ToLower();
            switch (bpcString)
            {
                case "8":
                    return BitsPerChannel.EightByte;
                    break;
                case "16":
                    return BitsPerChannel.SixteenShort;
                    break;
                case "32":
                    return BitsPerChannel.ThirtyTwoFloat;
                    break;
                case "64":
                    return BitsPerChannel.SixtyFourDouble;
                    break;
            }
            throw new ArgumentException("The string '" + bitsPerChannelString + "' does not identify a recognized number of bits per channel.");
        }

        public static ScalingMode ScalingModeFromString(string scalingModeString)
        {
            string smString = scalingModeString.Trim().ToLower();
            switch (smString)
            {
                case "none":
                    return ScalingMode.None;
                    break;
                case "longestedgeabsolute":
                    return ScalingMode.LongestEdgeAbsolute;
                    break;
                case "percentage":
                    return ScalingMode.Percentage;
                    break;
            }
            throw new ArgumentException("The string '" + scalingModeString + "' does not identify a recognized scaling mode.");
        }


    }
}
