// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    public enum ScriptReadState
    {
        Default,
        Image,
        MultiImage,
        OtherContent
    }

    public struct ScriptQuantum
    {
        public bool Executed;
        public string ScriptText;
        public string SectionName;
        public string ImageName;
        public int FirstImageNumber;
        public int LastImageNumber;
        public int TotalImages;
        public string[] RequiredImages;
        public string[] OutputImages;
    }

    public class TMSBDVScriptReader
    {
        protected string _ScriptPath;

        protected TMSBEngine _ParentEngine;

        protected bool _ScriptWasGeneratedByTMSB;

        // Some versions of DaVinci do not execute the last line of a script file.
        // Therefore, it's important to have a last line which can safely be ignored.
        protected const string _SplitScriptEndPadding = "# This line intentionally left blank.";

        public const string LogMarkerText = "TMSB Log Parse Marker";
        public const string LogEndMarkerText = "TMSB Log Parse Marker - End";

        protected const bool _WriteWarningsToSplitScripts = true;

        #region Properties

        public string ScriptPath
        {
            get
            {
                return _ScriptPath;
            }
            set
            {
                _ScriptPath = value;
            }
        }

        #endregion

        public TMSBDVScriptReader(TMSBEngine parentEngine, bool scriptWasGeneratedByTMSB)
        {
            _ParentEngine = parentEngine;
            _ScriptWasGeneratedByTMSB = scriptWasGeneratedByTMSB;
            _ScriptPath = "";
        }

        public TMSBDVScriptReader(TMSBEngine parentEngine, bool scriptWasGeneratedByTMSB, string scriptPath)
            : this(parentEngine, scriptWasGeneratedByTMSB)
        {
            _ScriptPath = scriptPath;
        }

        ~TMSBDVScriptReader()
        {
            _ParentEngine = null;
        }

        public ScriptQuantum[] GetScriptQuanta(ref int maximumImageNumberInScript)
        {
            ArrayList quantaList = new ArrayList();

            try
            {
                FileStream iStream = new FileStream(ScriptPath, FileMode.Open, FileAccess.Read);
                StreamReader iReader = new StreamReader(iStream);

                string currentSectionName = "Initializing";
                string currentImageNumber = "0";
                string newImageNumber = "0";
                string currentImageName = "Initializing";
                string newSectionName = currentSectionName;
                string newImageName = currentImageName;
                bool foundAtLeastOneSectionHeader = false;
                bool foundSectionHeader = false;
                bool foundEndMarker = false;
                StringBuilder qBuilder = new StringBuilder();
                ScriptReadState readState = ScriptReadState.Default;
                ScriptReadState newReadState = ScriptReadState.Default;

                Regex sectionNameRex = new Regex(@"^# Section '[0-9A-Za-z _\-/\\]{1,2048}'$");
                Regex imageBeginRex = new Regex(@"^# Begin image number [0-9]{1,20}");
                Regex multiImageBeginRex = new Regex(@"^# Begin image numbers [0-9]{1,20} - [0-9]{1,20}");
                Regex imageEndRex = new Regex(@"^# End image number [0-9]{1,20}");
                Regex multiImageEndRex = new Regex(@"^# End image numbers [0-9]{1,20} - [0-9]{1,20}");

                //Regex imageNameRex = new Regex(@"^# ((Create the '[0-9A-Za-z _\-/\\()]{1,2048})|(Begin image[^']{1,512}'[^']{1,2048}'))'");
                Regex imageNameRex = new Regex(@"^# ((Create the)|(Begin image))[^']{1,512}'[^']{1,2048}'");
                Regex refinedImageNameRex = new Regex(@"'[^']{1,2048}'");

                Regex endMarkerRex = new Regex("Execution Complete");

                Regex imageNumberRex = new Regex(@" [0-9]{1,20} ");
                Regex multiImageNumberRex = new Regex(@"[0-9]{1,20} - [0-9]{1,20}");

                Regex requiredImageRex = new Regex(@"read\(""[^""]{1,2048}""");
                Regex outputImageRex = new Regex(@"write\([^,]{1,512},[ ]{0,16}""[^""]{1,2048}""");
                Regex referencedFileNameRex = new Regex(@"""[^""]{1,2048}""");

                Regex[] allRexes = new Regex[] { sectionNameRex, imageBeginRex, 
                    multiImageBeginRex, imageEndRex, multiImageEndRex, imageNameRex, refinedImageNameRex, 
                    endMarkerRex, imageNumberRex,
                    multiImageNumberRex, requiredImageRex, outputImageRex, referencedFileNameRex };

                //for (int rN = 0; rN < allRexes.Length; rN++)
                //{
                //    allRexes[rN]. = RegexOptions.Multiline;
                //}

                ArrayList requiredImageList = new ArrayList();
                ArrayList outputImageList = new ArrayList();

                bool foundImageName = false;

                while (!iReader.EndOfStream)
                {
                    string currentLine = iReader.ReadLine();

                    // this is here in order to make sure the script being read was fully generated
                    // as opposed to the generation process failing in the middle
                    if (endMarkerRex.IsMatch(currentLine))
                    {
                        foundEndMarker = true;
                    }

                    if ((!foundSectionHeader) && (sectionNameRex.IsMatch(currentLine)))
                    {
                        MatchCollection nameMatches = refinedImageNameRex.Matches(currentLine);
                        if (nameMatches.Count > 0)
                        {
                            string secName = nameMatches[0].Value.Replace("'", "");
                            if (secName.Trim() != "")
                            {
                                newSectionName = secName;
                                foundSectionHeader = true;
                                foundAtLeastOneSectionHeader = true;
                            }
                        }
                    }

                    if ((currentImageName == "Unknown") && (currentSectionName != "Unknown"))
                    {
                        currentImageName = currentSectionName;
                    }

                    if (imageBeginRex.IsMatch(currentLine))
                    {
                        string imageNum = GetImageNumberFromLine(currentLine, ref qBuilder, imageNumberRex);
                        if (imageNum != "Unknown")
                        {
                            newImageNumber = imageNum;
                        }
                        newReadState = ScriptReadState.Image;
                        foundImageName = false;
                    }

                    if (multiImageBeginRex.IsMatch(currentLine))
                    {
                        string imageNum = GetImageNumberFromLine(currentLine, ref qBuilder, multiImageNumberRex);
                        if (imageNum != "Unknown")
                        {
                            newImageNumber = imageNum;
                        }
                        newReadState = ScriptReadState.MultiImage;
                        foundImageName = false;
                    }

                    if ((!foundImageName) && (imageNameRex.IsMatch(currentLine)))
                    {
                        MatchCollection nameMatches = refinedImageNameRex.Matches(currentLine);
                        if (nameMatches.Count > 0)
                        {
                            string imageName = nameMatches[0].Value.Replace("'", "");
                            if (imageName.Trim() != "")
                            {
                                newImageName = imageName;
                                foundImageName = true;
                            }
                        }
                    }

                    if (imageEndRex.IsMatch(currentLine))
                    {
                        string imageNum = GetImageNumberFromLine(currentLine, ref qBuilder, imageNumberRex);
                        if (imageNum != "Unknown")
                        {
                            if (imageNum != currentImageNumber)
                            {
                                qBuilder.AppendLine("# Warning: the beginning of the script for image number(s) " +
                                    currentImageNumber + " was detected, but the end of image number(s) " +
                                    imageNum + " was detected instead. It is very likely that a script generation error has occurred.");
                            }
                        }
                        newReadState = ScriptReadState.Default;
                    }

                    if (multiImageEndRex.IsMatch(currentLine))
                    {
                        string imageNum = GetImageNumberFromLine(currentLine, ref qBuilder, multiImageNumberRex);
                        if (imageNum != "Unknown")
                        {
                            if (imageNum != currentImageNumber)
                            {
                                qBuilder.AppendLine("# Warning: the beginning of the script for image number(s) " +
                                    currentImageNumber + " was detected, but the end of image number(s) " +
                                    imageNum + " was detected instead. It is very likely that a script generation error has occurred.");
                            }
                        }
                        newReadState = ScriptReadState.Default;
                    }

                    UpdateReferencedImageList(ref requiredImageList, requiredImageRex, referencedFileNameRex, currentLine);

                    UpdateReferencedImageList(ref outputImageList, outputImageRex, referencedFileNameRex, currentLine);

                    bool completeQuantum = false;
                    if ((readState != newReadState) && (newReadState != ScriptReadState.Default))
                    {
                        completeQuantum = true;
                    }

                    readState = newReadState;

                    // store the current script as a standalone entity and start a new one
                    if (completeQuantum)
                    {
                        AddCurrentQuantum(ref quantaList, ref qBuilder, currentSectionName, currentImageNumber, currentImageName, ref requiredImageList,
                            ref outputImageList, ref maximumImageNumberInScript);
                        //foundImageName = false;
                        currentSectionName = newSectionName;
                        currentImageNumber = newImageNumber;
                        currentImageName = newImageName;
                        foundImageName = false;
                        foundSectionHeader = false;
                        requiredImageList = new ArrayList();
                        outputImageList = new ArrayList();
                    }

                    //currentImageNumber = newImageNumber;
                    //currentImageName = newImageName;

                    qBuilder.AppendLine(currentLine);
                }

                iReader.Close();
                iStream.Close();

                // add any remaining text from the buffer
                if (qBuilder.Length > 0)
                {
                    AddCurrentQuantum(ref quantaList, ref qBuilder, currentSectionName, currentImageNumber, currentImageName, ref requiredImageList,
                        ref outputImageList, ref maximumImageNumberInScript);
                    foundImageName = false;
                }

                // make sure at least one section header exists in the script
                // but only if the script was generated by TMSB. If it was customized, there may not be any
                // even if it's a legitimate script
                if ((_ScriptWasGeneratedByTMSB) && (!foundAtLeastOneSectionHeader))
                {
                    throw new Exception("No section headers could be located in the script file '" + ScriptPath +
                        "'. This indicates that the script was not generated correctly, and it will not be processed.");
                }

                // same for the end marker
                if ((_ScriptWasGeneratedByTMSB) && (!foundEndMarker))
                {
                    throw new Exception("An end marker could not be located in the script file '" + ScriptPath +
                        "'. This indicates that the script was not generated correctly, and it will not be processed.");
                }
            }
            catch (Exception ex)
            {
                throw new IOException("An error occurred while re-reading the script '" + ScriptPath + "' for execution. " +
                    "The error description was '" + ex.Message + "'.");
            }
            ScriptQuantum pQ = new ScriptQuantum();
            return (ScriptQuantum[])quantaList.ToArray(pQ.GetType());
        }

        protected void UpdateReferencedImageList(ref ArrayList referencedImageList, Regex lineMatchRex, Regex nameMatchRex, string currentLine)
        {
            if (lineMatchRex.IsMatch(currentLine))
            {
                MatchCollection rawMatches = lineMatchRex.Matches(currentLine);
                foreach (Match m in rawMatches)
                {
                    string riRaw = m.Value;
                    MatchCollection refinedMatches = nameMatchRex.Matches(riRaw);
                    foreach (Match rm in refinedMatches)
                    {
                        string ri = rm.Value.Replace("\"", "");
                        if (!referencedImageList.Contains(ri))
                        {
                            referencedImageList.Add(ri);
                        }
                    }
                }
            }
        }

        protected void AddCurrentQuantum(ref ArrayList quantaList, ref StringBuilder qBuilder, string currentSectionName, string currentImageNumber,
            string currentImageName, ref ArrayList requiredImageList, ref ArrayList outputImageList, ref int maximumImageNumberInScript)
        {
            int firstImageNum = 0;
            int lastImageNum = 0;
            Regex numberRex = new Regex("[0-9]{1,20}");
            MatchCollection nM = numberRex.Matches(currentImageNumber);
            if (nM.Count > 0)
            {
                if (nM.Count == 1)
                {
                    firstImageNum = int.Parse(nM[0].Value);
                    lastImageNum = firstImageNum;
                }
                else
                {
                    firstImageNum = int.Parse(nM[0].Value);
                    if (nM.Count == 2)
                    {
                        lastImageNum = int.Parse(nM[1].Value);
                    }
                    else
                    {
                        ArrayList inList = new ArrayList();
                        for (int i = 1; i < nM.Count; i++)
                        {
                            inList.Add(nM[i].Value);
                        }
                        inList.Sort();
                        lastImageNum = int.Parse(inList[inList.Count - 1].ToString());
                    }
                }
            }

            if (lastImageNum > maximumImageNumberInScript)
            {
                maximumImageNumberInScript = lastImageNum;
            }
            if (firstImageNum > maximumImageNumberInScript)
            {
                maximumImageNumberInScript = firstImageNum;
            }

            StringBuilder q2 = new StringBuilder();
            q2.Append(_ParentEngine.GetDaVinciScriptHeader());
            q2.Append(Environment.NewLine);
            q2.Append("# Section: ");
            q2.Append(currentSectionName);
            q2.Append(Environment.NewLine);
            q2.Append("# Image: ");
            q2.Append(currentImageName);
            q2.Append(Environment.NewLine);
            q2.Append("# First Image Number: ");
            q2.Append(firstImageNum.ToString());
            q2.Append(Environment.NewLine);
            q2.Append("# Last Image Number: ");
            q2.Append(lastImageNum.ToString());
            q2.Append(Environment.NewLine);

            if (requiredImageList.Count > 0)
            {
                requiredImageList.Sort();
                foreach (string rI in requiredImageList)
                {
                    q2.Append("# Requires: ");
                    q2.Append(rI);
                    q2.Append(Environment.NewLine);
                }
            }

            if (outputImageList.Count > 0)
            {
                outputImageList.Sort();
                foreach (string oI in outputImageList)
                {
                    q2.Append("# Output: ");
                    q2.Append(oI);
                    q2.Append(Environment.NewLine);
                }
            }

            q2.AppendLine("printf(\"" + LogMarkerText + "\\n\")");
            q2.Append(qBuilder.ToString());
            q2.AppendLine("printf(\"" + LogMarkerText + "\\n\")");
            q2.AppendLine("printf(\"" + LogEndMarkerText + "\\n\")");

            string currentQuantum = q2.ToString();

            // to do: figure out why I put this check here instead of just doing the work
            if (currentQuantum.Trim() != "")
            {
                ScriptQuantum newQuantum = new ScriptQuantum();
                newQuantum.Executed = false;
                newQuantum.ScriptText = currentQuantum;
                newQuantum.SectionName = currentSectionName;
                newQuantum.FirstImageNumber = firstImageNum;
                newQuantum.LastImageNumber = lastImageNum;
                newQuantum.TotalImages = lastImageNum - (Math.Max(0, (firstImageNum - 1)));
                newQuantum.ImageName = currentImageName;
                newQuantum.RequiredImages = (string[])requiredImageList.ToArray("".GetType());
                newQuantum.OutputImages = (string[])outputImageList.ToArray("".GetType());
                quantaList.Add(newQuantum);
            }

            qBuilder = new StringBuilder();
            //requiredImageList = new ArrayList();
        }

        protected string GetImageNumberFromLine(string currentLine, ref StringBuilder qBuilder, Regex imageNumberRex)
        {
            string refinedLine = currentLine;
            Regex lineRex = new Regex(@"'[^']{1,2048}'");
            refinedLine = lineRex.Replace(refinedLine, "");
            MatchCollection inMatches = imageNumberRex.Matches(refinedLine);
            string result = "Unknown";
            if (inMatches.Count == 0)
            {
                if (_WriteWarningsToSplitScripts)
                {
                    qBuilder.AppendLine("# Warning: unable to determine the image number(s) correctly from the preceeding line - no matches found.");
                }
            }
            else if (inMatches.Count > 1)
            {
                if (_WriteWarningsToSplitScripts)
                {
                    qBuilder.AppendLine(String.Format("# Warning: unable to determine the image number(s) correctly from the preceeding line - {0} matches found.", inMatches.Count));
                }
            }
            else
            {
                result = inMatches[0].Value;
            }
            return result;
        }

        public static string GetQuantumDescriptionString(ScriptQuantum quantum)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("Section Name: '");
            builder.Append(quantum.SectionName);
            builder.Append("'");
            builder.Append(", Image Name: '");
            builder.Append(quantum.ImageName);
            builder.Append("', First Image Number: ");
            builder.Append(quantum.FirstImageNumber.ToString());
            builder.Append(", Last Image Number: ");
            builder.Append(quantum.LastImageNumber.ToString());
            builder.Append(", Total Images: ");
            builder.Append(quantum.TotalImages.ToString());
            return builder.ToString();
        }
    }

}
