// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public class TMSBEngineMonitor : IDisposable
    {
        protected TMSBEngineController _Controller;

        protected OutputLevel _OutputLevel;

        protected string _CurrentSection;
        //protected TimeTracker _Tracker;
        protected TimeTracker _MetaTracker;

        protected MessageCollection _ErrorMessages;

        protected bool _Disposed;

        #region Properties

        public TMSBEngineController Controller
        {
            get
            {
                return _Controller;
            }
            set
            {
                _Controller = value;
            }
        }

        public TimeTracker MetaTracker
        {
            get
            {
                return _MetaTracker;
            }
            set
            {
                _MetaTracker = value;
            }
        }

        public MessageCollection ErrorMessages
        {
            get
            {
                return _ErrorMessages;
            }
            set
            {
                _ErrorMessages = value;
            }
        }

        #endregion

        // longest constant name evar?
        // this is the number of remaining images above which the estimated remaining time
        // will not be displayed if it is zero seconds
        protected const int _MaximumNumberOfRemainingImagesBeforeDisplayingNoRemainingTime = 1;

        public TMSBEngineMonitor(TMSBEngineController controller, OutputLevel outputLevel)
        {
            _Disposed = false;
            _Controller = controller;
            //_Tracker = new TimeTracker();
            _MetaTracker = new TimeTracker();
            //_Tracker.Start();
            _MetaTracker.Start();
            _CurrentSection = "Initializing";
            _OutputLevel = outputLevel;
            _ErrorMessages = new MessageCollection();
            controller.EngineControllerExecutionError += OnEngineControllerExecutionError;
            controller.UnhandledExceptionHandler += OnUnhandledExceptionThrown;
            controller.ScriptExecutionErrorHandler += OnScriptExecutionError;
            controller.ScriptExecutionAbortedHandler += OnScriptExecutionAborted;
        }

        public void Dispose()
        {
            Dispose(true);

            // Use SupressFinalize in case a subclass
            // of this type implements a finalizer.
            //GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            // If you need thread safety, use a lock around these 
            // operations, as well as in your methods that use the resource.
            if (!_Disposed)
            {
                //if (disposing)
                //{
                    _Controller.EngineControllerExecutionError -= OnEngineControllerExecutionError;
                    _Controller.UnhandledExceptionHandler -= OnUnhandledExceptionThrown;
                    _Controller.ScriptExecutionErrorHandler -= OnScriptExecutionError;
                    _Controller.ScriptExecutionAbortedHandler -= OnScriptExecutionAborted;
                //}
                _Disposed = true;
            }
        }

        public double GetTimeRemainingMultiplier()
        {
            double total = 0.0d;
            // add in the number of items that are not completed
            total = Math.Max(0.0d, (double)(_MetaTracker.TotalElementCount - _MetaTracker.CurrentElementNumber));
            // remove the fraction of items being processed that are complete
            double current = 0.0d;
            for (int i = 0; i < Controller.Engines.Length; i++)
            {
                if ((Controller.Engines[i].Engine != null) && (Controller.Engines[i].Engine.Tracker.CurrentState == TimeTracker.TimeTrackerState.Running))
                {
                    current += (Controller.Engines[i].Engine.Tracker.CurrentElementNumber / Controller.Engines[i].Engine.Tracker.TotalElementCount);
                }
            }
            current = Math.Max(0.0d, current);
            current = Math.Min(current, (double)Controller.Engines.Length);

            total -= current;

            total = Math.Max(0.0d, total);
            total = Math.Min(total, (double)_MetaTracker.TotalElementCount);

            total /= Controller.EngineControllerConfiguration.EngineThreadCount;
            
            // return the reciprocal since that's what is easiest to use in the TimeTracker code
            //total = 1.0d / total;

            return total;
        }

        protected bool DisplayEstimatedTimeRemaining(TimeRemainingInformation tri, int elementsRemaining)
        {
            // the scaling factor is padded by 10% because it's better to finish early than to finish 
            // later than the estimate
            _MetaTracker.TimeEstimationScalingFactor = TimeTracker.GetTimeEstimationScalingFactor((double)_Controller.GetActiveEngineCount());

            if ((tri.EstimatedTimeRemainingAverage.TotalSeconds == TimeTracker.DefaultTimeSpanSeconds) ||
                (tri.EstimatedTimeRemainingMedian.TotalSeconds == TimeTracker.DefaultTimeSpanSeconds))
            {
                return false;
            }

            if ((tri.EstimatedTimeRemainingAverage.TotalSeconds == 0) &&
                (tri.EstimatedTimeRemainingMedian.TotalSeconds == 0) &&
                (elementsRemaining >
                _MaximumNumberOfRemainingImagesBeforeDisplayingNoRemainingTime))
            {
                return false;
            }

            return true;
        }

        protected static void ThrowFileIOException(string filePath, string activityDescription, Exception innerException)
        {
            throw new Exception("Unable to " + activityDescription + " '" + filePath +
                "'. The error description was '" + innerException.Message + "'. Verify that a valid path " +
                "was specified, that access permissions to that location allow writing, " +
                "and that there is sufficient space available on the storage device.", innerException);
        }

        protected string GetDirectoryInfoString(DirectoryProcessingArgs d)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("directory ");
            builder.Append(_MetaTracker.CurrentElementNumber.ToString());
            if (_MetaTracker.TotalElementCount > 0)
            {
                builder.Append(" of ");
                builder.Append(_MetaTracker.TotalElementCount.ToString());
            }
            builder.Append(" - '");
            builder.Append(d.DirectoryPath);
            builder.Append("' on ");
            builder.Append(_Controller.Engines[d.EngineNumber].Engine.Tracker.StartTime.ToLongDateString());
            builder.Append(" at ");
            builder.Append(_Controller.Engines[d.EngineNumber].Engine.Tracker.StartTime.ToLongTimeString());
            builder.Append(".");
            return builder.ToString();
        }

        protected static string GetCurrentDateTimeString()
        {
            StringBuilder builder = new StringBuilder();
            DateTime n = DateTime.Now;
            builder.Append(n.Year.ToString());
            builder.Append("-");
            builder.Append(String.Format("{0:00}", n.Month));
            builder.Append("-");
            builder.Append(String.Format("{0:00}", n.Day));
            builder.Append("@");
            builder.Append(String.Format("{0:00}", n.Hour));
            builder.Append(":");
            builder.Append(String.Format("{0:00}", n.Minute));
            builder.Append(":");
            builder.Append(String.Format("{0:00}", n.Second));
            return builder.ToString();
        }

        public void OnEngineControllerExecutionError(object sender, EngineManagementEventArgs e)
        {
            //WriteOutput(e.EngineNumber, e.GlobalThreadNumber, e.Message);
            ErrorMessages.AddNewMessage("_Controller", "[Engine: " + e.EngineNumber + "Global Thread: " + e.GlobalThreadNumber + "] - " + e.Message);
        }

        public void OnUnhandledExceptionThrown(object sender, UnhandledExceptionArgs u)
        {
            //WriteOutput(u.EngineNumber, u.GlobalThreadNumber, "Unhandled exception: " + u.Exception.Message);
            ErrorMessages.AddNewMessage("_Controller", "[Engine: " + u.EngineNumber + "Global Thread: " + 
                u.GlobalThreadNumber + "] - Unhandled exception: " + u.Exception);
        }

        public void OnScriptExecutionError(object sender, ScriptExecutionErrorArgs e)
        {
            for (int i = 0; i < e.ErrorMessages.Length; i++)
            {
                ErrorMessages.AddNewMessage(e.DirectoryPath, e.ErrorMessages[i]);
            }
        }

        public void OnScriptExecutionAborted(object sender, ScriptExecutionEventArgs s)
        {
            //WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Script execution aborted on " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongDateString() +
            //    " at " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongTimeString() + " Reason: '" + s.Message + "'.");
            ErrorMessages.AddNewMessage(s.DirectoryPath, "Script execution aborted. Reason: '" + s.Message + "'.");
        }
    }
}
