// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class TMSBInputConfiguration
    {
        protected SpectralBand[] _SpectralBands;
        //protected string _RedBandName;
        //protected string _GreenBandName;
        //protected string _BlueBandName;

        #region Properties

        public SpectralBand[] SpectralBands
        {
            get
            {
                return _SpectralBands;
            }
            set
            {
                _SpectralBands = value;
            }
        }


        /// <summary>
        /// The name of the band which contains the colour red (as it appears to humans). 
        /// Used for generating the luminance channel and/or renaming non-false-colour output, 
        /// if either of those options are enabled.
        /// </summary>
        public string RedBandName
        {
            get
            {
                //return _RedBandName;
                if (ContainsSpecialSpectralBand(SpecialBandName.Red))
                {
                    return GetSpecialSpectralBand(SpecialBandName.Red).BandName;
                }
                return "_not_present_";
            }
            //set
            //{
            //    _RedBandName = value;
            //}
        }

        /// <summary>
        /// The name of the band which contains the colour green (as it appears to humans). 
        /// Used for generating the luminance channel and/or renaming non-false-colour output, 
        /// if either of those options are enabled.
        /// </summary>
        public string GreenBandName
        {
            get
            {
                if (ContainsSpecialSpectralBand(SpecialBandName.Green))
                {
                    return GetSpecialSpectralBand(SpecialBandName.Green).BandName;
                }
                return "_not_present_";
            }
            //set
            //{
            //    _GreenBandName = value;
            //}
        }

        /// <summary>
        /// The name of the band which contains the colour blue (as it appears to humans). 
        /// Used for generating the luminance channel and/or renaming non-false-colour output, 
        /// if either of those options are enabled.
        /// </summary>
        public string BlueBandName
        {
            get
            {
                if (ContainsSpecialSpectralBand(SpecialBandName.Blue))
                {
                    return GetSpecialSpectralBand(SpecialBandName.Blue).BandName;
                }
                return "_not_present_";
            }
            //set
            //{
            //    _BlueBandName = value;
            //}
        }

        #endregion

        public TMSBInputConfiguration()
        {
            //_RedBandName = "R";
            //_GreenBandName = "G";
            //_BlueBandName = "B";
            _SpectralBands = new SpectralBand[0];
        }

        public void AddSpectralBand(SpectralBand newBand)
        {
            Array.Resize(ref _SpectralBands, _SpectralBands.Length + 1);
            int bandNum = _SpectralBands.Length - 1;
            _SpectralBands[bandNum] = newBand;
            _SpectralBands[bandNum].BandNumber = bandNum;
        }

        public SpectralBand[] GetBandsByCategory(BandCategory category)
        {
            ArrayList bands = new ArrayList();

            foreach (SpectralBand band in SpectralBands)
            {
                if (band.IsBandMember(category))
                {
                    bands.Add(band);
                }
            }

            SpectralBand proto = new SpectralBand();
            return (SpectralBand[])bands.ToArray(proto.GetType());
        }

        public bool ContainsSpecialSpectralBand(SpecialBandName specialBand)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (band.SpecialBandNames == specialBand)
                {
                    return true;
                }
            }
            return false;
        }

        public SpectralBand GetSpecialSpectralBand(SpecialBandName specialBand)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (band.SpecialBandNames == specialBand)
                {
                    return band;
                }
            }
            throw new ArgumentException("The input configuration does not contain the special band '" + specialBand.ToString() + "'.");
        }

        public bool ContainsBand(string bandName)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (band.BandName == bandName)
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// This method applies the properties of an input configuration file to the real-world contents of a source directory.
        /// For example, if the configuration specifies a band whose source file is "NIR.TIF" and that file is not present
        /// in the source directory, that band (and any variations which depend on it) will not be present in the 
        /// "actualized" configuration.
        /// </summary>
        public static TMSBInputConfiguration CreateActualizedInputConfig(TMSBInputConfiguration abstractConfig, string baseInputDirectory)
        {
            TMSBInputConfiguration result = new TMSBInputConfiguration();
            //result.RedBandName = abstractConfig.RedBandName;
            //result.GreenBandName = abstractConfig.GreenBandName;
            //result.BlueBandName = abstractConfig.BlueBandName;

            foreach (SpectralBand band in abstractConfig.SpectralBands)
            {
                SpectralBand newBand = new SpectralBand(band, false);
                bool foundSourceFile = false;
                string filePath = baseInputDirectory + Path.DirectorySeparatorChar + band.BandData[0].FileName;
                if (File.Exists(filePath))
                {
                    foundSourceFile = true;
                    newBand.SetPrimaryImageData(band.GetPrimaryImageData());
                }
                filePath = baseInputDirectory + Path.DirectorySeparatorChar + band.BandData[1].FileName;
                if (File.Exists(filePath))
                {
                    foundSourceFile = true;
                    newBand.SetSecondaryImageData(band.GetSecondaryImageData());
                }
                if (foundSourceFile == true)
                {
                    result.AddSpectralBand(newBand);
                }
            }

            // copy over all tinting colours - if any of them are unreferenced, no harm done
            //result.TintingColours = abstractConfig.TintingColours;

            // copy over all the tint mappings - any that are unreferenced will be cleaned up
            // by code in the ImageCube class after the greyscale channels have been created
            //result.TintMappings = abstractConfig.TintMappings;

            return result;
        }
    }
}
