// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci.TransformationProfiles
{
    public class NormalizationAboutArbitraryValueProfile : NormalizationProfile
    {
        protected float _InputCenterPoint;
        protected bool _NormalizeUpperAndLowerIndependently;
        protected float _OutputLowerBound;
        protected float _OutputUpperBound;

        #region Properties

        /// <summary>
        /// NAAV only:
        /// The centerpoint about which to normalize the input data.
        /// Note: setting this value sets OutputCenterPoint as well, so if OutputCenterPoint
        /// is being specified, set it after setting this value, not before.
        /// </summary>
        public float InputCenterPoint
        {
            get
            {
                return _InputCenterPoint;
            }
            set
            {
                _InputCenterPoint = value;
                //OutputCenterPoint = value;
            }
        }

        /// <summary>
        /// NAAV only:
        /// If this value is false, then the multiplier used to normalize the data is the lesser of 
        /// the two which would make the lowest negative value equal to the lower bound, and 
        /// the highest positive value equal to the upper bound.
        /// If this value is true, then the positive and negative values are normalized independently
        /// of each other. This maximizes the expansion of the data of interest to fit the available
        /// space, but masks the relationship between the positive and negative data.
        /// </summary>
        public bool NormalizeUpperAndLowerIndependently
        {
            get
            {
                return _NormalizeUpperAndLowerIndependently;
            }
            set
            {
                _NormalizeUpperAndLowerIndependently = value;
            }
        }

        /// <summary>
        /// NAAV only:
        /// The number to which negative values should be scaled so that the lowest negative value
        /// is equal to it.
        /// </summary>
        public float OutputLowerBound
        {
            get
            {
                return _OutputLowerBound;
            }
            set
            {
                if (value >= OutputUpperBound)
                {
                    throw new ArgumentOutOfRangeException("The lower bound cannot be greater than or equal to the upper bound.");
                }
                _OutputLowerBound = value;
            }
        }

        /// <summary>
        /// NAAV only:
        /// The number to which positive values should be scaled so that the highest positive value
        /// is equal to it.
        /// </summary>
        public float OutputUpperBound
        {
            get
            {
                return _OutputUpperBound;
            }
            set
            {
                if (value <= OutputLowerBound)
                {
                    throw new ArgumentOutOfRangeException("The upper bound cannot be less than or equal to the lower bound.");
                }
                _OutputUpperBound = value;
            }
        }

        #endregion

        public NormalizationAboutArbitraryValueProfile()
            : base()
        {
            InputCenterPoint = 0.0f;
            NormalizeUpperAndLowerIndependently = false;
            OutputLowerBound = -1.0f;
            OutputUpperBound = 1.0f;
        }

        public override string WrapScriptObjectInTransformationCode(string scriptObject)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("naav(");
            builder.Append(scriptObject);
            builder.Append(", center = ");
            builder.Append(InputCenterPoint.ToString());
            builder.Append(", ubound = ");
            builder.Append(OutputUpperBound.ToString());
            builder.Append(", lbound = ");
            builder.Append(OutputLowerBound.ToString());
            builder.Append(", independent = ");
            if (NormalizeUpperAndLowerIndependently)
            {
                builder.Append("1");
            }
            else
            {
                builder.Append("0");
            }
            builder.Append(", anormalize = ");
            if (Stance == NormalizationStance.Aggressive)
            {
                builder.Append("1");
            }
            else
            {
                builder.Append("0");
            }

            if (Stance == NormalizationStance.Aggressive)
            {
                builder.Append(", steps = ");
                builder.Append(QuantizationStepCount.ToString());
                builder.Append(", backlash = ");
                builder.Append(Backlash.ToString());
                builder.Append(", threshold = ");
                builder.Append(Threshold.ToString());
                builder.Append(", threshrepeat = ");
                builder.Append(ThresholdRepetitions.ToString());
                builder.Append(", searchwidth = ");
                builder.Append(SearchWidth.ToString());
            }

            builder.Append(")");

            //if (OutputCenterPoint != InputCenterPoint)
            //{
            //    builder.Insert(0, "(");
            //    builder.Append(" + ");
            //    builder.Append(OutputCenterPoint.ToString());
            //    builder.Append(")");
            //}

            return builder.ToString();
        }

        public override string GetProfileTypeName()
        {
            return "Normalization About Arbitrary Value";
        }

    }
}
