// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class NormalizationProfile : TransformationProfile
    {
        public enum NormalizationStance
        {
            Standard,
            Aggressive
        }

        protected NormalizationProfile.NormalizationStance _Stance;

        // aggressive normalization parameters
        protected int _QuantizationStepCount;
        protected float _Threshold;
        protected int _ThresholdRepetitions;
        protected int _SearchWidth;
        protected int _Backlash;

        #region Properties

        /// <summary>
        /// The basic type of normalization to use - standard (susceptible to outliers) versus aggressive (may clip data more severely than desired).
        /// </summary>
        public NormalizationStance Stance
        {
            get
            {
                return _Stance;
            }
            set
            {
                _Stance = value;
            }
        }

        /// <summary>
        /// Aggressive normalization only:
        /// The number of discrete segments to divide the data into for purposes of normalization.
        /// Can also be thought of as the number of columns in the histogram.
        /// </summary>
        public int QuantizationStepCount
        {
            get
            {
                return _QuantizationStepCount;
            }
            set
            {
                if (value < 3)
                {
                    throw new ArgumentOutOfRangeException("The quantization step count must be a positive integer of 3 or greater. Low values are unlikely to yield useful results.");
                }
                _QuantizationStepCount = value;
            }
        }

        /// <summary>
        /// Aggressive normalization only:
        /// The percentage (represented as a float between 0.0 and 1.0) that a particular 
        /// quantization step must achieve (of the maximum value attained by any quantization step)
        /// in order to be considered a "hit".
        /// </summary>
        public float Threshold
        {
            get
            {
                return _Threshold;
            }
            set
            {
                if (value < 0.0f)
                {
                    throw new ArgumentOutOfRangeException("The threshold must be greater than or equal to zero.");
                }
                if (value > 1.0f)
                {
                    throw new ArgumentOutOfRangeException("The threshold must be less than or equal to one.");
                }
                _Threshold = value;
            }
        }

        /// <summary>
        /// Aggressive normalization only:
        /// The number of times within a given search window that the threshold must be met or 
        /// exceeded in order for the first "hit" out of that window to be used as the high or 
        /// low value for the data.
        /// </summary>
        public int ThresholdRepetitions
        {
            get
            {
                return _ThresholdRepetitions;
            }
            set
            {
                if (value < 1)
                {
                    throw new ArgumentOutOfRangeException("The number of threshold repetitions must be a positive, non-zero integer.");
                }
                //if (value > SearchWidth)
                //{
                //    throw new ArgumentOutOfRangeException("The number of threshold repetitions must be less than or equal to the search width.");
                //}
                _ThresholdRepetitions = value;
            }
        }

        /// <summary>
        /// Aggressive normalization only:
        /// If ThresholdRepetitions is greater than 1, this value represents the maximum number 
        /// of quantization segments that are allowed to separate the first "hit" from the last.
        /// </summary>
        public int SearchWidth
        {
            get
            {
                return _SearchWidth;
            }
            set
            {
                if (value < 1)
                {
                    throw new ArgumentOutOfRangeException("The search width must be a positive, non-zero integer.");
                }
                //if (value < ThresholdRepetitions)
                //{
                //    throw new ArgumentOutOfRangeException("The search width must be at least equal to the number of threshold repetitions.");
                //}
                _SearchWidth = value;
            }
        }

        /// <summary>
        /// Aggressive normalization only:
        /// When the high or low value for the data is found, it will be increased (or 
        /// decreased, respectively) by this many quantization steps. This is used to capture 
        /// a buffer of data near the normalization threshold which would otherwise be clipped.
        /// </summary>
        public int Backlash
        {
            get
            {
                return _Backlash;
            }
            set
            {
                if (value < 0)
                {
                    throw new ArgumentOutOfRangeException("Backlash cannot be a negative number. Note: Forwardlash is not currently implemented.");
                }
                _Backlash = value;
            }
        }

        #endregion

        public NormalizationProfile()
            : base()
        {
            Stance = NormalizationStance.Standard;
            QuantizationStepCount = 100;
            Threshold = 0.05f;
            SearchWidth = 5;
            ThresholdRepetitions = 3;
            Backlash = 1;
        }

        public override string WrapScriptObjectInTransformationCode(string scriptObject)
        {
            StringBuilder builder = new StringBuilder();

            if (Stance == NormalizationStance.Standard)
            {
                builder.Append("normalize(" + scriptObject);
            }
            else
            {
                builder.Append("anormalize(");
                builder.Append(scriptObject);
                builder.Append(", steps = ");
                builder.Append(QuantizationStepCount.ToString());
                builder.Append(", backlash = ");
                builder.Append(Backlash.ToString());
                builder.Append(", threshold = ");
                builder.Append(Threshold.ToString());
                builder.Append(", threshrepeat = ");
                builder.Append(ThresholdRepetitions.ToString());
                builder.Append(", searchwidth = ");
                builder.Append(SearchWidth.ToString());
            }

            builder.Append(")");

            return builder.ToString();
        }

        public static NormalizationStance NormalizationStanceFromString(string stanceName)
        {
            string sn = stanceName.Trim().ToLower();

            switch (sn)
            {
                case "standard":
                    return NormalizationStance.Standard;
                    break;
                case "aggressive":
                    return NormalizationStance.Aggressive;
                    break;
            }
            throw new ArgumentException("Unable to convert the string '" + stanceName +
                "' into a recognized NormalizationStance.");
        }

        public override string GetProfileTypeName()
        {
            return "Normalization";
        }


    }
}
