// BenLincoln.Utility - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.Utility.

// BenLincoln.Utility is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.Utility is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.Utility (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace BenLincoln.Utility
{
    public struct TimeRemainingInformation
    {
        public TimeSpan TotalElapsedTime;
        public TimeSpan AverageTimePerElement;
        public TimeSpan MedianTimePerElement;
        public TimeSpan MaxTimePerElement;
        public string AverageTimePerElementLongString;
        public string MedianTimePerElementLongString;
        public string AverageTimePerElementShortString;
        public string MedianTimePerElementShortString;
        public TimeSpan EstimatedTimeRemainingAverage;
        public TimeSpan EstimatedTimeRemainingMedian;
        public TimeSpan EstimatedTimeRemainingMax;
        public float EstimatedTimeRemainingAverageConfidence;
        public float EstimatedTimeRemainingMedianConfidence;
        public string TotalElapsedTimeLongString;
        public string TotalElapsedTimeShortString;
        public string EstimatedTimeRemainingAverageLongString;
        public string EstimatedTimeRemainingMedianLongString;
        public string EstimatedTimeRemainingMaxLongString;
        public string EstimatedTimeRemainingAverageShortString;
        public string EstimatedTimeRemainingMedianShortString;
        public string EstimatedTimeRemainingMaxShortString;
        public string EstimatedTimeRemainingLongString;
        public string EstimatedTimeRemainingShortString;
    }

    public class TimeTracker : NumericValueTracker
    {
        // the value to multiply estimated time remaining by
        // the current value of 1.1 adds 10% because it's better to finish earlier than the estimate
        // than later
        protected const double _TimeEstimationScalingFactorMultiplier = 1.1d;

        public enum TimeTrackerState
        {
            Stopped,
            Running
        }

        protected TimeTrackerState _CurrentState;

        protected double _TimeEstimationScalingFactor;

        protected DateTime _StartTime;
        protected DateTime _EndTime;
        //protected TimeSpan[] _ElapsedTimes;
        //protected TimeRemainingInformation _LastResult;
        protected int _CurrentElementNumber;
        protected int _TotalElementCount;

        protected int _PreviousEstimationValuesToRetain = 16;
        protected double[] _PreviousAverageValues;
        protected double[] _PreviousMedianValues;
        protected double _NoPreviousDataValue = -1.0d;

        // the length of time returned if a timespan cannot be calculated
        // this may seem like kind of a hack, but the reason I didn't use 
        // TimeSpan.MinValue or TimeSpan.MaxValue is that as soon as one 
        // starts throwing those around, many calculations are likely to 
        // generate exceptions when they go one tick above or below the 
        // value (depending on which is used)
        public const int DefaultTimeSpanSeconds = 26354000;
        //public const int DefaultTimeSpanSeconds = 0;

        protected DateTime _DefaultEndTime;

        public const string NoTimeRemainingString = "Less than one minute";

        // the percent difference (represented as a float between 0.0 and 1.0) 
        // below which only the maximum estimated time remaining is returned,
        // instead of both the maximum and the minimum
        protected float _EstimatedTimeThresholdPercent;

        protected TimeRemainingInformation _LastTimeRemainingInformation;
        protected bool _UpdatingTimeRemainingInformation;

        protected readonly object UpdateLockObject = new object();
        protected readonly object TimeRemainingLockObject = new object();

        #region Properties

        public TimeTrackerState CurrentState
        {
            get
            {
                return _CurrentState;
            }
        }

        /// <summary>
        /// when estimates are returned, multiply then by this value
        /// e.g. to take multithreaded processing times into account
        /// </summary>
        public double TimeEstimationScalingFactor
        {
            get
            {
                return _TimeEstimationScalingFactor;
            }
            set
            {
                _TimeEstimationScalingFactor = value;
            }
        }

        public DateTime StartTime
        {
            get
            {
                return _StartTime;
            }
        }

        public DateTime EndTime
        {
            get
            {
                return _EndTime;
            }
        }

        public TimeSpan[] ElapsedTimes
        {
            get
            {
                TimeSpan[] spans = new TimeSpan[_Values.Length];
                for (int i = 0; i < _Values.Length; i++)
                {
                    spans[i] = TimeSpan.FromSeconds(_Values[i]);
                }
                return spans;
            }
        }

        public float EstimatedTimeThresholdPercent
        {
            get
            {
                return _EstimatedTimeThresholdPercent;
            }
            set
            {
                lock (this)
                {
                    _EstimatedTimeThresholdPercent = value;
                }
            }
        }

        public int CurrentElementNumber
        {
            get
            {
                return _CurrentElementNumber;
            }
            set
            {
                lock (UpdateLockObject)
                {
                    if (value < 0)
                    {
                        _CurrentElementNumber = 0;
                    }
                    else if (value > _TotalElementCount)
                    {
                        _CurrentElementNumber = _TotalElementCount;
                    }
                    else
                    {
                        _CurrentElementNumber = value;
                    }
                }
            }
        }

        public int TotalElementCount
        {
            get
            {
                return _TotalElementCount;
            }
            set
            {
                lock (UpdateLockObject)
                {
                    if (value >= 0)
                    {
                        _TotalElementCount = value;
                    }
                    else
                    {
                        _TotalElementCount = 0;
                    }
                }
            }
        }

        #endregion

        public TimeTracker()
        {
            _UpdatingTimeRemainingInformation = false;
            _CurrentState = TimeTrackerState.Stopped;
            TimeEstimationScalingFactor = TimeTracker.GetTimeEstimationScalingFactor(1.0d);
            _DefaultEndTime = DateTime.Parse("2982-06-27");
            _EstimatedTimeThresholdPercent = 0.1f;
            _StartTime = DateTime.Now;
            _CurrentElementNumber = 0;
            _TotalElementCount = 0;
            InitializePreviousValueArray(ref _PreviousAverageValues);
            InitializePreviousValueArray(ref _PreviousMedianValues);

            // the length of time returned if a timespan cannot be calculated
            // this may seem like kind of a hack, but the reason I didn't use 
            // TimeSpan.MinValue or TimeSpan.MaxValue is that as soon as one 
            // starts throwing those around, many calculations are likely to 
            // generate exceptions when they go one tick above or below the 
            // value (depending on which is used)
            _DefaultValue = DefaultTimeSpanSeconds;
        }

        protected void InitializePreviousValueArray(ref double[] targetArray)
        {
            targetArray = new double[_PreviousEstimationValuesToRetain];
            for (int i = 0; i < _PreviousEstimationValuesToRetain; i++)
            {
                targetArray[i] = _NoPreviousDataValue;
            }
        }

        public void Start()
        {
            lock (this)
            {
                _StartTime = DateTime.Now;
                _EndTime = _DefaultEndTime;
                _CurrentState = TimeTrackerState.Running;
            }
        }

        public void Stop()
        {
            lock (this)
            {
                _EndTime = DateTime.Now;
                _CurrentState = TimeTrackerState.Stopped;
            }
        }

        public void Reset()
        {
            lock (this)
            {
                _Values = new double[0];
                _CurrentElementNumber = 0;
            }
            Start();
        }

        protected double[] GetValidPreviousDataValues(double[] previousDataValues)
        {
            ArrayList result = new ArrayList();
            foreach (double d in previousDataValues)
            {
                bool includeValue = true;
                if (d == _NoPreviousDataValue)
                {
                    includeValue = false;
                }
                if (includeValue)
                {
                    result.Add(d);
                }
            }
            return (double[])result.ToArray(_NoPreviousDataValue.GetType());
        }

        protected void AddPreviousDataValue(ref double[] previousDataArray, double newValue)
        {
            double[] newArray = new double[previousDataArray.Length];
            Array.Copy(previousDataArray, 1, newArray, 0, (previousDataArray.Length - 1));
            newArray[previousDataArray.Length - 1] = newValue;
            previousDataArray = newArray;
        }

        protected float GetConfidence(double[] previousValues, double currentValue)
        {
            double[] validValues = GetValidPreviousDataValues(previousValues);
            if (validValues.Length == 0)
            {
                //Console.WriteLine("Debug: no previous values, using default confidence of 0%");
                return 0.0f;
            }
            //Console.WriteLine("Debug: current value is " + currentValue.ToString());
            //Console.WriteLine("Debug: previous values:");
            //foreach (double d in validValues)
            //{
            //    Console.WriteLine("\t" + d.ToString());
            //}
            double averageOfPrevious = GetAverageValue(validValues);
            //Console.WriteLine("Debug: average of previous values is " + averageOfPrevious.ToString());
            float percentChange = GetPercentChange(averageOfPrevious, currentValue);
            //Console.WriteLine("Debug: percent change is " + percentChange.ToString());
            float rawConfidence = 1.0f - percentChange;
            return rawConfidence;
            //1.0f - GetPercentChange(result.EstimatedTimeRemainingAverage.TotalSeconds, _LastResult.EstimatedTimeRemainingAverage.TotalSeconds)
        }

        public static double GetTimeEstimationScalingFactor(double simultaneousProcessCount)
        {
            return (_TimeEstimationScalingFactorMultiplier / Math.Max(1.0d, simultaneousProcessCount));
        }

        public TimeRemainingInformation GetTimeRemainingInformation()
        {
            //return GetTimeRemainingInformation(TimeEstimationScalingFactor);
            return GetTimeRemainingInformation(GetRemainingElementCount());
        }

        public TimeRemainingInformation GetTimeRemainingInformation(double remainingElementCount)
        {
            bool updateValue = true;
            lock (TimeRemainingLockObject)
            {
                if (_UpdatingTimeRemainingInformation)
                {
                    updateValue = false;
                }
                else
                {
                    _UpdatingTimeRemainingInformation = true;
                }
            }
            TimeRemainingInformation result = new TimeRemainingInformation();
            //lock (this)
            //{
            if (updateValue)
            {
                result.TotalElapsedTime = GetTotalElapsedTime();
                result.AverageTimePerElement = GetAverageTimePerElement();
                result.MedianTimePerElement = GetMedianTimePerElement();
                result.MaxTimePerElement = TimeSpan.FromSeconds(Maximum);
                result.AverageTimePerElementLongString = GetLongFormattedTimeSpanString(result.AverageTimePerElement);
                result.MedianTimePerElementLongString = GetLongFormattedTimeSpanString(result.MedianTimePerElement);
                result.MedianTimePerElementLongString = GetLongFormattedTimeSpanString(result.MaxTimePerElement);
                result.AverageTimePerElementShortString = GetShortFormattedTimeSpanString(result.AverageTimePerElement);
                result.MedianTimePerElementShortString = GetShortFormattedTimeSpanString(result.MedianTimePerElement);
                result.MedianTimePerElementShortString = GetShortFormattedTimeSpanString(result.MaxTimePerElement);

                //result.EstimatedTimeRemainingAverage = GetEstimatedRemainingTime(result.AverageTimePerElement);
                //result.EstimatedTimeRemainingMedian = GetEstimatedRemainingTime(result.MedianTimePerElement);

                try
                {
                    result.EstimatedTimeRemainingAverage = GetEstimatedRemainingTime(TimeSpan.FromSeconds(result.AverageTimePerElement.TotalSeconds * TimeEstimationScalingFactor), remainingElementCount);
                }
                catch (Exception ex)
                {
                    result.EstimatedTimeRemainingAverage = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
                }
                try
                {
                    result.EstimatedTimeRemainingMedian = GetEstimatedRemainingTime(TimeSpan.FromSeconds(result.MedianTimePerElement.TotalSeconds * TimeEstimationScalingFactor), remainingElementCount);
                }
                catch (Exception ex)
                {
                    result.EstimatedTimeRemainingMedian = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
                }
                try
                {
                    result.EstimatedTimeRemainingMax = GetEstimatedRemainingTime(TimeSpan.FromSeconds(Maximum * TimeEstimationScalingFactor), remainingElementCount);
                }
                catch (Exception ex)
                {
                    result.EstimatedTimeRemainingMax = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
                }

                result.EstimatedTimeRemainingAverageConfidence = GetConfidence(_PreviousAverageValues, result.AverageTimePerElement.TotalSeconds);
                result.EstimatedTimeRemainingMedianConfidence = GetConfidence(_PreviousMedianValues, result.MedianTimePerElement.TotalSeconds);

                result.TotalElapsedTimeLongString = GetLongFormattedTimeSpanString(result.TotalElapsedTime);
                result.TotalElapsedTimeShortString = GetShortFormattedTimeSpanString(result.TotalElapsedTime);
                result.EstimatedTimeRemainingAverageLongString = GetLongFormattedTimeSpanString(result.EstimatedTimeRemainingAverage);
                result.EstimatedTimeRemainingMedianLongString = GetLongFormattedTimeSpanString(result.EstimatedTimeRemainingMedian);
                result.EstimatedTimeRemainingMedianLongString = GetLongFormattedTimeSpanString(result.EstimatedTimeRemainingMax);
                result.EstimatedTimeRemainingAverageShortString = GetShortFormattedTimeSpanString(result.EstimatedTimeRemainingAverage);
                result.EstimatedTimeRemainingMedianShortString = GetShortFormattedTimeSpanString(result.EstimatedTimeRemainingMedian);
                result.EstimatedTimeRemainingMaxShortString = GetShortFormattedTimeSpanString(result.EstimatedTimeRemainingMax);
                result.EstimatedTimeRemainingLongString = GetEstimatedTimeRemainingLongString(result.EstimatedTimeRemainingAverage,
                    result.EstimatedTimeRemainingMedian, result.EstimatedTimeRemainingMax);
                result.EstimatedTimeRemainingShortString = GetEstimatedTimeRemainingShortString(result.EstimatedTimeRemainingAverage,
                    result.EstimatedTimeRemainingMedian, result.EstimatedTimeRemainingMax);

                //if ((_LastResult.EstimatedTimeRemainingAverage == null) ||
                //    (_LastResult.EstimatedTimeRemainingMedian == null))
                //{
                //    result.EstimatedTimeRemainingAverageConfidence = 0.0f;
                //    result.EstimatedTimeRemainingMedianConfidence = 0.0f;
                //}
                //else
                //{
                //    result.EstimatedTimeRemainingAverageConfidence = 1.0f - GetPercentChange(result.EstimatedTimeRemainingAverage.TotalSeconds,
                //        _LastResult.EstimatedTimeRemainingAverage.TotalSeconds);
                //    result.EstimatedTimeRemainingMedianConfidence = 1.0f - GetPercentChange(result.EstimatedTimeRemainingMedian.TotalSeconds,
                //        _LastResult.EstimatedTimeRemainingMedian.TotalSeconds);
                //}

                //_LastResult = result;

                _LastTimeRemainingInformation = result;

                AddPreviousValuesFromTimeRemainingInformationFunction(result.AverageTimePerElement.TotalSeconds, result.MedianTimePerElement.TotalSeconds);

                _UpdatingTimeRemainingInformation = false;

                return result;
            }
            else
            {
                //if (_LastTimeRemainingInformation == null)
                //{
                //    return result;
                //}
                //else
                //{
                    return _LastTimeRemainingInformation;
                //}
            }
        }

        protected void AddPreviousValuesFromTimeRemainingInformationThreaded(double totalSecondsAverage, double totalSecondsMedian)
        {
            try
            {
                Thread addThread = new Thread(new ParameterizedThreadStart(AddPreviousValuesFromTimeRemainingInformationThreadTarget));
                addThread.Start(new double[] { totalSecondsAverage, totalSecondsMedian });
            }
            catch (Exception ex)
            {
                // this is not a critical function, so fail silently
            }
        }

        protected void AddPreviousValuesFromTimeRemainingInformationThreadTarget(object obj)
        {
            if (obj is double[])
            {
                try
                {
                    double[] vals = (double[])obj;
                    if (vals.Length == 2)
                    {
                        AddPreviousValuesFromTimeRemainingInformationFunction(vals[0], vals[1]);
                    }
                }
                catch (Exception ex)
                {
                    // this is not a critical function, so fail silently
                }
            }
        }

        //delegate void AddPreviousValuesFromTimeRemainingInformationCallback(double totalSecondsAverage, double totalSecondsMedian);

        protected void AddPreviousValuesFromTimeRemainingInformationFunction(double totalSecondsAverage, double totalSecondsMedian)
        {
            //if (this.InvokeRequired)
            //{
            //    AddPreviousValuesFromTimeRemainingInformationCallback c =
            //        new AddPreviousValuesFromTimeRemainingInformationCallback(AddPreviousValuesFromTimeRemainingInformationFunction);
            //    this.Invoke(c, new object[] { totalSecondsAverage, totalSecondsMedian });
            //}
            //else
            //{
                lock (UpdateLockObject)
                {
                    if (totalSecondsAverage != DefaultTimeSpanSeconds)
                    {
                        AddPreviousDataValue(ref _PreviousAverageValues, totalSecondsAverage);
                    }
                    if (totalSecondsMedian != DefaultTimeSpanSeconds)
                    {
                        AddPreviousDataValue(ref _PreviousMedianValues, totalSecondsMedian);
                    }
                }
            //}
        }

        public TimeSpan GetTotalElapsedTime()
        {
            return DateTime.Now - _StartTime;
        }

        public string GetTotalElapsedTimeShortString()
        {
            return GetShortFormattedTimeSpanString(GetTotalElapsedTime());
        }

        public string GetTotalElapsedTimeLongString()
        {
            return GetLongFormattedTimeSpanString(GetTotalElapsedTime());
        }

        public string GetEstimatedTimeRemainingLongString()
        {
            return GetEstimatedTimeRemainingLongString(GetEstimatedRemainingTimeAverage(), GetEstimatedRemainingTimeMedian(), TimeSpan.FromSeconds(Maximum));
        }

        public string GetEstimatedTimeRemainingLongString(TimeSpan avgTime, TimeSpan medTime, TimeSpan maxTime)
        {
            return GetEstimatedTimeRemainingString(avgTime, medTime, maxTime, true);
        }

        //public string GetEstimatedTimeRemainingLongString(TimeSpan avgTime, float confidenceAvg, TimeSpan medTime, float confidenceMed)
        //{
        //    return GetEstimatedTimeRemainingString(avgTime, confidenceAvg, medTime, confidenceMed, true);
        //}

        public string GetEstimatedTimeRemainingShortString()
        {
            return GetEstimatedTimeRemainingShortString(GetEstimatedRemainingTimeAverage(), GetEstimatedRemainingTimeMedian(), TimeSpan.FromSeconds(Maximum));
        }

        public string GetEstimatedTimeRemainingShortString(TimeSpan avgTime, TimeSpan medTime, TimeSpan maxTime)
        {
            return GetEstimatedTimeRemainingString(avgTime, medTime, maxTime, false);
        }

        //protected string GetEstimatedTimeRemainingString(TimeSpan avgTime, TimeSpan medTime, bool useLongFormat)
        //{
        //    return GetEstimatedTimeRemainingString(avgTime, -1.0f, medTime, -1.0f, useLongFormat);
        //}

        protected string GetEstimatedTimeRemainingString(TimeSpan avgTime, TimeSpan medTime, TimeSpan maxTime,
            bool useLongFormat)
        {
            //if (_Values.Length > 0)
            //{
                try
                {
                    //TimeSpan maxTime = avgTime;
                    TimeSpan mt = maxTime;
                    TimeSpan minTime = medTime;

                    if (medTime > avgTime)
                    {
                        //maxTime = medTime;
                        minTime = avgTime;
                    }

                    if (minTime > maxTime)
                    {
                        TimeSpan t = minTime;
                        minTime = maxTime;
                        maxTime = t;
                    }

                    TimeSpan diff = maxTime - minTime;
                    if ((diff.TotalSeconds == 0) || (((float)diff.TotalSeconds / (float)maxTime.TotalSeconds) < _EstimatedTimeThresholdPercent))
                    {
                        if (useLongFormat)
                        {
                            return GetLongFormattedTimeSpanString(maxTime);
                        }
                        else
                        {
                            return GetShortFormattedTimeSpanString(maxTime);
                        }
                    }

                    if (useLongFormat)
                    {
                        if ((minTime.TotalSeconds == 0) && (maxTime.TotalSeconds > 0))
                        {
                            return GetLongFormattedTimeSpanString(maxTime) + " or less";
                        }
                        return "Between " + GetLongFormattedTimeSpanString(minTime) + " and " +
                            GetLongFormattedTimeSpanString(maxTime);
                    }
                    else
                    {
                        return GetShortFormattedTimeSpanString(minTime) + " - " + GetShortFormattedTimeSpanString(maxTime);
                    }
                }
                catch (Exception ex)
                {
                    return "Error during calculation: '" + ex.Message + "'";
                }
            //}
            //else
            //{
            //    return DefaultString;
            //}
        }

        public TimeSpan GetMedianTimePerElement()
        {
            TimeSpan result = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
            try
            {
                result = TimeSpan.FromSeconds(GetMedianValue());
            }
            catch (Exception ex)
            {
                result = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
            }
            return result;
        }

        public TimeSpan GetAverageTimePerElement()
        {
            TimeSpan result = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
            try
            {
                result = TimeSpan.FromSeconds(GetAverageValue());
            }
            catch (Exception ex)
            {
                result = TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
            }
            return result;
        }

        //public TimeSpan GetMedianTimePerElement()
        //{
        //    lock (this)
        //    {
        //        if (_ElapsedTimes.Length > 0)
        //        {
        //            long seconds = DefaultTimeSpanSeconds;
        //            if (_ElapsedTimes.Length > 1)
        //            {
        //                TimeSpan[] sortedTimes = new TimeSpan[_ElapsedTimes.Length];
        //                Array.Copy(_ElapsedTimes, sortedTimes, _ElapsedTimes.Length);
        //                Array.Sort(sortedTimes);
        //                if ((sortedTimes.Length % 2) == 1)
        //                {
        //                    int midPoint = ((sortedTimes.Length - 1) / 2);
        //                    seconds = (long)sortedTimes[midPoint].TotalSeconds;
        //                }
        //                else
        //                {
        //                    int sP = (sortedTimes.Length / 2) - 1;
        //                    seconds = (long)((sortedTimes[sP].TotalSeconds + sortedTimes[sP + 1].TotalSeconds) / 2);
        //                }
        //            }
        //            else
        //            {
        //                seconds = (long)_ElapsedTimes[0].TotalSeconds;
        //            }
        //            return TimeSpan.FromSeconds((double)seconds);
        //        }
        //        else
        //        {
        //            return TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
        //        }
        //    }
        //}

        //public TimeSpan GetAverageTimePerElement()
        //{
        //    lock (this)
        //    {
        //        if (_ElapsedTimes.Length > 0)
        //        {
        //            long seconds = 0;
        //            if (_ElapsedTimes.Length > 1)
        //            {
        //                foreach (TimeSpan span in _ElapsedTimes)
        //                {
        //                    seconds += (long)span.TotalSeconds;
        //                }
        //                seconds /= _ElapsedTimes.Length;
        //            }
        //            else
        //            {
        //                seconds = (long)_ElapsedTimes[0].TotalSeconds;
        //            }
        //            return TimeSpan.FromSeconds((double)seconds);
        //        }
        //        else
        //        {
        //            return TimeSpan.FromSeconds(DefaultTimeSpanSeconds);
        //        }
        //    }
        //}

        public int GetRemainingElementCount()
        {
            return (_TotalElementCount - _CurrentElementNumber);
        }

        public TimeSpan GetEstimatedRemainingTimeAverage()
        {
            return GetEstimatedRemainingTime(GetAverageTimePerElement(), GetRemainingElementCount());
        }

        public TimeSpan GetEstimatedRemainingTimeMedian()
        {
            return GetEstimatedRemainingTime(GetMedianTimePerElement(), GetRemainingElementCount());
        }

        protected TimeSpan GetEstimatedRemainingTime(TimeSpan timePerElement, double remainingElements)
        {
            //if ((_CurrentElementNumber < 1) || (_TotalElementCount < 1))
            if (_TotalElementCount < 1)
            {
                return TimeSpan.FromSeconds(_DefaultValue);
            }
            return TimeSpan.FromSeconds((int)(timePerElement.TotalSeconds * remainingElements));
        }

        public static string GetShortFormattedTimeSpanString(TimeSpan span)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(String.Format("{0:00}:", span.Days));
            builder.Append(String.Format("{0:00}:", span.Hours));
            builder.Append(String.Format("{0:00}:", span.Minutes));
            builder.Append(String.Format("{0:00}:", span.Seconds));

            return builder.ToString();
        }

        public static string GetLongFormattedTimeSpanString(TimeSpan span)
        {
            return GetLongFormattedTimeSpanString(span, -1.0f);
        }

        public static string GetLongFormattedTimeSpanString(TimeSpan span, float confidence)
        {
            StringBuilder builder = new StringBuilder();
            bool includeSeconds = true;
            if (AddElapsedTimeElement(ref builder, span.Days, "Day"))
            {
                includeSeconds = false;
            }
            if (AddElapsedTimeElement(ref builder, span.Hours, "Hour"))
            {
                includeSeconds = false;
            }
            AddElapsedTimeElement(ref builder, span.Minutes, "Minute");
            if (includeSeconds == true)
            {
                AddElapsedTimeElement(ref builder, span.Seconds, "Second");
            }
            string result = builder.ToString().Trim();
            if (result.Length > 1)
            {
                if (result.Substring(result.Length - 1) == ",")
                {
                    result = result.Substring(0, result.Length - 1);
                }
                if (confidence >= 0.0f)
                {
                    //Console.WriteLine("Debug: confidence is " + confidence.ToString());
                    int conf = (int)(confidence * 100.0f);
                    //Console.WriteLine("Debug: confidence x100 is " + conf.ToString());
                    result = result + " (Confidence: " + String.Format("{0:0}", conf) + "%)";
                }
            }
            else
            {
                //if (span.TotalSeconds > 0)
                //{
                //    throw new Exception("The string generated by the GetLongFormattedTimeSpanString function was empty, but the TimeSpan passed to it represented a non-zero number of seconds.");
                //}
                result = NoTimeRemainingString;
            }
            return result;
        }

        protected static bool AddElapsedTimeElement(ref StringBuilder builder, int elementValue, string elementName)
        {
            bool result = false;
            if (elementValue > 0)
            {
                builder.Append(elementValue.ToString());
                builder.Append(" ");
                builder.Append(elementName);
                if (elementValue > 1)
                {
                    builder.Append("s");
                }
                builder.Append(", ");
                result = true;
            }
            return result;
        }

        public void AddElapsedTime(TimeSpan newSpan)
        {
            AddValue(newSpan.TotalSeconds);
            //lock (_ElapsedTimes)
            //{
            //    if ((_ElapsedTimes == null) || (_ElapsedTimes.Length == 0))
            //    {
            //        _ElapsedTimes = new TimeSpan[1];
            //    }
            //    else
            //    {
            //        Array.Resize(ref _ElapsedTimes, _ElapsedTimes.Length + 1);
            //    }
            //    _ElapsedTimes[_ElapsedTimes.Length - 1] = newSpan;
            //}
        }
    }
}
