// The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of The Mirror's Surface Breaks.

// The Mirror's Surface Breaks is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Mirror's Surface Breaks is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Mirror's Surface Breaks (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.IO;
using System.Text;
using System.Windows.Forms;

using BenLincoln.DaVinci;

namespace TMSBGUI
{
    public partial class TMSBConfigurationSelector : UserControl
    {
        protected TMSBConfigurationFileType _ConfigFileType;
        protected TMSBConfigurationFile[] _ConfigFiles;
        protected TMSBConfigurationFile _CurrentConfigFile;

        protected string _DescriptionAppendText;

        protected System.EventHandler _DropdownChangedEventHandler;

        #region Properties

        public TMSBConfigurationFileType ConfigFileType
        {
            get
            {
                return _ConfigFileType;
            }
            set
            {
                _ConfigFileType = value;
                PopulateDropDown();
            }
        }

        public TMSBConfigurationFile[] ConfigFiles
        {
            get
            {
                return _ConfigFiles;
            }
            set
            {
                _ConfigFiles = value;
                UpdateControls();
            }
        }

        public TMSBConfigurationFile CurrentConfigFile
        {
            get
            {
                return _CurrentConfigFile;
            }
            set
            {
                _CurrentConfigFile = value;
                UpdateControls();
            }
        }

        public string DescriptionAppendText
        {
            get
            {
                return _DescriptionAppendText;
            }
            set
            {
                _DescriptionAppendText = value;
                if (_CurrentConfigFile != null)
                {
                    UpdateControls();
                }
            }
        }

        #endregion

        public TMSBConfigurationSelector()
        {
            InitializeComponent();

            _DescriptionAppendText = "";

            _ConfigFileType = TMSBConfigurationFileType.Input;

            //_ConfigFiles = new TMSBConfigurationFile[0];
        }

        public void UpdateControls()
        {
            if (CurrentConfigFile == null)
            {
                txtDescription.Text = "";
                txtWarnings.Text = "";
                txtFileName.Text = "";
                txtAuthor.Text = "";
                txtCreated.Text = "";
                txtLastModified.Text = "";
                btnExportXML.Enabled = false;
            }
            else
            {
                txtDescription.Text = CurrentConfigFile.Description + Environment.NewLine + DescriptionAppendText;
                txtWarnings.Text = CurrentConfigFile.Warnings;
                txtFileName.Text = CurrentConfigFile.FilePath;
                txtAuthor.Text = CurrentConfigFile.Author;
                txtCreated.Text = GetDateString(CurrentConfigFile.Created.ToLocalTime());
                txtLastModified.Text = GetDateString(CurrentConfigFile.LastModified.ToLocalTime());
                btnExportXML.Enabled = true;
            }
        }

        public void PopulateDropDown()
        {
            string selectedFilePath = "";
            if (CurrentConfigFile != null)
            {
                selectedFilePath = CurrentConfigFile.FilePath;
            }

            _ConfigFiles = TMSBConfigurationFile.GetConfigurationFiles(false, ConfigFileType);
            Array.Sort(_ConfigFiles);

            cbConfigFiles.Items.Clear();

            foreach (TMSBConfigurationFile f in ConfigFiles)
            {
                cbConfigFiles.Items.Add(f.Name);
            }

            UpdateControls();

            if (selectedFilePath != "")
            {
                try
                {
                    SetSelectedItemByFilePath(selectedFilePath);
                }
                catch (Exception ex)
                {
                    // do nothing - item no longer exists
                }
            }
        }

        protected string GetDateString(DateTime dt)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(dt.Year.ToString());
            builder.Append("-");
            builder.Append(String.Format("{0:00}", dt.Month));
            builder.Append("-");
            builder.Append(String.Format("{0:00}", dt.Day));
            builder.Append(" @ ");
            builder.Append(String.Format("{0:00}", dt.Hour));
            builder.Append(":");
            builder.Append(String.Format("{0:00}", dt.Minute));
            builder.Append(":");
            builder.Append(String.Format("{0:00}", dt.Second));

            return builder.ToString();
        }

        private void cbConfigFiles_SelectedIndexChanged(object sender, EventArgs e)
        {
            int newIndex = cbConfigFiles.SelectedIndex;
            if (newIndex < ConfigFiles.Length)
            {
                CurrentConfigFile = ConfigFiles[cbConfigFiles.SelectedIndex];
            }
            else
            {
                CurrentConfigFile = null;
            }
        }

        public void SetSelectedItemByFilePath(string configFilePath)
        {
            cbConfigFiles.SelectedIndex = GetIndexByFilePath(configFilePath);
        }

        public void SetSelectedItemByFileName(string configFileName)
        {
            cbConfigFiles.SelectedIndex = GetIndexByFileName(configFileName);
        }

        public int GetIndexByFilePath(string configFilePath)
        {
            for (int i = 0; i < ConfigFiles.Length; i++)
            {
                if (ConfigFiles[i].FilePath == configFilePath)
                {
                    return i;
                }
            }
            return -1;
        }

        public int GetIndexByFileName(string configFileName)
        {
            for (int i = 0; i < ConfigFiles.Length; i++)
            {
                if (ConfigFiles[i].Name == configFileName)
                {
                    return i;
                }
            }
            return -1;
        }

        public void SetDropdownEventHandler(System.EventHandler newHandler)
        {
            RemoveDropdownEventHandler();
            _DropdownChangedEventHandler = newHandler;
            cbConfigFiles.SelectedIndexChanged += newHandler;
        }

        public void RemoveDropdownEventHandler()
        {
            if (_DropdownChangedEventHandler != null)
            {
                cbConfigFiles.SelectedIndexChanged -= _DropdownChangedEventHandler;
                _DropdownChangedEventHandler = null;
            }
        }

        private void btnExportXML_Click(object sender, EventArgs e)
        {
            ExportXML();
        }

        public void ExportXML()
        {
            if (CurrentConfigFile != null)
            {
                SaveFileDialog sf = new SaveFileDialog();
                sf.AddExtension = true;
                sf.OverwritePrompt = true;
                sf.CheckPathExists = true;
                sf.DefaultExt = "xml";
                sf.Filter = "XML Files (*.xml)|*.xml|All Files (*.*)|*.*";
                sf.FileName = Path.GetFileName(CurrentConfigFile.FilePath);
                sf.InitialDirectory = TMSBConfigurationFile.GetPersonalConfigDirectoryName();
                sf.Title = "Export Configuration As XML";

                DialogResult r = sf.ShowDialog();
                if (r == DialogResult.OK)
                {
                    try
                    {
                        Stream iStream = TMSBConfigurationFile.GetConfigurationStream(Path.GetFileName(CurrentConfigFile.FilePath));
                        StreamReader iReader = new StreamReader(iStream, UnicodeEncoding.Unicode);
                        FileStream oStream = new FileStream(sf.FileName, FileMode.Create, FileAccess.Write);
                        StreamWriter oWriter = new StreamWriter(oStream, UnicodeEncoding.Unicode);
                        oWriter.Write(iReader.ReadToEnd());
                        oWriter.Close();
                        oStream.Close();
                        iReader.Close();
                        iStream.Close();
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show("An error occurred while exporting the file: '" + ex.Message + "'",
                            "Export Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
            }
        }
    }
}
