// The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of The Mirror's Surface Breaks.

// The Mirror's Surface Breaks is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Mirror's Surface Breaks is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Mirror's Surface Breaks (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using BenLincoln.DaVinci;
using BenLincoln.DaVinci.GUI;
  
namespace TMSBGUI
{
    public partial class frmMain : Form, IDisposable
    {
        protected const string _DefaultExecutionOptionsFileWindows = "Execution-Default-Windows.xml";
        protected const string _DefaultExecutionOptionsFileUnixLike = "Execution-Default-UnixLike.xml";
        protected const string _LastUsedOptionsFileName = "Execution-LastUsed.xml";

        protected const int _WaitLengthWaitWindow = 100;

        protected TMSBEngineControllerConfigurationFile _ExecutionOptions;

        protected frmAbout _AboutWindow;
        protected frmProcessingStatus _ProcessingStatusWindow;
        protected frmPleaseWait _WaitWindow;

        protected FormWindowState _PreviousWindowState;

        protected bool _WaitWindowShouldBeOpen;

        public frmMain()
        {
            InitializeComponent();

            _WaitWindowShouldBeOpen = false;

            csInput.ConfigFileType = TMSBConfigurationFileType.Input;
            tpInputConfig.Text = TMSBConfigurationFile.TypeToTypeName(TMSBConfigurationFileType.Input);

            csProcessing.ConfigFileType = TMSBConfigurationFileType.Processing;
            tpProcessingConfig.Text = TMSBConfigurationFile.TypeToTypeName(TMSBConfigurationFileType.Processing);

            csInput.SetDropdownEventHandler(ProcessingConfigOptionChanged);
            csProcessing.SetDropdownEventHandler(ProcessingConfigOptionChanged);
            csOutput.SetDropdownEventHandler(ProcessingConfigOptionChanged);
            this.ProcessingConfigOptionChanged(this, new EventArgs());

            csOutput.ConfigFileType = TMSBConfigurationFileType.Output;
            tpOutputConfig.Text = TMSBConfigurationFile.TypeToTypeName(TMSBConfigurationFileType.Output);

            //pqgDirectoryQueue.AddDragDropHandler(Queue_DragDrop);
            pqgDirectoryQueue.EnableDragAndDrop = true;
            pqgDirectoryQueue.AddItemsRecursively = true;
            pqgDirectoryQueue.SortBy(ProcessableItemQueueGrid.SortColumn.Name);

            // this was causing problems, so I'm just not going to bother with it now
            //pqgDirectoryQueue.SetWaitStartedHandler(TriggerWaitWindowOpen);
            //pqgDirectoryQueue.SetWaitFinishedHandler(TriggerWaitWindowClose);
            
            // for debugging (apps run in VS2005 under UAC don't allow drag-and-drop)
            //pqgDirectoryQueue.AddDirectories(new string[] { "E:\\Adobe" });
            //pqgDirectoryQueue.AddDirectories(new string[] { @"D:\DaVinci\TMSB_Test\GUI_Small\01", @"D:\DaVinci\TMSB_Test\GUI_Small\02", 
            //    @"D:\DaVinci\TMSB_Test\GUI_Small\03", @"D:\DaVinci\TMSB_Test\GUI_Small\05", @"D:\DaVinci\TMSB_Test\GUI_Small\07", @"D:\DaVinci\TMSB_Test\GUI_Small\08",
            //    @"D:\DaVinci\TMSB_Test\GUI_Small\09" }, false);

            InitializeExecutionOptions();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing && (components != null))
            {
                components.Dispose();
            }
            if (_AboutWindow != null)
            {
                _AboutWindow.Dispose();
                _AboutWindow = null;
            }
            if (_ProcessingStatusWindow != null)
            {
                _ProcessingStatusWindow.Dispose();
                _ProcessingStatusWindow = null;
            }

            if (_ExecutionOptions != null)
            {
                _ExecutionOptions = null;
            }

            

            try
            {
                this.Events.Dispose();
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error disposing of delegate list: " + ex.Message);
                // do nothing
            }

            base.Dispose(disposing);
        }

        private void frmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            bool writeOptions = SetExecutionOptionsFromUI(false);
            if (writeOptions)
            {
                WriteLastUsedExecutionOptions();
            }
            //this.Dispose();
            Application.Exit();
        }

        protected void InitializeExecutionOptions()
        {
            string fullUserXOPath = GetLastUsedExecutionOptionsFilePath();

            if (File.Exists(fullUserXOPath))
            {
                ReadLastUsedExecutionOptions(fullUserXOPath);
            }
            else
            {
                SetDefaultExecutionOptions();
            }
        }

        protected void SetDefaultExecutionOptions()
        {
            string xoFile = _DefaultExecutionOptionsFileWindows;
            switch (TMSBEngineControllerConfiguration.AutoDetectCurrentExecutionPlatform())
            {
                case ExecutionPlatform.Windows:
                    xoFile = TMSBConfigurationFile.GetApplicationConfigDirectoryName() + Path.DirectorySeparatorChar + _DefaultExecutionOptionsFileWindows;
                    break;
                case ExecutionPlatform.UnixLike:
                    xoFile = TMSBConfigurationFile.GetApplicationConfigDirectoryName() + Path.DirectorySeparatorChar + _DefaultExecutionOptionsFileUnixLike;
                    break;
            }

            if (!File.Exists(xoFile))
            {
                MessageBox.Show("TMSB was unable to locate the default execution options file ('" +
                    xoFile + "'). You may be able to resolve this problem by reinstalling the application.",
                    "Missing File", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                try
                {
                    _ExecutionOptions = (TMSBEngineControllerConfigurationFile)TMSBConfigurationFile.GetConfigurationFile(xoFile, true);
                    SetUIFromExecutionOptions();
                    WriteLastUsedExecutionOptions();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("TMSB was unable to read the default execution options file. ('" +
                        xoFile + "'). The specific error message was '" +
                        ex.Message + "'. You may be able to resolve this problem by reinstalling the application.",
                        "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

            }
        }

        protected void ReadLastUsedExecutionOptions(string filePath)
        {
            try
            {
                _ExecutionOptions = (TMSBEngineControllerConfigurationFile)TMSBConfigurationFile.GetConfigurationFile(filePath, true);
                SetUIFromExecutionOptions();
            }
            catch (Exception ex)
            {
                MessageBox.Show("TMSB was unable to read the last used execution options file. The specific error message was '" +
                    ex.Message + "'. You may be able to resolve this problem by manually deleting the file '" + filePath +
                    "', then restarting this application.", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        protected void WriteLastUsedExecutionOptions()
        {
            WriteLastUsedExecutionOptions(GetLastUsedExecutionOptionsFilePath());
        }

        protected void WriteLastUsedExecutionOptions(string filePath)
        {
            //SetExecutionOptionsFromUI();

            string configDirectoryPath = Path.GetDirectoryName(filePath);

            bool proceed = true;

            if (!Directory.Exists(configDirectoryPath))
            {
                try
                {
                    Directory.CreateDirectory(configDirectoryPath);
                }
                catch (Exception ex)
                {
                    proceed = false;
                    MessageBox.Show("TMSB was unable to create the user configuration directory ('" +
                        configDirectoryPath + "'). The specific error message was '" +
                        ex.Message + "'. You may be able to resolve this problem by reinstalling the application.",
                        "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.WriteData(filePath);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("TMSB was unable to write the last used execution options files ('" +
                            filePath + "'). The specific error message was '" +
                            ex.Message + "'. You may be able to resolve this problem by reinstalling the application.",
                            "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        protected void SetUIFromExecutionOptions()
        {
            // processing criteria
            txtMinimumSourceBands.Text = _ExecutionOptions.EngineControllerConfig.MinimumSourceBandsInInputDirectory.ToString();
            pqgDirectoryQueue.AddItemsRecursively = _ExecutionOptions.EngineControllerConfig.ScanInputDirectoriesRecursively;

            // config files
            csInput.SetSelectedItemByFilePath(_ExecutionOptions.EngineControllerConfig.InputConfigFilePath);
            csProcessing.SetSelectedItemByFilePath(_ExecutionOptions.EngineControllerConfig.ProcessingConfigFilePath);
            csOutput.SetSelectedItemByFilePath(_ExecutionOptions.EngineControllerConfig.OutputConfigFilePath);

            // script processor
            SetComboBoxSelectedValue(cbGenerateScripts, 
                TMSBEngineControllerConfiguration.ScriptGenerationModeToString(_ExecutionOptions.EngineControllerConfig.GenerateScripts, 
                true));
            chkExecuteScripts.Checked = _ExecutionOptions.EngineControllerConfig.ExecuteScripts;
            txtScriptProcessorPath.Text = _ExecutionOptions.EngineControllerConfig.DaVinciCommand;
            txtScriptProcessorArgs.Text = _ExecutionOptions.EngineControllerConfig.DaVinciCommandArgs;
            SetComboBoxSelectedValue(cbScriptProcessPriority,
                TMSBEngineControllerConfiguration.ProcessPriorityClassToString(_ExecutionOptions.EngineControllerConfig.DaVinciProcessPriority));
            txtMaxDirectoryThreads.Text = _ExecutionOptions.EngineControllerConfig.EngineThreadCount.ToString();
            txtMaxProcessingThreads.Text = _ExecutionOptions.EngineControllerConfig.GlobalThreadCount.ToString();
            txtSubscriptTimeLimit.Text = _ExecutionOptions.EngineControllerConfig.ScriptExecutionTimeLimit.ToString();

            // statistics
            chkCollectPerfData.Checked = _ExecutionOptions.EngineControllerConfig.CollectPerformanceData;
            SetComboBoxSelectedValue(cbMemoryFormat,
                TMSBEngineControllerConfiguration.MemoryRepresentationToString(_ExecutionOptions.EngineControllerConfig.MemoryUnits, true));

            // debugging/troubleshooting
            SetComboBoxSelectedValue(cbLogLevel, TMSBEngineControllerConfiguration.OutputLevelToString(_ExecutionOptions.EngineControllerConfig.OutputLevel));
            txtLogPath.Text = _ExecutionOptions.EngineControllerConfig.LogFilePath;
            SetComboBoxSelectedValue(cbLogMode, TMSBEngineControllerConfiguration.LogModeToString(_ExecutionOptions.EngineControllerConfig.LogFileMode));
            chkIgnoreScriptErrors.Checked = !_ExecutionOptions.EngineControllerConfig.HaltOnErrors;
            chkKeepTemporaryFiles.Checked = !_ExecutionOptions.EngineControllerConfig.DeleteTemporaryFiles;
            chkAlwaysWriteScriptCode.Checked = _ExecutionOptions.EngineControllerConfig.WriteScriptCodeWhenNoImagesWillBeGenerated;
            chkDisplayThumbnails.Checked = _ExecutionOptions.EngineControllerConfig.DisplayImageThumbnails;
            SetComboBoxSelectedValue(cbOS, TMSBEngineControllerConfiguration.ExecutionPlatformModeToString(_ExecutionOptions.EngineControllerConfig.ExecutionPlatformMode));

        }

        protected bool SetExecutionOptionsFromUI(bool showPopupOnFailedParse)
        {
            bool proceed = true;

            // processing criteria
            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.MinimumSourceBandsInInputDirectory = int.Parse(txtMinimumSourceBands.Text);
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("The minimum number source bands must be a positive integer greater than 1. Please correct the value and try again.");
                    }
                }
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.ScanInputDirectoriesRecursively = pqgDirectoryQueue.AddItemsRecursively;
            }


            // config files
            if (proceed)
            {
                if (csInput.CurrentConfigFile != null)
                {
                    _ExecutionOptions.EngineControllerConfig.InputConfigFilePath = csInput.CurrentConfigFile.FilePath;
                }
            }

            if (proceed)
            {
                if (csProcessing.CurrentConfigFile != null)
                {
                    _ExecutionOptions.EngineControllerConfig.ProcessingConfigFilePath = csProcessing.CurrentConfigFile.FilePath;
                }
            }

            if (proceed)
            {
                if (csOutput.CurrentConfigFile != null)
                {
                    _ExecutionOptions.EngineControllerConfig.OutputConfigFilePath = csOutput.CurrentConfigFile.FilePath;
                }
            }



            // script processor
            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.GenerateScripts =
                        TMSBEngineControllerConfiguration.ScriptGenerationModeFromString(cbGenerateScripts.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the script generation mode from the settings in the UI. Please restart the application.");
                    }
                }
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.ExecuteScripts = chkExecuteScripts.Checked;
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.DaVinciCommand = txtScriptProcessorPath.Text;
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.DaVinciCommandArgs = txtScriptProcessorArgs.Text;
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.DaVinciProcessPriority =
                        TMSBEngineControllerConfiguration.ProcessPriorityClassFromString(cbScriptProcessPriority.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the script process priority from the settings in the UI. Please restart the application.");
                    }
                }
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.EngineThreadCount = int.Parse(txtMaxDirectoryThreads.Text);
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("The maximum number of simultaneous directories must be a positive integer. Please correct the value and try again.");
                    }
                }
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.GlobalThreadCount = int.Parse(txtMaxProcessingThreads.Text);
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("The maximum number of simultaneous images must be a positive integer. Please correct the value and try again.");
                    }
                }
            }


            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.ScriptExecutionTimeLimit = int.Parse(txtSubscriptTimeLimit.Text);
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("The subscript execution time limit must be an integer which is greater than or equal to -1. Please correct the value and try again.");
                    }
                }
            }
            


            // statistics
            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.CollectPerformanceData = chkCollectPerfData.Checked;
            }
            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.MemoryUnits =
                                        TMSBEngineControllerConfiguration.MemoryRepresentationFromString(cbMemoryFormat.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the memory representation mode from the settings in the UI. Please restart the application.");
                    }
                }
            }


            // debugging/troubleshooting
            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.OutputLevel =
                                        TMSBEngineControllerConfiguration.OutputLevelFromString(cbLogLevel.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the log level from the settings in the UI. Please restart the application.");
                    }
                }
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.LogFilePath = txtLogPath.Text;
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.LogFileMode =
                                        TMSBEngineControllerConfiguration.LogModeFromString(cbLogMode.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the log mode from the settings in the UI. Please restart the application.");
                    }
                }
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.HaltOnErrors = !chkIgnoreScriptErrors.Checked;
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.DeleteTemporaryFiles = !chkKeepTemporaryFiles.Checked;
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.WriteScriptCodeWhenNoImagesWillBeGenerated = chkAlwaysWriteScriptCode.Checked;
            }

            if (proceed)
            {
                _ExecutionOptions.EngineControllerConfig.DisplayImageThumbnails = chkDisplayThumbnails.Checked;
            }

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.ExecutionPlatformMode =
                                        TMSBEngineControllerConfiguration.ExecutionPlatformModeFromString(cbOS.SelectedItem.ToString());
                }
                catch (Exception ex)
                {
                    proceed = false;
                    if (showPopupOnFailedParse)
                    {
                        ShowOptionError("TMSB was unable to determine the operating system/execution platform mode from the settings in the UI. Please restart the application.");
                    }
                }
            }


            return proceed;
        }

        protected void ShowOptionError(string errorMessage)
        {
            MessageBox.Show(errorMessage, "Option Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        protected string GetLastUsedExecutionOptionsFilePath()
        {
            return TMSBConfigurationFile.GetPersonalConfigDirectoryName() +
                Path.DirectorySeparatorChar + _LastUsedOptionsFileName;
        }

        #region Config-to-UI (and vice versa) mappings

        protected void SetComboBoxSelectedValue(ComboBox target, string selectedValue)
        {
            int selectedIndex = -1;
            for (int i = 0; i < target.Items.Count; i++)
            {
                if (target.Items[i].ToString().Trim().ToLower() == selectedValue.Trim().ToLower())
                {
                    selectedIndex = i;
                }
            }
            if (selectedIndex > -1)
            {
                target.SelectedIndex = selectedIndex;
            }
        }

        #endregion

        #region Menu Items

        private void addDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pqgDirectoryQueue.DisplayAddDirectoryDialogue(false);
        }

        private void addDirectoryRecursivelyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pqgDirectoryQueue.DisplayAddDirectoryDialogue(true);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //Application.Exit();
            this.Close();
        }

        private void copyCurrentDirectoryToClipboardToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pqgDirectoryQueue.CopyCurrentNameToClipboard();
        }

        private void copyCurrentInforToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pqgDirectoryQueue.CopyCurrentInformationToClipboard();
        }

        private void copyAllQueueInformationToClipboardToolStripMenuItem_Click(object sender, EventArgs e)
        {
            pqgDirectoryQueue.CopyAllInformationToClipboard();
        }

        private void beginProcessingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            BeginProcessing();
        }

        private void resetExecutionOptionsToDefaultToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ResetExecutionOptionsByUserRequest();
        }

        private void reloadConfigurationFilesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ReloadConfigurationFiles();
        }

        private void aboutTheMirrorsSurfaceBreaksToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ShowAboutWindow();
        }

        #endregion


        #region Buttons

        private void btnScriptProcessorBrowse_Click(object sender, EventArgs e)
        {
            BrowseForScriptProcessor();
        }

        private void btnLogPathBrowse_Click(object sender, EventArgs e)
        {
            BrowseForLogFile();
        }

        private void btnBeginProcessing_Click(object sender, EventArgs e)
        {
            BeginProcessing();
        }

        #endregion


        #region Shared Actions

        protected void BrowseForScriptProcessor()
        {
            OpenFileDialog fd = new OpenFileDialog();
            fd.CheckFileExists = true;
            fd.CheckPathExists = true;
            fd.Multiselect = false;
            if (Directory.Exists(Path.GetDirectoryName(txtScriptProcessorPath.Text)))
            {
                fd.InitialDirectory = Path.GetDirectoryName(txtScriptProcessorPath.Text);
                if (File.Exists(txtScriptProcessorPath.Text))
                {
                    fd.FileName = Path.GetFileName(txtScriptProcessorPath.Text);
                }
            }
            DialogResult r = fd.ShowDialog();
            if ((r == DialogResult.OK) && (fd.FileName.Trim() != ""))
            {
                txtScriptProcessorPath.Text = fd.FileName;
            }
        }

        protected void BrowseForLogFile()
        {
            SaveFileDialog fd = new SaveFileDialog();
            fd.CheckFileExists = true;
            fd.CheckPathExists = true;
            if (Directory.Exists(Path.GetDirectoryName(txtLogPath.Text)))
            {
                fd.InitialDirectory = Path.GetDirectoryName(txtLogPath.Text);
                if (File.Exists(txtLogPath.Text))
                {
                    fd.FileName = Path.GetFileName(txtLogPath.Text);
                }
            }
            DialogResult r = fd.ShowDialog();
            if ((r == DialogResult.OK) && (fd.FileName.Trim() != ""))
            {
                txtLogPath.Text = fd.FileName;
            }
        }

        public void BeginProcessing()
        {
            //frmProcessingStatusTest testForm = new frmProcessingStatusTest();
            //testForm.Show();

            //return;

            //frmProcessingStatus testForm2 = new frmProcessingStatus();
            //testForm2.Show();

            //return;

            bool proceed = SetExecutionOptionsFromUI(true);

            // in the GUI, the recursion is handled before the hand-off to the processing engine
            bool currentRecurseSetting = _ExecutionOptions.EngineControllerConfig.ScanInputDirectoriesRecursively;
            _ExecutionOptions.EngineControllerConfig.ScanInputDirectoriesRecursively = false;

            if (proceed)
            {
                try
                {
                    _ExecutionOptions.EngineControllerConfig.InputDirectories = pqgDirectoryQueue.Items.GetDirectoryList();

                    _ExecutionOptions.EngineControllerConfig.CheckRequiredValues();



                    // do the processing
                    TMSBEngineController controller = new TMSBEngineController(_ExecutionOptions.EngineControllerConfig);
                    TMSBGUIEngineMonitor guiMonitor = new TMSBGUIEngineMonitor(controller, OutputLevel.Debug);

                    if (_ExecutionOptions.EngineControllerConfig.OutputLevel != OutputLevel.None)
                    {
                        TMSBCommandLineEngineMonitor clMonitor = new TMSBCommandLineEngineMonitor(controller, _ExecutionOptions.EngineControllerConfig.OutputLevel);
                        clMonitor.LogToConsole = false;
                        bool overwrite = true;
                        if (_ExecutionOptions.EngineControllerConfig.LogFileMode == LogMode.Append)
                        {
                            overwrite = false;
                        }
                        clMonitor.EnableFileLogging(_ExecutionOptions.EngineControllerConfig.LogFilePath, overwrite);
                        ShowProcessingWindow(guiMonitor, clMonitor);
                    }
                    else
                    {
                        ShowProcessingWindow(guiMonitor, null);
                    }
                    

                    //frmProcessingStatus statusForm = new frmProcessingStatus(this, guiMonitor);
                    //statusForm.Show();
                    
                    
                }
                catch (Exception ex)
                {
                    MessageBox.Show("TMSB could not begin processing because of the following error: '" + ex.Message +
                        "'. Please check all options and then try again.", "Unable to Process", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            // revert to the previous recursion setting
            _ExecutionOptions.EngineControllerConfig.ScanInputDirectoriesRecursively = currentRecurseSetting;
        }

        protected void SetupChildForm(Form childForm, bool centerChild)
        {
            childForm.Icon = this.Icon;
            //childForm.Owner = this;
            childForm.TopLevel = true;
            childForm.ShowInTaskbar = true;
            if (centerChild)
            {
                childForm.StartPosition = FormStartPosition.Manual;
                CenterChildForm(childForm, this);
            }
        }

        public void TriggerWaitWindowOpen(object sender, EventArgs e)
        {
            ShowWaitWindow();
        }

        public void TriggerWaitWindowClose(object sender, EventArgs e)
        {
            CloseWaitWindow();
        }

        protected void ShowWaitWindow()
        {
            //_WaitWindowShouldBeOpen = true;
            //this.Enabled = false;
            ////_WaitWindow.Show();
            ////_WaitWindow.ShowThreaded();

            //Thread wwThread = new Thread(new ThreadStart(ShowWaitWindowThreaded));
            //wwThread.Start();
            ShowWaitWindowFunction();
        }

        delegate void ShowWaitWindowCallback();

        protected void ShowWaitWindowFunction()
        {
            if (this.InvokeRequired)
            {
                ShowWaitWindowCallback c = new ShowWaitWindowCallback(ShowWaitWindowFunction);
                this.Invoke(c);
            }
            else
            {
                _WaitWindowShouldBeOpen = true;
                this.Enabled = false;

                Thread wwThread = new Thread(new ThreadStart(ShowWaitWindowThreaded));
                wwThread.Start();
            }
        }

        protected void ShowWaitWindowThreaded()
        {
            frmPleaseWait ww = new frmPleaseWait();
            SetupChildForm(ww, true);
            //ww.StartPosition = FormStartPosition.Manual;
            //CenterChildForm(ww, this);
            //Thread wwThread2 = new Thread(new ThreadStart(ww.Show));
            //wwThread2.Start();
            ////ww.Show();
            //ww.BringToFront();
            //ww.Activate();
            //ww.ShowThreaded();
            ww.Show();
            ww.BringToFront();
            ww.Activate();
            while (_WaitWindowShouldBeOpen == true)
            {
                Thread.Sleep(_WaitLengthWaitWindow);
                //ww.Invalidate();
            }
            //wwThread2.Abort();
            //ww.Close();
            ww.CloseThreaded();
            ww.Dispose();
            ww = null;
        }

        protected void CloseWaitWindow()
        {
            ////_WaitWindow.Close();
            ////_WaitWindow.CloseThreaded();
            //_WaitWindowShouldBeOpen = false;
            //this.Enabled = true;
            //this.BringToFront();
            //this.Activate();

            ////_WaitWindow.Dispose();
            ////_WaitWindow = null;
            CloseWaitWindowFunction();
        }

        delegate void CloseWaitWindowCallback();

        protected void CloseWaitWindowFunction()
        {
            if (this.InvokeRequired)
            {
                CloseWaitWindowCallback c = new CloseWaitWindowCallback(CloseWaitWindowFunction);
                this.Invoke(c);
            }
            else
            {
                _WaitWindowShouldBeOpen = false;
                this.Enabled = true;
                this.BringToFront();
                this.Activate();
            }
        }

        public void ShowAboutWindow()
        {
            _AboutWindow = new frmAbout();
            if (_AboutWindow.Visible)
            {
                _AboutWindow.Close();
            }
            _AboutWindow.FormClosed += new FormClosedEventHandler(_AboutWindow_FormClosed);
            //_AboutWindow.Icon = this.Icon;
            //_AboutWindow.Owner = this;
            //_AboutWindow.TopLevel = true;
            //_AboutWindow.ShowInTaskbar = true;
            //_AboutWindow.StartPosition = FormStartPosition.Manual;
            //CenterChildForm(_AboutWindow, this);
            SetupChildForm(_AboutWindow, true);
            this.Enabled = false;
            _AboutWindow.Show();

            //frmAbout about = new frmAbout();
            //about.Show();
        }

        void _AboutWindow_FormClosed(object sender, FormClosedEventArgs e)
        {
            this.Enabled = true;
            if (_AboutWindow != null)
            {
                _AboutWindow.Dispose();
            }
        }

        public void ShowProcessingWindow(TMSBGUIEngineMonitor guiMonitor, TMSBCommandLineEngineMonitor clMonitor)
        {
            //if (_ProcessingStatusWindow != null)
            //{
            //    _ProcessingStatusWindow.Dispose();
            //}
            _ProcessingStatusWindow = new frmProcessingStatus(this, guiMonitor, clMonitor);

            _ProcessingStatusWindow.FormClosed += _ProcessingStatusWindow_FormClosed;
            //_ProcessingStatusWindow.StartPosition = FormStartPosition.Manual;
            //CenterChildForm(_ProcessingStatusWindow, this);
            SetupChildForm(_ProcessingStatusWindow, false);
            this.Enabled = false;
            _PreviousWindowState = this.WindowState;
            this.SendToBack();
            this.WindowState = FormWindowState.Minimized;
            _ProcessingStatusWindow.Show();
            _ProcessingStatusWindow.WindowState = FormWindowState.Maximized;
            _ProcessingStatusWindow.Activate();
            _ProcessingStatusWindow.BringToFront();
        }

        void _ProcessingStatusWindow_FormClosed(object sender, FormClosedEventArgs e)
        {
            this.Enabled = true;
            this.WindowState = _PreviousWindowState;
            this.BringToFront();
            this.Activate();
            _ProcessingStatusWindow.FormClosed -= _ProcessingStatusWindow_FormClosed;
            if (_ProcessingStatusWindow != null)
            {
                _ProcessingStatusWindow.Dispose();
            }
        }

        public void ResetExecutionOptionsByUserRequest()
        {
            DialogResult result = MessageBox.Show("Reset all execution options to their default values? This action cannot be undone.",
                "Reset Execution Options?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            if (result == DialogResult.Yes)
            {
                SetDefaultExecutionOptions();
            }
        }

        public void ReloadConfigurationFiles()
        {
            try
            {
                csInput.PopulateDropDown();
                csProcessing.PopulateDropDown();
                csOutput.PopulateDropDown();
            }
            catch (Exception ex)
            {
                MessageBox.Show("An error occurred while reloading the configuration files. The error message was '" +
                    ex.Message + "'.", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        public void UpdateQueueItems(ProcessableItemCollection itemList)
        {
            UpdateQueueItemsFunction(itemList);
        }

        delegate void UpdateQueueItemsCallback(ProcessableItemCollection itemList);

        protected void UpdateQueueItemsFunction(ProcessableItemCollection itemList)
        {
            if (this.InvokeRequired)
            {
                UpdateQueueItemsCallback c = new UpdateQueueItemsCallback(UpdateQueueItemsFunction);
                this.Invoke(c, new object[] { itemList });
            }
            else
            {
                //pqgDirectoryQueue.Items = itemList;
                pqgDirectoryQueue.Items = new ProcessableItemCollection(itemList);
                pqgDirectoryQueue.ForceQueueUpdate();
            }
        }

        protected void CenterChildForm(Form childForm, Form parentForm)
        {
            // manually center the window in relation to its parent, since Windows doesn't do it correctly
            childForm.Location = new Point
            (
            parentForm.Location.X + ((parentForm.Width - childForm.Width) / 2),
            parentForm.Location.Y + ((parentForm.Height - childForm.Height) / 2)
            );
        }


        public void ProcessingConfigOptionChanged(object sender, EventArgs e)
        {
            if ((csInput.CurrentConfigFile != null) && 
                (csProcessing.CurrentConfigFile != null) && 
                (csOutput.CurrentConfigFile != null))
            {
                string result = "";
                try
                {
                    string tempDirectory = Path.GetTempPath();
                    TransformationProfileConfigurationFile tcFile =  new TransformationProfileConfigurationFile(TMSBEngineControllerConfiguration.GetTransformationProfileConfigFilePath());
                    TMSBColourConfigurationFile ccFile = new TMSBColourConfigurationFile(TMSBEngineControllerConfiguration.GetColourConfigFilePath());
                    TMSBInputConfigurationFile icFile =  new TMSBInputConfigurationFile(csInput.CurrentConfigFile.FilePath, tcFile.TransformationProfiles);
                    TMSBProcessingConfigurationFile pcFile = new TMSBProcessingConfigurationFile(csProcessing.CurrentConfigFile.FilePath, tcFile.TransformationProfiles);
                    TMSBOutputConfigurationFile ocFile = new TMSBOutputConfigurationFile(csOutput.CurrentConfigFile.FilePath, tcFile.TransformationProfiles);

                    ImageCube tempCube = ImageCube.GetTemporaryCube(tempDirectory, tcFile.TransformationProfiles, icFile.InputConfig, ccFile.ColourConfig, 
                        pcFile.ProcessingConfig, ocFile.OutputConfig);
                    tempCube.GenerateChannels();
                    tempCube.GeneratePermutations();
                    tempCube.GetSortedGreyscaleChannels();

                    int totalImages = 0;

                    foreach (GreyscaleChannelSet channelSet in tempCube.GreyscaleChannelSets)
                    {
                        totalImages += channelSet.Channels.Length;
                    }

                    totalImages += tempCube.ThreeChannelImages.Length;
                    totalImages += tempCube.TintedChannels.Length;
                    totalImages += tempCube.ColourGradientImages.Length;
                    totalImages += tempCube.DCSImages.Length;
                    totalImages += tempCube.LumaColourImages.Length;

                    tempCube = null;

                    int totalFiles = totalImages * ocFile.OutputConfig.OutputFormats.Length;

                    result = "Estimated variations based on the current input and processing configurations: " + totalImages.ToString() +
                        Environment.NewLine + "Estimated output file count based on the current configuration: " + totalFiles.ToString();

                }
                catch (Exception ex)
                {
                    result = "An error occurred calculating the number of variations: '" + ex.Message + "'";
                }

                csProcessing.DescriptionAppendText = result;
            }
        }



    }
}